/*****************************************************************************
 * Copyright (c) 2008, Mooneer Salem
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Kite Language organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "config.h"
#include "objs/kite_object.h"
#include "vm/kite_vm.h"

#ifdef HAVE_READLINE
#include <readline/readline.h>
#endif /* HAVE_READLINE */

#ifndef WIN32
#include <unistd.h>
#else
#include <XGetopt.h>
#endif /* WIN32 */

#undef PACKAGE_NAME
#define PACKAGE_NAME "ikt"

/*****************************************************************************
 * Output software version
 ****************************************************************************/
void version_short()
{
    printf("This is %s, version %s (built on %s)\n", PACKAGE_NAME, 
           PACKAGE_VERSION, TARGET_ARCH);
    printf("Copyright (C) 2009 Mooneer Salem. All rights reserved\n");
}

void version() {
    version_short();

    /* BSD License text below copyright notice */
    printf("\nRedistribution and use in source and binary forms, with or without\n");
    printf("modification, are permitted provided that the following conditions are met:\n");
    printf("    * Redistributions of source code must retain the above copyright\n");
    printf("      notice, this list of conditions and the following disclaimer.\n");
    printf("    * Redistributions in binary form must reproduce the above copyright\n");
    printf("      notice, this list of conditions and the following disclaimer in the\n");
    printf("      documentation and/or other materials provided with the distribution.\n");
    printf("    * Neither the name of the Kite Language organization nor the\n");
    printf("      names of its contributors may be used to endorse or promote products\n");
    printf("      derived from this software without specific prior written permission.\n\n");
    printf("THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY\n");
    printf("EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED\n");
    printf("WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE\n");
    printf("DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY\n");
    printf("DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES\n");
    printf("(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;\n");
    printf("LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND\n");
    printf("ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT\n");
    printf("(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF\n");
    printf("THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.\n");
}

/*****************************************************************************
 * Output help message
 ****************************************************************************/
void help() {
    version();
    printf("\n");
    printf("Usage: %s [-h|-v]\n", PACKAGE_NAME);
    printf("      -v: Print version information\n");
    printf("      -h: This message\n\n");
}

/*****************************************************************************
 * Fetch line from standard input.
 ****************************************************************************/
char *fetch_line(char *prompt)
{
    char *buf;
#ifdef HAVE_READLINE
    buf = readline(prompt);
    if (buf != NULL) add_history(buf);
    return buf;
#else
    char characters[2];
    if (feof(stdin)) return NULL;
    printf("%s", prompt);
    buf = calloc(2, 1);
    while(fgets(characters, 2, stdin) != NULL)
    {
#ifdef WIN32
        strncat_s(buf, strlen(buf) + 2, characters, 2);
#else
        strncat(buf, characters, strlen(buf) + 2);
#endif /* WIN32 */
        if (characters[0] == '\n') break;
        buf = realloc(buf, strlen(buf) + 2);
    }
    return buf;
#endif /* HAVE_READLINE */
}

/*****************************************************************************
 * Main application entry point.
 ****************************************************************************/
int main(int argc, char **argv) {
    int ch, num_braces = 0, cur_line = 1, tmp_line = 0, oldstack;
    kite_vm_t *vm;
    kite_object_t *obj;
    kite_thread_t *thd;
    char *line, *tmp;
#ifdef WIN32
    char *prompt = _strdup("ikt> ");
#else
    char *prompt = strdup("ikt> ");
#endif /* WIN32 */
    char *block = NULL;
    kite_funccall_info_t *entry, *fc;
    
#ifdef WIN32
    char buf[2048];
    char *file;
    GetModuleFileName(NULL, buf, 2048);
    file = strstr(buf, "ikt.exe");
    file[0] = 0;
    strncat_s(file, 2048, "modules\\:.\\modules\\:.\\", 2048);
    kite_set_dylib_path(buf);
#endif /* WIN32 */

    kite_app_init();
    
    /* Parse command arguments first */
#ifdef __linux
    while ((ch = getopt(argc, argv, "+hv")) != -1) {
#else
    while ((ch = getopt(argc, argv, "hv")) != -1) {
#endif /* LINUX */
        switch(ch) {
            case 'v':
                version();
                exit(0);
                break;
            case 'h':
            default:
                help();
                exit(0);
                break;
        }
    }
    argc -= optind;
    argv += optind;
    
    version_short();

    vm = kite_new_vm(argv);
    thd = kite_vm_creator_thread(vm);

#ifdef HAVE_GC_H
    entry = (void*)GC_malloc(sizeof(kite_funccall_info_t));
#else
    entry = calloc(sizeof(kite_funccall_info_t), 1);
#endif /* HAVE_GC_H */
    entry->this = NULL;
    entry->file = "(ikt)";
    kite_push_stack(&thd->func_stack, FALSE, entry);

    /* Enter loop. Read line(s), and run with Kite interpreter. */
    while((line = fetch_line(prompt)) != NULL)
    {
        tmp_line++;
    
        if (block != NULL) block = realloc(block, strlen(block) + strlen(line) + 1);
        else
        {
            block = malloc(strlen(line) + 1);
            block[0] = 0;
        }

#ifdef WIN32
        strncat_s(block, strlen(block) + strlen(line) + 1, line, strlen(line) + 1);
#else
        strncat(block, line, strlen(block) + strlen(line) + 1);
#endif

        /* Count number of braces, and fix prompt as appropriate. */
        for (tmp = line; *tmp; tmp++)
        {
            if (*tmp == '[') num_braces++;
            if (*tmp == ']') num_braces--;
        }
        if (num_braces > 0)
        {
            free(prompt);
            prompt = malloc(32);
#ifndef WIN32
            snprintf(prompt, 32, "ikt(%d)> ", num_braces);
#else
            sprintf_s(prompt, 32, "ikt(%d)> ", num_braces);
#endif /* WIN32 */
        }
        else
        {
            num_braces = 0;
#ifdef WIN32
            prompt = _strdup("ikt> ");
#else
            prompt = strdup("ikt> ");
#endif /* WIN32 */
            obj = kite_vm_compile_from_string(thd, block, cur_line);
            cur_line = tmp_line;
            free(block); block = NULL;
            if (thd->exception) 
            {   
                kite_vm_execute_exception(thd, FALSE);
                kite_dereference_object(thd->exception);
                thd->exception = NULL;

                do
                {
                    fc = kite_pop_stack(&thd->func_stack);
                    if (fc == entry) 
                    {
                        kite_push_stack(&thd->func_stack, FALSE, entry);
                        break;
                    }
#ifndef HAVE_GC_H
                    free(fc);
#endif /* HAVE_GC_H */
                } while (fc);
                
                continue;
            }
            
            thd->start = obj;
            if (thd->running_stack)
                oldstack = thd->running_stack->length - 1;
            else
                oldstack = 0;
            kite_vm_call_object(thd, thd->start, thd->start, kite_new_null(thd));
            if (thd->exception)
            {
                kite_vm_execute_exception(thd, FALSE);
                kite_dereference_object(thd->exception);
                thd->exception = NULL;
                
                do
                {
                    fc = kite_pop_stack(&thd->func_stack);
                    if (fc == entry) 
                    {
                        kite_push_stack(&thd->func_stack, FALSE, entry);
                        break;
                    }
#ifndef HAVE_GC_H
                    free(fc);
#endif /* HAVE_GC_H */
                } while (fc);
            }
            else
            {
                if (oldstack != (thd->running_stack->length - 1))
                {
                    kite_object_t *ret = kite_vm_pop(thd);
                    kite_object_t *retstr = kite_string_object(thd, ret);
                    printf("%s\n", KITE_GET_STRING_VALUE(retstr));
                    kite_dereference_object(retstr);
                    kite_dereference_object(ret);
                }
            }
            kite_dereference_object(obj);
        }
    }

    /* Wait for thread to stop, then free everything and quit. */
    printf("\n");
    kite_join_thread(vm, thd);
    kite_free_vm(&vm);

    return 0;
}
