/*****************************************************************************
 * Copyright (c) 2007, Mooneer Salem
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Kite Language organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/

#include "kite_object.h"
#include "vm/kite_vm.h"

/* Common code for the integer operations */
#define PERFORM_INTEGER_BOOLEAN_OPERATION(op) \
    result = this->builtin_data.intvalue op rhs->builtin_data.intvalue; \
    kite_vm_return(thd, kite_new_boolean(thd, result));
    
#define TYPECHECK_OPERANDS \
    kite_object_t *rhs; \
    kite_object_t *exc; \
    if (args->type == OBJ_LIST) { \
        rhs = args->builtin_data.listvalue.car; \
    } else { \
        rhs = args; \
    } \
    if (this->type != rhs->type) { \
        exc = kite_new_exception(thd, "System.exceptions.TypeMismatch", \
                            "Both operands must be of the same type."); \
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE); \
        return; \
    }

#define INTEGER_BOOLEAN_OPERATION(op) \
    { \
        int result; \
        TYPECHECK_OPERANDS \
        PERFORM_INTEGER_BOOLEAN_OPERATION(op) \
    }
    
KITE_CLASS_METHOD(Boolean_and) {
    INTEGER_BOOLEAN_OPERATION(&&);
}

KITE_CLASS_METHOD(Boolean_or) {
    INTEGER_BOOLEAN_OPERATION(||);
}

KITE_CLASS_METHOD(Boolean_xor) {
    int result;

    TYPECHECK_OPERANDS;
    result = 
        (this->builtin_data.intvalue && !rhs->builtin_data.intvalue) ||
        (!this->builtin_data.intvalue && rhs->builtin_data.intvalue);
    kite_vm_push(thd, kite_new_boolean(thd, result));
}

KITE_CLASS_METHOD(Boolean_equals) {
    INTEGER_BOOLEAN_OPERATION(==);
}

KITE_CLASS_METHOD(Boolean_notequals) {
    INTEGER_BOOLEAN_OPERATION(!=);
}

KITE_CLASS_METHOD(Boolean_not) {
    KITE_NO_ARGS;
    kite_vm_return(thd, kite_new_boolean(thd, !this->builtin_data.intvalue));
}

KITE_CLASS_METHOD(Boolean_str) {
    KITE_NO_ARGS;
    kite_vm_return(thd, kite_string_object(thd, this));
}

KITE_CLASS_METHOD(Boolean_int) {
    KITE_NO_ARGS;
    kite_vm_return(thd, kite_int_object(thd, this));
}

KITE_CLASS_METHOD(Boolean_float) {
    KITE_NO_ARGS;
    kite_vm_return(thd, kite_float_object(thd, this));
}

KITE_CLASS_METHOD(Boolean_bool) {
    KITE_NO_ARGS;
    kite_vm_return(thd, kite_reference_object(this));
}

KITE_MODULE_INITIALIZER(boolean) {
    /*
     * Create new class object. inherits is typically NULL for packages
     * and System.object.
     */
    kite_object_t *inherits = 
        kite_dereference_and_load(thread, "System.object");
    kite_object_t *newclass = 
        kite_new_class(thread, inherits, "boolean");
    kite_set_docstring(newclass, "Class representing Boolean values.");
    
    /*
     * Add operators
     */
    kite_add_operator(thread, newclass, OP_AND, 
        kite_new_method_compiled_with_docs(thread, Boolean_and,
                                  "True if both sides are true", 1,
                                  "rhs", "The right hand side of the expression."));
    kite_add_operator(thread, newclass, OP_OR, 
        kite_new_method_compiled_with_docs(thread, Boolean_or,
                                  "True if either or both sides are true", 1,
                                  "rhs", "The right hand side of the expression."));
    kite_add_operator(thread, newclass, OP_XOR, 
        kite_new_method_compiled_with_docs(thread, Boolean_xor,
                                  "True if either side is true (but false if both).", 1,
                                  "rhs", "The right hand side of the expression."));
    kite_add_operator(thread, newclass, OP_EQUALS, 
        kite_new_method_compiled_with_docs(thread, Boolean_equals,
                                  "True if both sides are equal to each other.", 1,
                                  "rhs", "The right hand side of the expression."));
    kite_add_operator(thread, newclass, OP_NOT_EQUALS, 
        kite_new_method_compiled_with_docs(thread, Boolean_notequals,
                                  "True if both sides are not equal to each other.", 1,
                                  "rhs", "The right hand side of the expression."));
    kite_add_operator(thread, newclass, OP_NOT, 
        kite_new_method_compiled_with_docs(thread, Boolean_not,
                                  "Returns the opposite value of the current object", 1,
                                  "rhs", "The right hand side of the expression (ignored)."));
    
    /* Add methods. */
    kite_add_method(thread, newclass, "str", 
        kite_new_method_compiled_with_docs(thread, Boolean_str, 
                                  "Returns the string representation of the given object.", 0));
    kite_add_method(thread, newclass, "int", 
        kite_new_method_compiled_with_docs(thread, Boolean_int,
                                  "Returns the integer representation of the given object.", 0));
    kite_add_method(thread, newclass, "float", 
        kite_new_method_compiled_with_docs(thread, Boolean_float,
                                  "Returns the floating-point representation of the given object.", 0));
    kite_add_method(thread, newclass, "bool", 
        kite_new_method_compiled_with_docs(thread, Boolean_bool,
                                  "Returns the given object.", 0));
        
    /* Add to parent to complete loading. */
    kite_add_property(thread, parent, "boolean", 1, NULL);
    kite_set_property(thread, parent, "boolean", newclass);
}

