/*****************************************************************************
 * Copyright (c) 2007, Mooneer Salem
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Kite Language organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/

#include "kite_object.h"
#include "vm/kite_vm.h"

/* Common code for the integer operations */
#define PERFORM_FLOAT_OPERATION(op) \
    result = \
        this->builtin_data.floatvalue op rhs->builtin_data.floatvalue; \
    kite_vm_return(thd, kite_new_float(thd, result));

#define PERFORM_FLOAT_BOOLEAN_OPERATION(op) \
    result = \
        this->builtin_data.floatvalue op rhs->builtin_data.floatvalue; \
    kite_vm_return(thd, kite_new_boolean(thd, result));
    
#define TYPECHECK_OPERANDS \
    kite_object_t *rhs; \
    kite_object_t *exc; \
    if (args->type == OBJ_LIST) { \
        rhs = args->builtin_data.listvalue.car; \
    } else { \
        rhs = args; \
    } \
    if (this->type != rhs->type) { \
        exc = kite_new_exception(thd, "System.exceptions.TypeMismatch", \
                            "Both operands must be of the same type."); \
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE); \
        return; \
    }

#define CHECK_DIVIDE_BY_ZERO \
    if (rhs->builtin_data.floatvalue == 0) { \
        exc = kite_new_exception(thd, "System.exceptions.DivideByZero", \
                            "Division by zero is not possible."); \
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE); \
        return; \
    }
    
#define FLOAT_OPERATION(op) \
    { \
        double result; \
        TYPECHECK_OPERANDS \
        PERFORM_FLOAT_OPERATION(op) \
    }

#define FLOAT_BOOLEAN_OPERATION(op) \
    { \
        int result; \
        TYPECHECK_OPERANDS \
        PERFORM_FLOAT_BOOLEAN_OPERATION(op) \
    }
    
#define FLOAT_DIVISION_OPERATION(op) \
    { \
        double result; \
        TYPECHECK_OPERANDS \
        CHECK_DIVIDE_BY_ZERO \
        PERFORM_FLOAT_OPERATION(op) \
    }
    
KITE_CLASS_METHOD(Float_add) {
    FLOAT_OPERATION(+);
}

KITE_CLASS_METHOD(Float_subtract) {
    FLOAT_OPERATION(-);
}

KITE_CLASS_METHOD(Float_multiply) {
    FLOAT_OPERATION(*);
}

KITE_CLASS_METHOD(Float_divide) {
    FLOAT_DIVISION_OPERATION(/);
}

KITE_CLASS_METHOD(Float_lessthan) {
    FLOAT_BOOLEAN_OPERATION(<);
}

KITE_CLASS_METHOD(Float_lessequals) {
    FLOAT_BOOLEAN_OPERATION(<=);
}

KITE_CLASS_METHOD(Float_equals) {
    FLOAT_BOOLEAN_OPERATION(==);
}

KITE_CLASS_METHOD(Float_greaterequals) {
    FLOAT_BOOLEAN_OPERATION(>=);
}

KITE_CLASS_METHOD(Float_greaterthan) {
    FLOAT_BOOLEAN_OPERATION(>);
}

KITE_CLASS_METHOD(Float_notequals) {
    FLOAT_BOOLEAN_OPERATION(!=);
}

KITE_CLASS_METHOD(Float_unaryplus) {
    int result = this->builtin_data.intvalue;

    KITE_NO_ARGS;

    if (result < 0) result = -result;
    kite_vm_return(thd, kite_new_integer(thd, result));
}

KITE_CLASS_METHOD(Float_unaryminus) {
    KITE_NO_ARGS;
    kite_vm_return(thd, kite_new_integer(thd, -this->builtin_data.intvalue));
}

KITE_CLASS_METHOD(Float_str) {
    KITE_NO_ARGS;
    kite_vm_return(thd, kite_string_object(thd, this));
}

KITE_CLASS_METHOD(Float_int) {
    KITE_NO_ARGS;
    kite_vm_return(thd, kite_int_object(thd, this));
}

KITE_CLASS_METHOD(Float_float) {
    KITE_NO_ARGS;
    kite_vm_return(thd, kite_reference_object(this));
}

KITE_CLASS_METHOD(Float_bool) {
    KITE_NO_ARGS;
    kite_vm_return(thd, kite_boolean_object(thd, this));
}

KITE_MODULE_INITIALIZER(float) {
    /*
     * Create new class object. inherits is typically NULL for packages
     * and System.object.
     */
    kite_object_t *inherits = 
        kite_dereference_and_load(thread, "System.object");
    kite_object_t *newclass = 
        kite_new_class(thread, inherits, "float");
    kite_set_docstring(newclass, "Class representing floating-point numbers.");
    
    /*
     * Add operators
     */
    kite_add_operator(thread, newclass, OP_ADD, 
        kite_new_method_compiled_with_docs(thread, Float_add, 
                                  "Adds two floating-point numbers together.", 1,
                                  "rhs", "The number to add to the current object."));
    kite_add_operator(thread, newclass, OP_SUBTRACT, 
        kite_new_method_compiled_with_docs(thread, Float_subtract, 
                                  "Subtracts two floating-point numbers.", 1,
                                  "rhs", "The number to subtract from the current object."));
    kite_add_operator(thread, newclass, OP_MULTIPLY, 
        kite_new_method_compiled_with_docs(thread, Float_multiply,
                                  "Multiplies two floating-point numbers.", 1,
                                  "rhs", "The number to multiply by."));
    kite_add_operator(thread, newclass, OP_DIVIDE, 
        kite_new_method_compiled_with_docs(thread, Float_divide,
                                  "Divides two floating-point numbers.", 1,
                                  "rhs", "The number to divide by."));
    kite_add_operator(thread, newclass, OP_LESS_THAN, 
        kite_new_method_compiled_with_docs(thread, Float_lessthan, 
                                  "Determines whether the current number is less than the given number.", 1,
                                  "rhs", "The number to compare the current number to."));
    kite_add_operator(thread, newclass, OP_LESS_OR_EQUALS, 
        kite_new_method_compiled_with_docs(thread, Float_lessequals,
                                  "Determines whether the current number is less than or equal to the given number.", 1,
                                  "rhs", "The number to compare the current number to."));
    kite_add_operator(thread, newclass, OP_EQUALS, 
        kite_new_method_compiled_with_docs(thread, Float_equals,
                                  "Determines whether the current number is equal to the given number.", 1,
                                  "rhs", "The number to compare the current number to."));
    kite_add_operator(thread, newclass, OP_GREATER_OR_EQUALS, 
        kite_new_method_compiled_with_docs(thread, Float_greaterequals,
                                  "Determines whether the current number is greater than or equal to the given number.", 1,
                                  "rhs", "The number to compare the current number to."));
    kite_add_operator(thread, newclass, OP_GREATER_THAN, 
        kite_new_method_compiled_with_docs(thread, Float_greaterthan,
                                  "Determines whether the current number is greater than the given number.", 1,
                                  "rhs", "The number to compare the current number to."));
    kite_add_operator(thread, newclass, OP_NOT_EQUALS, 
        kite_new_method_compiled_with_docs(thread, Float_notequals,
                                  "Determines whether the current number is not equal to the given number.", 1,
                                   "rhs", "The number to compare the current number to."));
    kite_add_operator(thread, newclass, OP_UNARY_PLUS, 
        kite_new_method_compiled_with_docs(thread, Float_unaryplus,
                                  "Makes current number positive.", 1,
                                  "rhs", "Ignored."));
    kite_add_operator(thread, newclass, OP_UNARY_MINUS, 
        kite_new_method_compiled_with_docs(thread, Float_unaryminus,
                                  "Inverts sign of the current number.", 1,
                                  "rhs", "Ignored."));
    
    /* Add methods. */
    kite_add_method(thread, newclass, "str", 
        kite_new_method_compiled_with_docs(thread, Float_str, 
                                  "Returns string representation of current object.", 0));
    kite_add_method(thread, newclass, "int", 
        kite_new_method_compiled_with_docs(thread, Float_int,
                                  "Returns integer representation of current object.", 0));
    kite_add_method(thread, newclass, "float", 
        kite_new_method_compiled_with_docs(thread, Float_float,
                                  "Returns floating-point representation of current object.", 0));
    kite_add_method(thread, newclass, "bool", 
        kite_new_method_compiled_with_docs(thread, Float_bool,
                                  "Returns Boolean representation of current object.", 0));
        
    /* Add to parent to complete loading. */
    kite_add_property(thread, parent, "float", 1, NULL);
    kite_set_property(thread, parent, "float", newclass);
}

