/*****************************************************************************
 * Copyright (c) 2007, Mooneer Salem
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Kite Language organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/

#include "kite_object.h"
#include "vm/kite_vm.h"

/* Common code for the integer operations */
#define PERFORM_INTEGER_OPERATION(op) \
    result = this->builtin_data.intvalue op rhs->builtin_data.intvalue; \
    kite_vm_return(thd, kite_new_integer(thd, result));

#define PERFORM_INTEGER_BOOLEAN_OPERATION(op) \
    result = this->builtin_data.intvalue op rhs->builtin_data.intvalue; \
    kite_vm_return(thd, kite_new_boolean(thd, result));
    
#define TYPECHECK_OPERANDS \
    kite_object_t *rhs; \
    kite_object_t *exc; \
    if (args->type == OBJ_LIST) { \
        rhs = args->builtin_data.listvalue.car; \
    } else { \
        rhs = args; \
    } \
    if (this->type != rhs->type) { \
        exc = kite_new_exception(thd, "System.exceptions.TypeMismatch", \
                            "Both operands must be of the same type."); \
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE); \
        return; \
    }

#define CHECK_DIVIDE_BY_ZERO \
    if (rhs->builtin_data.intvalue == 0) { \
        exc = kite_new_exception(thd, "System.exceptions.DivideByZero", \
                            "Division by zero is not possible."); \
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE); \
        return; \
    }
    
#define INTEGER_OPERATION(op) \
    { \
        int result; \
        TYPECHECK_OPERANDS \
        PERFORM_INTEGER_OPERATION(op) \
    }

#define INTEGER_BOOLEAN_OPERATION(op) \
    { \
        int result; \
        TYPECHECK_OPERANDS \
        PERFORM_INTEGER_BOOLEAN_OPERATION(op) \
    }
    
#define INTEGER_DIVISION_OPERATION(op) \
    { \
        int result; \
        TYPECHECK_OPERANDS \
        CHECK_DIVIDE_BY_ZERO \
        PERFORM_INTEGER_OPERATION(op) \
    }
    
KITE_CLASS_METHOD(Integer_add) {
    INTEGER_OPERATION(+);
}

KITE_CLASS_METHOD(Integer_subtract) {
    INTEGER_OPERATION(-);
}

KITE_CLASS_METHOD(Integer_multiply) {
    INTEGER_OPERATION(*);
}

KITE_CLASS_METHOD(Integer_divide) {
    INTEGER_DIVISION_OPERATION(/);
}

KITE_CLASS_METHOD(Integer_modulus) {
    INTEGER_DIVISION_OPERATION(%);
}

KITE_CLASS_METHOD(Integer_leftshift) {
    INTEGER_OPERATION(<<);
}

KITE_CLASS_METHOD(Integer_rightshift) {
    INTEGER_OPERATION(>>);
}

KITE_CLASS_METHOD(Integer_and) {
    INTEGER_OPERATION(&);
}

KITE_CLASS_METHOD(Integer_or) {
    INTEGER_OPERATION(|);
}

KITE_CLASS_METHOD(Integer_xor) {
    INTEGER_OPERATION(^);
}

KITE_CLASS_METHOD(Integer_lessthan) {
    INTEGER_BOOLEAN_OPERATION(<);
}

KITE_CLASS_METHOD(Integer_lessequals) {
    INTEGER_BOOLEAN_OPERATION(<=);
}

KITE_CLASS_METHOD(Integer_equals) {
    INTEGER_BOOLEAN_OPERATION(==);
}

KITE_CLASS_METHOD(Integer_greaterequals) {
    INTEGER_BOOLEAN_OPERATION(>=);
}

KITE_CLASS_METHOD(Integer_greaterthan) {
    INTEGER_BOOLEAN_OPERATION(>);
}

KITE_CLASS_METHOD(Integer_notequals) {
    INTEGER_BOOLEAN_OPERATION(!=);
}

KITE_CLASS_METHOD(Integer_not) {
    KITE_NO_ARGS;
    kite_vm_return(thd, kite_new_integer(thd, !this->builtin_data.intvalue));
}

KITE_CLASS_METHOD(Integer_unaryplus) {
    int result = this->builtin_data.intvalue;

    KITE_NO_ARGS;

    if (result < 0) result = -result;
    kite_vm_return(thd, kite_new_integer(thd, result));
}

KITE_CLASS_METHOD(Integer_unaryminus) {
    KITE_NO_ARGS;
    kite_vm_return(thd, kite_new_integer(thd, -this->builtin_data.intvalue));
}

KITE_CLASS_METHOD(Integer_str) {
    KITE_NO_ARGS;
    kite_vm_return(thd, kite_string_object(thd, this));
}

KITE_CLASS_METHOD(Integer_int) {
    KITE_NO_ARGS;
    kite_vm_return(thd, kite_reference_object(this));
}

KITE_CLASS_METHOD(Integer_float) {
    KITE_NO_ARGS;
    kite_vm_return(thd, kite_float_object(thd, this));
}

KITE_CLASS_METHOD(Integer_bool) {
    KITE_NO_ARGS;
    kite_vm_return(thd, kite_boolean_object(thd, this));
}

KITE_CLASS_METHOD(Integer_parse) {
    kite_object_t *str, *base;
    int res;
    (void)this;
    
    KITE_GET_METHOD_ARGUMENT(str, 1); 
    KITE_GET_METHOD_ARGUMENT(base, 2); 

    res = strtol(KITE_GET_STRING_VALUE(str), NULL, KITE_GET_INTEGER(base));
    kite_vm_return(thd, kite_new_integer(thd, res));
}

KITE_MODULE_INITIALIZER(integer) {
    /*
     * Create new class object. inherits is typically NULL for packages
     * and System.object.
     */
    kite_object_t *inherits = 
        kite_dereference_and_load(thread, "System.object");
    kite_object_t *newclass = 
        kite_new_class(thread, inherits, "integer");
    kite_set_docstring(newclass, "Class representing integer objects.");
    
    /*
     * Add operators
     */
    kite_add_operator(thread, newclass, OP_ADD, 
        kite_new_method_compiled_with_docs(thread, Integer_add, 
                                  "Adds two numbers.", 1,
                                  "rhs", "The number to add to the current number."));
    kite_add_operator(thread, newclass, OP_SUBTRACT, 
        kite_new_method_compiled_with_docs(thread, Integer_subtract,
                                  "Subtract two numbers.", 1,
                                  "rhs", "The number to subtract from the current number."));
    kite_add_operator(thread, newclass, OP_MULTIPLY, 
        kite_new_method_compiled_with_docs(thread, Integer_multiply, 
                                  "Multiplies two numbers.", 1,
                                  "rhs", "The number to multiply to the current number."));
    kite_add_operator(thread, newclass, OP_DIVIDE, 
        kite_new_method_compiled_with_docs(thread, Integer_divide,
                                  "Divides two numbers.", 1,
                                   "rhs", "The number to divide with the current number."));
    kite_add_operator(thread, newclass, OP_MODULUS, 
        kite_new_method_compiled_with_docs(thread, Integer_modulus,
                                  "Returns the modulus of two numbers.", 1,
                                   "rhs", "The number to divide with the current number."));
    kite_add_operator(thread, newclass, OP_LEFT_SHIFT, 
        kite_new_method_compiled_with_docs(thread, Integer_leftshift,
                                  "Left-shift current number by given number of bits.", 1,
                                  "bits", "The number of bits to shift the current number by."));
    kite_add_operator(thread, newclass, OP_RIGHT_SHIFT, 
        kite_new_method_compiled_with_docs(thread, Integer_rightshift,
                                  "Right-shift current number by given number of bits.", 1,
                                  "bits", "The number of bits to shift the current number by."));
    kite_add_operator(thread, newclass, OP_AND, 
        kite_new_method_compiled_with_docs(thread, Integer_and,
                                  "Return the bitwise AND of two numbers.", 1,
                                  "rhs", "The number to perform the operation with."));
    kite_add_operator(thread, newclass, OP_OR, 
        kite_new_method_compiled_with_docs(thread, Integer_or,
                                  "Return the bitwise OR of two numbers.", 1,
                                  "rhs", "The number to perform the operation with."));
    kite_add_operator(thread, newclass, OP_XOR, 
        kite_new_method_compiled_with_docs(thread, Integer_xor, 
                                  "Return the bitwise XOR of two numbers.", 1,
                                  "rhs", "The number to perform the operation with."));
    kite_add_operator(thread, newclass, OP_LESS_THAN, 
        kite_new_method_compiled_with_docs(thread, Integer_lessthan, 
                                  "Returns true if the current number is less than the given number.", 1,
                                  "rhs", "The number to compare to."));
    kite_add_operator(thread, newclass, OP_LESS_OR_EQUALS, 
        kite_new_method_compiled_with_docs(thread, Integer_lessequals,
                                  "Returns true if the current number is less than or equal to the given number.", 1,
                                  "rhs", "The number to compare to."));
    kite_add_operator(thread, newclass, OP_EQUALS, 
        kite_new_method_compiled_with_docs(thread, Integer_equals,
                                  "Returns true if the current number is equal to the given number.", 1,
                                  "rhs", "The number to compare to."));
    kite_add_operator(thread, newclass, OP_GREATER_OR_EQUALS, 
        kite_new_method_compiled_with_docs(thread, Integer_greaterequals,
                                  "Returns true if the current number is greater than or equal to the given number.", 1,
                                  "rhs", "The number to compare to."));
    kite_add_operator(thread, newclass, OP_GREATER_THAN, 
        kite_new_method_compiled_with_docs(thread, Integer_greaterthan,
                                  "Returns true if the current number is greater than the given number.", 1,
                                  "rhs", "The number to compare to."));
    kite_add_operator(thread, newclass, OP_NOT_EQUALS, 
        kite_new_method_compiled_with_docs(thread, Integer_notequals,
                                  "Returns true if the current number is not equal to the given number.", 1,
                                  "rhs", "The number to compare to."));
    kite_add_operator(thread, newclass, OP_NOT, 
        kite_new_method_compiled_with_docs(thread, Integer_not,
                                  "Returns the bitwise inverse of the current number.", 1,
                                  "rhs", "Ignored."));
    kite_add_operator(thread, newclass, OP_UNARY_PLUS, 
        kite_new_method_compiled_with_docs(thread, Integer_unaryplus,
                                  "Returns the given number made positive.", 1,
                                  "rhs", "Ignored."));
    kite_add_operator(thread, newclass, OP_UNARY_MINUS, 
        kite_new_method_compiled_with_docs(thread, Integer_unaryminus,
                                  "Inverts the given number's sign, then returns the result.", 1,
                                  "rhs", "Ignored."));
    
    /* Add methods. */
    kite_add_method(thread, newclass, "str", 
        kite_new_method_compiled_with_docs(thread, Integer_str, 
                                  "Returns string representation of current object.", 0));
    kite_add_method(thread, newclass, "int", 
        kite_new_method_compiled_with_docs(thread, Integer_int,
                                  "Returns integer representation of current object.", 0));
    kite_add_method(thread, newclass, "float", 
        kite_new_method_compiled_with_docs(thread, Integer_float,
                                  "Returns floating-point representation of current object.", 0));
    kite_add_method(thread, newclass, "bool", 
        kite_new_method_compiled_with_docs(thread, Integer_bool,
                                  "Returns Boolean representation of current object.", 0));
    kite_add_method(thread, newclass, "parse", 
        kite_new_method_compiled_with_docs(thread, Integer_parse, 
                                  "Returns integer extracted from string.", 2,
                                  "str", "String to parse.",
                                  "base", "The base the number is in."));
        
    /* Add to parent to complete loading. */
    kite_add_property(thread, parent, "integer", 1, NULL);
    kite_set_property(thread, parent, "integer", newclass);
}

