/*****************************************************************************
 * Copyright (c) 2007, Mooneer Salem
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Kite Language organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/

#include <string.h>
#include "kite_object.h"
#include "vm/kite_vm.h"

#define TYPECHECK_OPERAND(v, t) \
    { \
        kite_object_t *rhs = v; \
        kite_object_t *exc; \
        if (rhs->type != t) { \
            exc = kite_new_exception(thd, "System.exceptions.TypeMismatch", \
                                "Second operand is not of the requested type."); \
            kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE); \
            return; \
        } \
    }

/*****************************************************************************
 * Get list of method parameters.
 ****************************************************************************/  
KITE_CLASS_METHOD(Method_getParamNames) {
    kite_object_t *paramList = kite_new_list(thd), *tmpParam = paramList;
    kite_object_t *params = this->builtin_data.funcvalue.arginfo;
    KITE_NO_ARGS;
    
    while(params && params->builtin_data.listvalue.car) {
        tmpParam->builtin_data.listvalue.car = 
            kite_new_string(thd, KITE_GET_STRING_VALUE(params->builtin_data.listvalue.car));
        
        if (params->builtin_data.listvalue.cdr) {
            tmpParam->builtin_data.listvalue.cdr = kite_new_list(thd);
            tmpParam = tmpParam->builtin_data.listvalue.cdr;
        }
        
        params = params->builtin_data.listvalue.cdr;
    }
    
    kite_vm_push(thd, paramList);
    kite_dereference_object(paramList);
}

/*****************************************************************************
 * Get docstring for a method parameter
 ****************************************************************************/
KITE_CLASS_METHOD(Method_getParamDoc) {
    kite_object_t *ret, *name, *tmp;
    KITE_NO_ARGS;
    
    KITE_GET_METHOD_ARGUMENT(name, 1);
    TYPECHECK_OPERAND(name, OBJ_STRING);
    
    tmp = this->builtin_data.funcvalue.arginfo;
    while(tmp) {
        ret = tmp->builtin_data.listvalue.car;
        
        if (!strcmp(KITE_GET_STRING_VALUE(ret), KITE_GET_STRING_VALUE(name))) {
            ret = kite_new_string(thd, kite_get_docstring(ret));
            kite_vm_push(thd, ret);
            kite_dereference_object(ret);
            return;
        }
        
        tmp = tmp->builtin_data.listvalue.cdr;
    }

    ret = kite_new_null(thd);
    kite_vm_push(thd, ret);
    kite_dereference_object(ret);
}

KITE_MODULE_INITIALIZER(method) {
    /*
     * Create new class object. inherits is typically NULL for packages
     * and System.object.
     */
    kite_object_t *inherits = 
        kite_dereference_and_load(thread, "System.object");
    kite_object_t *newclass = 
        kite_new_class(thread, inherits, "method");
    kite_set_docstring(newclass, "Object representing a block of executable code.");
    
    /* Add to parent, since we need this class to "exist" for the
     * remaining definitions. */
    kite_add_property(thread, parent, "method", 1, NULL);
    kite_set_property(thread, parent, "method", newclass);
    
    kite_add_method(thread, newclass, "get_param_names",
        kite_new_method_compiled_with_docs(thread, Method_getParamNames, 
                                  "Gets the names of this method's parameters.", 0));
    kite_add_method(thread, newclass, "get_param_doc",
        kite_new_method_compiled_with_docs(thread, Method_getParamDoc, 
                                  "Gets the documentation string for a given parameter.", 1,
                                  "param", "Name of parameter to retrieve documentation string for."));
}

