/*****************************************************************************
 * Copyright (c) 2007, Mooneer Salem
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Kite Language organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/

#include <stdlib.h>
#include <string.h>
#ifndef WIN32
#include <unistd.h>
#else
#include <io.h>
#endif /* WIN32 */
#include <errno.h>
#include "kite_object.h"
#include "vm/kite_vm.h"

#define CHECK_SOCKET_OPEN \
    if (this->builtin_data.intvalue < 0) { \
        kite_object_t *exc = \
            kite_new_exception(thd, "System.exceptions.SocketError", \
                               "Socket is not open"); \
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE); \
        return; \
    }

#define THROW_SOCKET_ERROR \
    { \
        char *buf; \
        kite_object_t *exc; \
        buf = strerror(errno); \
        exc =  \
            kite_new_exception(thd, "System.exceptions.SocketError",  \
                               buf); \
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE); \
        return; \
    }

#define SET_SOCK_CONSTANT(c) \
    kite_add_property(thread, newclass, # c, 1, # c); \
    kite_set_property(thread, newclass, # c, kite_new_integer(thread, c));  

#ifndef WIN32
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <netdb.h>
#else
#define WIN32_LEAN_AND_MEAN
/*#define _WIN32_WINNT 0x501*/
#include <winsock2.h>
#include <ws2tcpip.h>
#endif

/*****************************************************************************
 * Class constructor
 ****************************************************************************/
KITE_CLASS_METHOD(Socket_construct) {
    kite_object_t *domain, *type, *protocol;

    KITE_GET_METHOD_ARGUMENT(domain, 1);
    KITE_GET_METHOD_ARGUMENT(type, 2);
    KITE_GET_METHOD_ARGUMENT(protocol, 3);

    if (!KITE_IS_TYPE(domain, OBJ_INTEGER) || 
        !KITE_IS_TYPE(type, OBJ_INTEGER) || 
        !KITE_IS_TYPE(protocol, OBJ_INTEGER))
    {
        kite_object_t *exc = 
            kite_new_exception(thd, "System.exceptions.TypeMismatch", 
                               "All arguments must be integers");
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);
        return;
    }

    /*
     * Yes, we know it's not an integer, but we store the socket number
     * in intvalue.
     */
    KITE_SET_INTEGER(this, socket(KITE_GET_INTEGER(domain), 
                     KITE_GET_INTEGER(type), KITE_GET_INTEGER(protocol)));
    if (KITE_GET_INTEGER(this) < 0) 
        THROW_SOCKET_ERROR;
}

/*****************************************************************************
 * Write bytes to socket
 ****************************************************************************/
KITE_CLASS_METHOD(Socket_write) {
    kite_object_t *arg1 = args->builtin_data.listvalue.car;
    kite_object_t *bytes = kite_string_object(thd, arg1);
    int ret;

    CHECK_SOCKET_OPEN;
    ret = write(KITE_GET_INTEGER(this), KITE_GET_STRING_VALUE(bytes), 
                    KITE_GET_STRING_LENGTH(bytes));
    if (ret < 0)
        THROW_SOCKET_ERROR;

    kite_dereference_object(bytes);
    kite_vm_return(thd, kite_new_integer(thd, ret));
}
/*****************************************************************************
 * Read up to the given number of bytes, also with a timeout supplied
 ****************************************************************************/
KITE_CLASS_METHOD(Socket_read_with_timeout) {
    
    kite_object_t *bytes;
    kite_object_t *timeout;
    struct timeval timeout_struct;
    kite_object_t *rval;
    fd_set socklist;
    
    char *buf;
    int ret, can_read;

    KITE_GET_METHOD_ARGUMENT(bytes,1);
    KITE_GET_METHOD_ARGUMENT(timeout, 2);
    
    buf = calloc(bytes->builtin_data.intvalue + 1, 1);
    
    timeout_struct.tv_sec = KITE_GET_INTEGER(timeout) / 1000;
    timeout_struct.tv_usec = (float)(KITE_GET_INTEGER(timeout) % 1000) / 1000.0f;

    CHECK_SOCKET_OPEN;
    
    
    FD_ZERO(&socklist);
    FD_SET(KITE_GET_INTEGER(this),&socklist);
    can_read = select(FD_SETSIZE,&socklist,NULL,NULL,&timeout_struct);
    if (can_read <= 0) {
      free(buf);
      kite_vm_return(thd, kite_new_null(thd));
      return;
    }
    
    ret = read(KITE_GET_INTEGER(this), buf, KITE_GET_STRING_LENGTH(bytes));
    if (ret < 0 && errno != EAGAIN) {
        free(buf);
        THROW_SOCKET_ERROR;
    } else if (ret == 0) {
        free(buf);
        kite_vm_return(thd, kite_new_null(thd));
        return;
    }

    rval = kite_new_string_with_length(
            thd, 
            buf, 
            KITE_GET_INTEGER(bytes)
        );
    free(buf);
    kite_dereference_object(bytes);
    kite_vm_return(thd, rval);
}
/*************************************************************************
 * Check if the socket has data.
****************************************************************************/
KITE_CLASS_METHOD(Socket_eof) {
  kite_object_t *timeout;
  struct timeval timeout_struct;
  fd_set socklist;
  
  int can_read;

  KITE_GET_METHOD_ARGUMENT(timeout, 1);
  if (timeout == NULL || timeout->type == OBJ_NULL) {
    timeout_struct.tv_sec = 1;
    timeout_struct.tv_usec = 0;
  }
  else
  {
    timeout_struct.tv_sec = KITE_GET_INTEGER(timeout) / 1000;
    timeout_struct.tv_usec = (float)(KITE_GET_INTEGER(timeout) % 1000) / 1000.0f;
  }
  
  CHECK_SOCKET_OPEN;
  
  FD_ZERO(&socklist);
  FD_SET(KITE_GET_INTEGER(this),&socklist);
  can_read = select(FD_SETSIZE,&socklist,NULL,NULL,&timeout_struct);
  if (can_read <= 0) {
    kite_vm_return(thd, kite_new_boolean(thd,TRUE));
    return;
  }
  else {
    kite_vm_return(thd, kite_new_boolean(thd,FALSE));
    return;
  }
}   
/*****************************************************************************
 * Read up to the given number of bytes
 ****************************************************************************/
KITE_CLASS_METHOD(Socket_read) {
    kite_object_t *arg1 = args->builtin_data.listvalue.car;
    kite_object_t *bytes = kite_int_object(thd, arg1);
    kite_object_t *rval;
    char *buf = calloc(bytes->builtin_data.intvalue + 1, 1);
    int ret;
    
    CHECK_SOCKET_OPEN;

    ret = read(KITE_GET_INTEGER(this), buf, KITE_GET_INTEGER(bytes));
    if (ret < 0 && errno != EAGAIN) {
        free(buf);
        THROW_SOCKET_ERROR;
    } else if (ret == 0) {
        free(buf);
        kite_vm_return(thd, kite_new_null(thd));
        return;
    }

    rval = kite_new_string_with_length(
            thd, 
            buf, 
            ret
        );
    free(buf);
    kite_dereference_object(bytes);
    kite_vm_return(thd, rval);
}

/*****************************************************************************
 * Read a line of text from the socket.
 ****************************************************************************/
KITE_CLASS_METHOD(Socket_readline) {
    char *buf = NULL;
    int ret, idx = 0;
    kite_object_t *rval;
    // int can_read;
    // struct timeval timeout;
    // fd_set socklist;

    CHECK_SOCKET_OPEN;
    KITE_NO_ARGS;

    buf = calloc(1, 1);

    do {
        ret = read(KITE_GET_INTEGER(this), &buf[idx], 1);
        if (ret < 0 && errno != EAGAIN) {
            free(buf);
            THROW_SOCKET_ERROR;
        } else if (ret == 0 || errno == EAGAIN) {
            break;
        }
        idx++;
        buf = realloc(buf, idx + 1);
    } while(buf[idx - 1] != '\n');
    buf[idx] = 0;

    rval = kite_new_string(thd, buf);
    free(buf);
    kite_vm_return(thd, rval);
}

/*****************************************************************************
 * Close given file
 ****************************************************************************/
KITE_CLASS_METHOD(Socket_close) {
    int ret;

    KITE_NO_ARGS;

    CHECK_SOCKET_OPEN;
    ret = close(KITE_GET_INTEGER(this));
    if (ret < 0)
        THROW_SOCKET_ERROR;

    KITE_SET_INTEGER(this, 0);
}

/*****************************************************************************
 * Accept new connection
 ****************************************************************************/
KITE_CLASS_METHOD(Socket_accept) {
    struct sockaddr_storage addr;
    socklen_t len = sizeof(addr);
    int ret;
    char hostname[1024];
    kite_object_t *isUnix, *retval, *sock, *tmpret;

    KITE_NO_ARGS;
    CHECK_SOCKET_OPEN;

    ret = accept(KITE_GET_INTEGER(this), 
        (struct sockaddr*)&addr, &len);
    if (ret < 0) 
        THROW_SOCKET_ERROR;

    isUnix = kite_get_property(this, "unix_socket");
    
    if (!KITE_GET_INTEGER(isUnix))
        getnameinfo((struct sockaddr*)&addr, sizeof(addr), hostname, 1024, NULL,
                    0, NI_NUMERICHOST);
    else
        strcpy(hostname, "");

    retval = kite_new_list(thd);
    sock = kite_new_instance(thd, this->object_data.inherit_from);
    KITE_SET_INTEGER(sock, ret);
    retval->builtin_data.listvalue.car = sock;
    retval->builtin_data.listvalue.cdr = kite_new_list(thd);
    tmpret = retval->builtin_data.listvalue.cdr;
    tmpret->builtin_data.listvalue.car = kite_new_string(thd, hostname);
    tmpret->builtin_data.listvalue.cdr = kite_new_list(thd);
    tmpret = tmpret->builtin_data.listvalue.cdr;
    if (!KITE_GET_INTEGER(isUnix))
    {
        int port = 0;
        
        switch(addr.ss_family)
        {
            case AF_INET:
            {
                struct sockaddr_in *sin = (struct sockaddr_in*)&addr;
                port = sin->sin_port;
                break;
            }
            case AF_INET6:
            {
                struct sockaddr_in6 *sin = (struct sockaddr_in6*)&addr;
                port = sin->sin6_port;
                break;
            }
            default:
            {
                assert(0);
            }
        }
        tmpret->builtin_data.listvalue.car = kite_new_integer(thd, port);
    }
    else
        tmpret->builtin_data.listvalue.car = kite_new_integer(thd, 0);
        
    kite_vm_return(thd, retval);
}

/*****************************************************************************
 * Listen for connections on this socket
 ****************************************************************************/
KITE_CLASS_METHOD(Socket_listen) {
    kite_object_t *backlog;
    int ret;

    KITE_GET_METHOD_ARGUMENT(backlog, 1);

    if (!KITE_IS_TYPE(backlog, OBJ_INTEGER))
    {
        kite_object_t *exc = 
            kite_new_exception(thd, "System.exceptions.TypeMismatch", 
                               "Argument must be an integer");
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);
        return;
    }

    ret = listen(KITE_GET_INTEGER(this), KITE_GET_INTEGER(backlog));
    if (ret < 0)
        THROW_SOCKET_ERROR;
}

/*****************************************************************************
 * Perform DNS lookup
 ****************************************************************************/
static int 
Socket_dnsLookup(kite_thread_t *thd, kite_object_t *this, int sock, 
                 char *host, int port, int mode)
{
    struct sockaddr *sa;
    struct addrinfo hints;
    struct addrinfo *res;
    int ret = 0, error;
    size_t salen;
    char buf[128];
    int socktype;
    socklen_t len = sizeof(socklen_t);
#ifndef WIN32
    struct sockaddr_un local;
#endif /* WIN32 */

    /* Unix domain sockets */
    kite_add_property(thd, this, "unix_socket", 0, "Whether the socket is a Unix domain socket.");
#ifndef WIN32
    if (host[0] == '/')
    {
        kite_set_property(thd, this, "unix_socket", kite_new_boolean(thd, TRUE));
                
        local.sun_family = AF_UNIX;
        strcpy(local.sun_path, host);
        
        if (mode)
        {
            unlink(local.sun_path);
            ret = bind(sock, (struct sockaddr*)&local, sizeof(local));
        }
        else
        {
            ret = connect(sock, (struct sockaddr*)&local, sizeof(local));
        }
    }
    else
    {
#endif
        kite_set_property(thd, this, "unix_socket", kite_new_boolean(thd, FALSE));
        
        /* AF_INET[6] */
        memset(&hints, 0, sizeof(hints));
        sprintf(buf, "%d", port);
    
        getsockopt(sock, SOL_SOCKET, SO_TYPE, (char*)&socktype, &len);
    
        /* set-up hints structure */
        hints.ai_family = AF_UNSPEC;
        hints.ai_socktype = socktype;
        if (mode) hints.ai_flags = AI_PASSIVE;
        error = getaddrinfo(host, buf, &hints, &res);
        if (error)
        {
            kite_object_t *exc = 
                kite_new_exception(thd, "System.exceptions.DnsError", 
                                   (char*)gai_strerror(error));
            kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);
            if (res) freeaddrinfo(res);
            return -1;
        }
        else 
        {
            struct addrinfo *top = res;
            while (res) {
                sa = res->ai_addr;
                salen = res->ai_addrlen;
            
                if (mode)
                {
                    /* bind() */
                    ret = bind(sock, sa, salen);
                    if (ret < 0) goto end_socket;
                }
                else
                {
                    /* connect() */
                    ret = connect(sock, sa, salen);
                    if (ret == 0) goto end_socket;
                }
            
                res = res->ai_next;
            }
end_socket:
            freeaddrinfo(top);
        }
#ifndef WIN32
    }
#endif
    return ret;
}

/*****************************************************************************
 * Bind address to socket
 ****************************************************************************/
KITE_CLASS_METHOD(Socket_bind) {
    kite_object_t *host, *port;
    int ret;

    KITE_GET_METHOD_ARGUMENT(host, 1);
    KITE_GET_METHOD_ARGUMENT(port, 2);

    if (!KITE_IS_TYPE(host, OBJ_STRING) || !KITE_IS_TYPE(port, OBJ_INTEGER)) {
        kite_object_t *exc = 
            kite_new_exception(thd, "System.exceptions.TypeMismatch", 
                               "Arguments must be of the correct type (string, integer)");
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);
        return;
    }

    ret = Socket_dnsLookup(
        thd, this, KITE_GET_INTEGER(this), KITE_GET_STRING_VALUE(host), 
        KITE_GET_INTEGER(port), 1
    );

    if (ret < 0)
        THROW_SOCKET_ERROR;
}

/*****************************************************************************
 * Connect to server
 ****************************************************************************/
KITE_CLASS_METHOD(Socket_connect) {
    kite_object_t *host, *port;
    int ret;

    KITE_GET_METHOD_ARGUMENT(host, 1);
    KITE_GET_METHOD_ARGUMENT(port, 2);

    if (!KITE_IS_TYPE(host, OBJ_STRING) || !KITE_IS_TYPE(port, OBJ_INTEGER)) {
        kite_object_t *exc = 
            kite_new_exception(thd, "System.exceptions.TypeMismatch", 
                               "Arguments must be of the correct type (string, integer)");
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);
        return;
    }

    ret = Socket_dnsLookup(
        thd, this, KITE_GET_INTEGER(this), KITE_GET_STRING_VALUE(host), 
        KITE_GET_INTEGER(port), 0
    );

    if (ret < 0)
        THROW_SOCKET_ERROR;
}

KITE_MODULE_INITIALIZER(socket) {
    /*
     * Create new class object. inherits is typically NULL for packages
     * and System.object.
     */
    kite_object_t *inherits =
        kite_dereference_and_load(thread, "System.object");
    kite_object_t *newclass = kite_new_class(thread, inherits, "socket");
    kite_set_docstring(newclass, "Internet client and server support.");

    /* Add methods common to all classes. */
    kite_add_method(thread, newclass, "__construct__",
        kite_new_method_compiled_with_docs(thread, Socket_construct, 
                                  "Constructs a new socket.", 3,
                                  "domain", "The socket's domain",
                                  "type", "The socket's type",
                                  "protocol", "The socket's protocol."));
    kite_add_method(thread, newclass, "accept", 
        kite_new_method_compiled_with_docs(thread, Socket_accept, 
                                  "Accepts a new connection from the listening socket.", 0));
    kite_add_method(thread, newclass, "bind", 
        kite_new_method_compiled_with_docs(thread, Socket_bind, 
                                  "Binds socket to given host and port.", 2,
                                  "host", "Host/filename to bind to.",
                                  "port", "Port to bind to (for Unix sockets, ignored)."));
    kite_add_method(thread, newclass, "connect", 
        kite_new_method_compiled_with_docs(thread, Socket_connect, 
                                  "Connects to given host and port.", 2,
                                  "host", "Host/filename to connect to.",
                                  "port", "Port to connect to (for Unix sockets, ignored)."));
    kite_add_method(thread, newclass, "listen", 
        kite_new_method_compiled_with_docs(thread, Socket_listen, 
                                  "Listen for connections on given socket.", 1,
                                  "queue", "The size of the listen queue."));
    kite_add_method(thread, newclass, "eof", 
        kite_new_method_compiled_with_docs(thread, Socket_eof, 
                                  "Checks if the socket has any more data to read. 1 second timeout.", 0));
    kite_add_method(thread, newclass, "eof", 
      kite_new_method_compiled_with_docs(thread, Socket_eof, 
                                "Checks if the socket has any more data to read. Timeout in parameter.", 1,
                                "timeout","How long (in milliseconds) to wait before timing out and returning TRUE."));                                                                    
    kite_add_method(thread, newclass, "read", 
        kite_new_method_compiled_with_docs(thread, Socket_read, 
                                  "Read bytes from socket and place in a string.", 1,
                                  "size", "Number of bytes to read."));
    kite_add_method(thread, newclass, "read", 
        kite_new_method_compiled_with_docs(thread, Socket_read_with_timeout, 
                                  "Read bytes from socket and place in a string. Takes a timeout argument", 2,
                                  "size", "Number of bytes to read.",
                                  "timeout", "Number of milliseconds to wait for a connection."));
    kite_add_method(thread, newclass, "readline", 
        kite_new_method_compiled_with_docs(thread, Socket_readline, 
                                  "Reads a line of text into a string.", 0));
    kite_add_method(thread, newclass, "write", 
        kite_new_method_compiled_with_docs(thread, Socket_write, 
                                  "Writes bytes to socket.", 1,
                                  "str", "Text to write."));
    kite_add_method(thread, newclass, "close", 
        kite_new_method_compiled_with_docs(thread, Socket_close, 
                                  "Closes socket and prevents further operations on it.", 0));

    /* Add standard socket constants */
    SET_SOCK_CONSTANT(SOCK_DGRAM);
    SET_SOCK_CONSTANT(SOCK_STREAM);
    SET_SOCK_CONSTANT(SOCK_SEQPACKET);
    SET_SOCK_CONSTANT(AF_INET);
    SET_SOCK_CONSTANT(AF_INET6);
    SET_SOCK_CONSTANT(AF_UNSPEC);
    SET_SOCK_CONSTANT(AF_UNIX);

    /* Set destructor. */
    kite_add_method(thread, newclass, "__destruct__", 
        kite_new_method_compiled_with_docs(thread, Socket_close, 
                                  "Class destructor.", 0));

    /* Add to parent to complete loading. */
    kite_add_property(thread, parent, "socket", 1, NULL);
    kite_set_property(thread, parent, "socket", newclass);
}
