/*****************************************************************************
 * Copyright (c) 2007, Mooneer Salem
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Kite Language organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <ctype.h>
#include <limits.h>
#include "kite_object.h"
#include "vm/kite_vm.h"
#include "config.h"

#define TYPECHECK_OPERANDS \
    kite_object_t *rhs; \
    kite_object_t *exc; \
    if (args->type == OBJ_LIST) { \
        rhs = args->builtin_data.listvalue.car; \
    } else { \
        rhs = args; \
    } \
    if (this->type != rhs->type) { \
        exc = kite_new_exception(thd, "System.exceptions.TypeMismatch", \
                            "Both operands must be of the same type."); \
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE); \
        return; \
    }

#define MAX(x,y) ((x < y) ? y : x)

#define PERFORM_STRING_BOOLEAN_OPERATION(op) \
    int result; \
    TYPECHECK_OPERANDS; \
    result = memcmp(this->builtin_data.stringvalue.string, \
                        rhs->builtin_data.stringvalue.string, \
                        MAX(KITE_GET_STRING_LENGTH(this), KITE_GET_STRING_LENGTH(rhs))) op 0; \
    kite_vm_return(thd, kite_new_boolean(thd, result));
    
#define TYPECHECK_INT_OPERAND \
    kite_object_t *rhs; \
    kite_object_t *exc; \
    if (args->type == OBJ_LIST) { \
        rhs = args->builtin_data.listvalue.car; \
    } else { \
        rhs = args; \
    } \
    if (rhs->type != OBJ_INTEGER) { \
        exc = kite_new_exception(thd, "System.exceptions.TypeMismatch", \
                            "Second operand must be an integer."); \
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE); \
        return; \
    }

#define CHECK_SINGLE_INDEX \
    index = rhs->builtin_data.intvalue; \
    if (index < 0 || index > this->builtin_data.stringvalue.length) { \
        exc = kite_new_exception(thd, "System.exceptions.ArrayBounds", \
                                 "Provided index is out of bounds."); \
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE); \
        return; \
    }
        
KITE_CLASS_METHOD(String_add) {
    kite_object_t *ret;
    char *buf;

    TYPECHECK_OPERANDS;

    buf = malloc(KITE_GET_STRING_LENGTH(this) + KITE_GET_STRING_LENGTH(rhs) + 1);
    memcpy(buf, KITE_GET_STRING_VALUE(this), KITE_GET_STRING_LENGTH(this));
    memcpy(buf + KITE_GET_STRING_LENGTH(this), KITE_GET_STRING_VALUE(rhs), KITE_GET_STRING_LENGTH(rhs));
    buf[KITE_GET_STRING_LENGTH(this) + KITE_GET_STRING_LENGTH(rhs)] = 0;
    
    ret = 
        kite_new_string_with_length(
            thd, 
            buf, 
            KITE_GET_STRING_LENGTH(this) + KITE_GET_STRING_LENGTH(rhs)
        );
    free(buf);
    kite_vm_return(thd, ret);
}

KITE_CLASS_METHOD(String_multiply) {
    kite_object_t *ret;
    int i, num;
    char *buf;

    TYPECHECK_INT_OPERAND;
    
    num = rhs->builtin_data.intvalue;
    if (num == 0) {
        kite_vm_return(thd, kite_new_string(thd, ""));
    } else if (num < 0) {
        exc = kite_new_exception(thd, "System.exceptions.InvalidArgument",
                            "Provided argument is invalid.");
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);
        return;
    }
    
    buf = malloc(KITE_GET_STRING_LENGTH(this) * num + 1);
    for(i = 0; i < num; i++)
    {
        memcpy(buf + (i * num), KITE_GET_STRING_VALUE(this), KITE_GET_STRING_LENGTH(this));
    }
    buf[KITE_GET_STRING_LENGTH(this) * num] = 0;
    ret = 
        kite_new_string_with_length(
            thd, 
            buf, 
            KITE_GET_STRING_LENGTH(this) * num
        );
    
    free(buf);
    kite_vm_return(thd, ret);
}

KITE_CLASS_METHOD(String_lessthan) {
    PERFORM_STRING_BOOLEAN_OPERATION(<);
}

KITE_CLASS_METHOD(String_lessequals) {
    PERFORM_STRING_BOOLEAN_OPERATION(<=);
}

KITE_CLASS_METHOD(String_equals) {
    PERFORM_STRING_BOOLEAN_OPERATION(==);
}

KITE_CLASS_METHOD(String_greaterequals) {
    PERFORM_STRING_BOOLEAN_OPERATION(>=);
}

KITE_CLASS_METHOD(String_greaterthan) {
    PERFORM_STRING_BOOLEAN_OPERATION(>);
}

KITE_CLASS_METHOD(String_notequals) {
    PERFORM_STRING_BOOLEAN_OPERATION(!=);
}

KITE_CLASS_METHOD(String_deref) {
    int index;
    char buf[2];

    TYPECHECK_INT_OPERAND;

    index = rhs->builtin_data.intvalue;
    if (index < 0 || 
        index >= KITE_GET_STRING_LENGTH(this)) {
        exc = kite_new_exception(thd, "System.exceptions.ArrayBounds",
                                      "Provided index is out of bounds.");
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);
        return;
    }
    
    buf[0] = this->builtin_data.stringvalue.string[index];
    buf[1] = 0;
    kite_vm_return(thd, kite_new_string_with_length(thd, buf, 1));
}

KITE_CLASS_METHOD(String_str) {
    KITE_NO_ARGS;
    kite_vm_return(thd, kite_reference_object(this));
}

KITE_CLASS_METHOD(String_int) {
    KITE_NO_ARGS;
    kite_vm_return(thd, kite_int_object(thd, this));
}

KITE_CLASS_METHOD(String_asc) {
    unsigned char *tmp = (unsigned char*)KITE_GET_STRING_VALUE(this);
    int ret = (int)*tmp;
    KITE_NO_ARGS;
    kite_vm_return(thd, kite_new_integer(thd, ret));
}

KITE_CLASS_METHOD(String_chr) {
    char buf[2];
    TYPECHECK_INT_OPERAND;
    (void)this;
    buf[0] = (unsigned char)KITE_GET_INTEGER(rhs);
    buf[1] = 0;
    kite_vm_return(thd, kite_new_string_with_length(thd, buf, 1));
}

KITE_CLASS_METHOD(String_float) {
    KITE_NO_ARGS;
    kite_vm_return(thd, kite_float_object(thd, this));
}

KITE_CLASS_METHOD(String_bool) {
    KITE_NO_ARGS;
    kite_vm_return(thd, kite_boolean_object(thd, this));
}

KITE_CLASS_METHOD(String_length) {
    KITE_NO_ARGS;
    kite_vm_return(thd, 
        kite_new_integer(thd, KITE_GET_STRING_LENGTH(this)));
}

KITE_CLASS_METHOD(String_ltrim) {
    char *buf = KITE_GET_STRING_VALUE(this);
    int l = KITE_GET_STRING_LENGTH(this);
    KITE_NO_ARGS;
    while(l > 0 && isspace(*buf++)) l--;

    kite_vm_return(thd, 
        kite_new_string_with_length(thd, l > 0 ? buf - 1 : "", l));
}

KITE_CLASS_METHOD(String_rtrim) {
    char *buf = KITE_GET_STRING_VALUE(this);
    int l = KITE_GET_STRING_LENGTH(this);
    KITE_NO_ARGS;
    while(l > 0 && (isspace(buf[l - 1]) || !buf[l - 1])) l--;

    kite_vm_return(thd, 
        kite_new_string_with_length(thd, buf, l));
}

KITE_CLASS_METHOD(String_trim) {
    kite_object_t *tmp;
    KITE_NO_ARGS;

    String_ltrim(thd, this, NULL);
    tmp = kite_vm_pop(thd);
    String_rtrim(thd, tmp, NULL);
    kite_dereference_object(tmp);
}

KITE_CLASS_METHOD(String_lower) {
    char *newstr = malloc(KITE_GET_STRING_LENGTH(this) + 1);
    int i;
    kite_object_t *ret;
    
    KITE_NO_ARGS;

    for(i = 0; i < KITE_GET_STRING_LENGTH(this); i++) 
    {
        newstr[i] = tolower(KITE_GET_STRING_VALUE(this)[i]);
    }
    newstr[i] = 0;
    
    ret = kite_new_string_with_length(thd, newstr, KITE_GET_STRING_LENGTH(this));
    free(newstr);
    
    kite_vm_push(thd, ret);
    kite_dereference_object(ret);
}

KITE_CLASS_METHOD(String_upper) {
    char *newstr = malloc(KITE_GET_STRING_LENGTH(this) + 1);
    int i;
    kite_object_t *ret;
    
    KITE_NO_ARGS;

    for(i = 0; i < KITE_GET_STRING_LENGTH(this); i++) 
    {
        newstr[i] = toupper(KITE_GET_STRING_VALUE(this)[i]);
    }
    newstr[i] = 0;
    
    ret = kite_new_string_with_length(thd, newstr, KITE_GET_STRING_LENGTH(this));
    free(newstr);
    
    kite_vm_push(thd, ret);
    kite_dereference_object(ret);
}

KITE_CLASS_METHOD(String_replace) {
    kite_object_t *match,*replace;
    kite_object_t *regexclass, *regex, *listargs;
    
    KITE_GET_METHOD_ARGUMENT(match,1);
    KITE_GET_METHOD_ARGUMENT(replace,2);
    
    regexclass = kite_dereference_and_load(thd,"System.regex");
    listargs = kite_list_object(thd,match);
    regex = kite_new_instance_with_constructor(thd,regexclass,listargs);
    kite_dereference_object(listargs);
    
    listargs = kite_new_list(thd);
    kite_append_list(thd,listargs,this);
    kite_append_list(thd,listargs,replace);
    
    kite_vm_call_method(thd,regex,"replace",listargs,1);
    
    kite_vm_return(thd,kite_vm_pop(thd));
    kite_dereference_object(regexclass);
    kite_dereference_object(regex);
    
}

KITE_CLASS_METHOD(String_substring) {
    int length, index;
    kite_object_t *ret;
    char *tmp;
    
    TYPECHECK_INT_OPERAND;
    CHECK_SINGLE_INDEX;

    if (args->builtin_data.listvalue.cdr) {
        args = args->builtin_data.listvalue.cdr;
        {
            TYPECHECK_INT_OPERAND;
            length = rhs->builtin_data.intvalue;
        }
    } else {
        length = INT_MAX;
    }
    
    if (length > KITE_GET_STRING_LENGTH(this)) {
        length = KITE_GET_STRING_LENGTH(this);
    } else if (length < 0) length = 0;
    
    tmp = (char*)calloc(length + 1, 1);
    if (length > 0)
        memcpy(tmp, KITE_GET_STRING_VALUE(this) + index, length);
    tmp[length] = 0;
    
    ret = kite_new_string(thd, tmp);
    free(tmp);
    kite_vm_return(thd, ret);
}

KITE_CLASS_METHOD(String_split) {
    kite_object_t *delimiter, *limit, *ret;
    int max = INT_MAX; 
    char *cur;

    KITE_GET_METHOD_ARGUMENT(delimiter, 1);
    KITE_GET_METHOD_ARGUMENT(limit, 2);

    if (!KITE_IS_TYPE(delimiter, OBJ_STRING) || 
        (limit && !KITE_IS_TYPE(limit, OBJ_INTEGER)))
    {
        kite_object_t *exc;
        exc = kite_new_exception(thd, "System.exceptions.TypeMismatch",
                                 "Operands must be of (string, integer)");
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);
        return;
    }

    if (limit) {
        max = KITE_GET_INTEGER(limit);
    }

    cur = this->builtin_data.stringvalue.string;
    ret = kite_new_list(thd);
    while(cur && *cur && max > 0) {
        char *tmp = strstr(cur, delimiter->builtin_data.stringvalue.string);
        char *ele;
        if (tmp && tmp != cur && max > 1) {
            ele = calloc(1, tmp - cur + 1);
            strncpy(ele, cur, tmp - cur);
            tmp += KITE_GET_STRING_LENGTH(delimiter);
        } else if (tmp == cur && max > 1) {
            ele = calloc(2, 1);
            if (KITE_GET_STRING_LENGTH(delimiter) == 0)
            {
                *ele = *cur;
                *(ele + 1) = 0;
                tmp++;
            }
            else
            {
                *ele = 0;
                tmp += KITE_GET_STRING_LENGTH(delimiter);
            }
        } else {
            ele = calloc(1, strlen(cur) + 1);
            strcpy(ele, cur);
        }
        kite_append_list(thd, ret, kite_new_string(thd, ele));
        free(ele);

        cur = tmp;
        max--;
    }
    
    kite_vm_return(thd, ret);
}

#define CHECK_TYPE(object, t) \
    if (object == NULL) { \
        exc = kite_new_exception(thd, "System.exceptions.InvalidArgument", \
            "Not enough objects have been passed in for the format " \
            "string given."); \
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE); \
        return; \
    } \
    if (object->type != t) { \
        exc = kite_new_exception(thd, "System.exceptions.TypeMismatch", \
            "Object passed to format is of invalid type for the format " \
            "specifier provided."); \
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE); \
        return; \
    }

#define FORMAT(out, type, val) \
    fmtstr = (char*)calloc(length + 1, 1); \
    strncpy(fmtstr, cur - length + 1, length); \
    asprintf(&out, fmtstr, \
        (type)cur_fmt->builtin_data.listvalue.car->builtin_data.val); \
    free(fmtstr);

#ifndef HAVE_ASPRINTF
void asprintf(char **out, char *fmt, ...)
{
    va_list ap;
    char *ret;

    va_start(ap, fmt);
    switch(*(fmt+1))
    {    
        case 's':
        {
            ret = malloc(strlen(va_arg(ap, char*) + 1));
            break;
        }
        default:
        {
            /* Really, we need to do this better to avoid buffer overflows. */
            ret = malloc(256);
            break;
        }
    }

    vsprintf(ret, fmt, ap);
    ret = realloc(ret, strlen(ret) + 1);
    va_end(ap);
    *out = ret;
}
#endif

KITE_CLASS_METHOD(String_format) {
    kite_object_t *format_list = args->builtin_data.listvalue.car;
    char *output = NULL, *ret_string = NULL, *cur, *fmtstr;
    int length = 0;
    kite_object_t *cur_fmt, *exc;
    
    if (format_list->type != OBJ_LIST) {
        format_list = kite_list_object(thd, format_list);
    }
    
    cur = this->builtin_data.stringvalue.string;
    cur_fmt = format_list;
    while(*cur && cur_fmt) {
        if (*cur == '%') {
            cur++;
            length++;
cont_parse:
            while(*cur && *cur != '%' && !isalpha(*cur)) {
                cur++;
                length++;
            }
            if (*cur == '%') goto cont_parse;
            
            length++;
            switch(*cur) {
                case 'd':
                case 'i':
                case 'o':
                case 'u':
                case 'x':
                case 'X':
                    /* Integer conversion. */
                    CHECK_TYPE(cur_fmt->builtin_data.listvalue.car, OBJ_INTEGER);
                    FORMAT(output, int, intvalue);
                    break;
                case 'D':
                case 'O':
                case 'U':
                    /* Long conversion. */
                    CHECK_TYPE(cur_fmt->builtin_data.listvalue.car, OBJ_INTEGER);
                    FORMAT(output, long, intvalue);
                    break;
                case 'e':
                case 'E':
                case 'F':
                case 'f':
                case 'G':
                case 'g':
                case 'A':
                case 'a':
                    /* Double conversion. */
                    CHECK_TYPE(cur_fmt->builtin_data.listvalue.car, OBJ_FLOAT);
                    FORMAT(output, double, floatvalue);
                    break;
                case 's':
                    /* String conversion */
                    CHECK_TYPE(cur_fmt->builtin_data.listvalue.car, OBJ_STRING);
                    FORMAT(output, char*, stringvalue.string);
                    break;
                case 'c':
                    /* Character conversion */
                    CHECK_TYPE(cur_fmt->builtin_data.listvalue.car, OBJ_INTEGER);
                    FORMAT(output, int, intvalue);
                    break;
                default:
                    /* Invalid format string. */
                    exc = kite_new_exception(thd, 
                                        "System.exceptions.InvalidArgument",
                                        "Provided format string is invalid.");
                    kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), 
                                        TRUE);
                    return;
            }
            
            length = 0;
            if (ret_string == NULL) {
                ret_string = output;
            } else {
                ret_string = realloc(ret_string, strlen(ret_string) + 
                                     strlen(output) + 1);
                strcat(ret_string, output);
                free(output);
            }
            cur++;
            cur_fmt = cur_fmt->builtin_data.listvalue.cdr;
            continue;
        }
        
        cur++;
        length++;
    }
    
    if (cur_fmt != NULL && cur_fmt->builtin_data.listvalue.car != NULL && !*cur) {
        exc = kite_new_exception(thd, 
                            "System.exceptions.InvalidArgument",
                            "Provided format string has fewer format "
                            "specifiers than required.");
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), 
                            TRUE);
        return;
    } else if (length > 0 || *cur) {
        asprintf(&output, cur - length);
        if (ret_string != NULL) {
            ret_string = 
                realloc(ret_string, strlen(ret_string) + strlen(output) + 1);
            strcat(ret_string, output);
            free(output);
        } else {
            ret_string = output;
        }
    } else if (!ret_string) {
        ret_string = strdup("");
    }
    
    cur_fmt = kite_new_string(thd, ret_string);
    free(ret_string);
    kite_vm_return(thd, cur_fmt);
}
    
KITE_MODULE_INITIALIZER(string) {
    /*
     * Create new class object. inherits is typically NULL for packages
     * and System.object.
     */
    kite_object_t *inherits = 
        kite_dereference_and_load(thread, "System.object");
    kite_object_t *newclass = 
        kite_new_class(thread, inherits, "string");
    kite_set_docstring(newclass, "Object representing a Kite string.");
    
    /*
     * Add operators
     */
    kite_add_operator(thread, newclass, OP_ADD, 
        kite_new_method_compiled_with_docs(thread, String_add, 
                                  "Concatenate two strings together.", 1,
                                  "rhs", "The string to append."));
    kite_add_operator(thread, newclass, OP_MULTIPLY, 
        kite_new_method_compiled_with_docs(thread, String_multiply, 
                                  "Duplicate string a given number of times.", 1,
                                  "num", "Number of times to duplicate."));
    kite_add_operator(thread, newclass, OP_LESS_THAN, 
        kite_new_method_compiled_with_docs(thread, String_lessthan, 
                                  "Returns true if the current string is less than the given string.", 1,
                                  "rhs", "The string to compare to."));
    kite_add_operator(thread, newclass, OP_LESS_OR_EQUALS, 
        kite_new_method_compiled_with_docs(thread, String_lessequals,
                                  "Returns true if the current string is less than or equal to the given string.", 1,
                                  "rhs", "The string to compare to."));
    kite_add_operator(thread, newclass, OP_EQUALS, 
        kite_new_method_compiled_with_docs(thread, String_equals,
                                  "Returns true if the current string is equal to the given string.", 1,
                                  "rhs", "The string to compare to."));
    kite_add_operator(thread, newclass, OP_GREATER_OR_EQUALS, 
        kite_new_method_compiled_with_docs(thread, String_greaterequals,
                                  "Returns true if the current string is greater than or equal to the given string.", 1,
                                  "rhs", "The string to compare to."));
    kite_add_operator(thread, newclass, OP_GREATER_THAN, 
        kite_new_method_compiled_with_docs(thread, String_greaterthan,
                                  "Returns true if the current string is greater than the given string.", 1,
                                  "rhs", "The string to compare to."));
    kite_add_operator(thread, newclass, OP_NOT_EQUALS, 
        kite_new_method_compiled_with_docs(thread, String_notequals, 
                                  "Returns true if the current string is not equal to the given string.", 1,
                                  "rhs", "The string to compare to."));
    kite_add_operator(thread, newclass, OP_ARRAY_DEREF, 
        kite_new_method_compiled_with_docs(thread, String_deref, 
                                  "Returns the character at the given index.", 1,
                                  "index", "The 0-based index to return."));
    
    /* Add methods. */
    kite_add_method(thread, newclass, "str", 
        kite_new_method_compiled_with_docs(thread, String_str, 
                                  "Return string representation of given string.", 0));
    kite_add_method(thread, newclass, "int", 
        kite_new_method_compiled_with_docs(thread, String_int, 
                                  "Return integer representation of given string.", 0));
    kite_add_method(thread, newclass, "float", 
        kite_new_method_compiled_with_docs(thread, String_float, 
                                  "Return floating-point representation of given string.", 0));
    kite_add_method(thread, newclass, "bool", 
        kite_new_method_compiled_with_docs(thread, String_bool, 
                                  "Return Boolean representation of given string.", 0));
    kite_add_method(thread, newclass, "asc", 
        kite_new_method_compiled_with_docs(thread, String_asc, 
                                "Return ASCII code for current character.", 0));
    kite_add_method(thread, newclass, "chr", 
        kite_new_method_compiled_with_docs(thread, String_chr, 
                                "Return ASCII character for given code.", 1,
                                "code", "Code to look up."));
    kite_add_method(thread, newclass, "length", 
        kite_new_method_compiled_with_docs(thread, String_length, 
                                  "Return length of given string.", 0));
    kite_add_method(thread, newclass, "ltrim", 
        kite_new_method_compiled_with_docs(thread, String_ltrim, 
                                  "Trim whitespace from beginning of string.", 0));
    kite_add_method(thread, newclass, "rtrim", 
        kite_new_method_compiled_with_docs(thread, String_rtrim, 
                                  "Trim whitespace from the end of string.", 0));
    kite_add_method(thread, newclass, "trim", 
        kite_new_method_compiled_with_docs(thread, String_trim, 
                                  "Trim whitespace from both sides of string.", 0));
    kite_add_method(thread, newclass, "lower", 
        kite_new_method_compiled_with_docs(thread, String_lower, 
                                  "Return lower-case representation of given string.", 0));
    kite_add_method(thread, newclass, "upper", 
        kite_new_method_compiled_with_docs(thread, String_upper, 
                                  "Return upper-case representation of given string.", 0));
    kite_add_method(thread, newclass, "append", 
        kite_new_method_compiled_with_docs(thread, String_add, 
                                  "Concatenate two strings together.", 1,
                                  "rhs", "The string to append."));
    kite_add_method(thread, newclass, "repeat", 
        kite_new_method_compiled_with_docs(thread, String_multiply,
                                  "Duplicate string a given number of times.", 1,
                                  "num", "Number of times to duplicate."));
    kite_add_method(thread, newclass, "charAt", 
        kite_new_method_compiled_with_docs(thread, String_deref,
                                  "Returns the character at the given index.", 1,
                                  "index", "The 0-based index to return."));
                                  
   kite_add_method(thread, newclass, "replace", 
       kite_new_method_compiled_with_docs(thread, String_replace, 
                                 "Replaces every instance of the first parameter with the second parameter.", 2,
                                 "match", "The substring to search for.",
                                 "replace", "The string with which to replace matched substrings."));
                                 
    kite_add_method(thread, newclass, "substring", 
        kite_new_method_compiled_with_docs(thread, String_substring, 
                                  "Return part of the given string.", 1,
                                  "start", "The 0-based position to extract from."));
    kite_add_method(thread, newclass, "substring", 
        kite_new_method_compiled_with_docs(thread, String_substring, 
                                  "Return part of the given string.", 2,
                                  "start", "The 0-based position to extract from.",
                                  "length", "The number of characters to extract."));
    kite_add_method(thread, newclass, "split", 
        kite_new_method_compiled_with_docs(thread, String_split, 
                                  "Split a string on a delimiter.", 1,
                                  "delim", "Delimiter to split on."));
    kite_add_method(thread, newclass, "split", 
        kite_new_method_compiled_with_docs(thread, String_split, 
                                  "Split a string on a delimiter.", 2,
                                  "delim", "Delimiter to split on.",
                                  "max", "Maximum number of times to split."));
    kite_add_method(thread, newclass, "format", 
        kite_new_method_compiled_with_docs(thread, String_format, 
                                  "Replace format codes within string with the given values (as in printf).", 1,
                                  "vals", "An array of values matching the types given in the format string."));
        
    /* Add to parent to complete loading. */
    kite_add_property(thread, parent, "string", 1, NULL);
    kite_set_property(thread, parent, "string", newclass);
}
