/*****************************************************************************
 * Copyright (c) 2007, Mooneer Salem
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Kite Language organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "kite_object.h"

/*****************************************************************************
 * Mark object and parents as referred (for garbage collector).
 ****************************************************************************/
kite_object_t *kite_reference_object(kite_object_t *obj) {
#ifndef HAVE_GC_H
    if (obj != NULL && obj->refcount == 0) {
        obj->refcount = 1;
        kite_list_remove(
            &obj->owner_thread->vm->gc_begin, 
            &obj->owner_thread->vm->gc_end, 
            obj->gc_entry
        );
        kite_list_add_end(
            &obj->owner_thread->vm->gc_begin, 
            &obj->owner_thread->vm->gc_end, 
            obj->gc_entry
        );
        if (obj->parent) kite_reference_object(obj);
        if (obj->object_data.inherit_from) kite_reference_object(obj);
    }
    else if (obj != NULL) obj->refcount++;
#endif /* HAVE_GC_H */    
    return obj;
}

/*****************************************************************************
 * Mark object as unreferenced (for garbage collector).
 ****************************************************************************/
void kite_dereference_object(kite_object_t *obj) {
#ifndef HAVE_GC_H
    if (obj != NULL) {
        obj->refcount--;
        if (obj->refcount == 0) {
            kite_list_remove(
                &obj->owner_thread->vm->gc_begin,
                &obj->owner_thread->vm->gc_end, 
                obj->gc_entry
            );
            kite_list_add_begin(
                &obj->owner_thread->vm->gc_begin, 
                &obj->owner_thread->vm->gc_end, 
                obj->gc_entry
            );
        }
    }
#else
    (void)obj;
#endif /* HAVE_GC_H */
}

/*****************************************************************************
 * Verify that we're really destructable
 ****************************************************************************/
#ifndef HAVE_GC_H
static int kite_gc_verify_noreferences(kite_object_t *obj) {
    int ret = TRUE;

    if (obj->refcount) return FALSE;
    if (obj->parent) ret &= kite_gc_verify_noreferences(obj->parent);
    if (!ret) kite_reference_object(obj);

    return ret;
}
#endif /* HAVE_GC_H */

/*****************************************************************************
 * Perform incremental garbage collection
 ****************************************************************************/
kite_object_t *kite_gc_incremental(kite_thread_t *thd) {
#ifndef HAVE_GC_H
    kite_list_t **gc_list = &thd->vm->gc_begin;

    if (*gc_list && ((kite_object_t*)(*gc_list)->obj)->refcount == 0) {
        kite_object_t *obj = (kite_object_t*)(*gc_list)->obj;

        // Make sure nothing references to us in hierachy
        if (kite_gc_verify_noreferences(obj)) {
            if (obj->type == OBJ_INSTANCE)
            {
                obj->refcount++;
                kite_vm_call_method(thd, obj, "__destruct__", 
                                    kite_new_list(thd), FALSE);
            }
            kite_destruct_object_nofree(thd, &obj, TRUE);
            return obj;
        }
    }
#else
    (void)thd;
#endif /* HAVE_GC_H */
    return NULL;
}

/*****************************************************************************
 * Perform complete end of program destruction
 ****************************************************************************/
void kite_gc_destroy_all(kite_thread_t *thd) {
#ifndef HAVE_GC_H
    kite_list_t **gc_list = &thd->vm->gc_begin;
    kite_list_t *cur = *gc_list;

    /* Call all destructors */
    while(cur) {
        kite_object_t *obj = cur->obj;
        if (obj->type == OBJ_INSTANCE) {
            kite_vm_call_method(thd, obj, "__destruct__", kite_new_list(thd), FALSE);
        }
        cur = cur->next;
    }

    /* Destroy all objects */
    while(*gc_list) {
        kite_object_t *obj = (*gc_list)->obj;
        kite_destruct_object(thd, &obj, FALSE);
    }
#else
    (void)thd;
#endif /* HAVE_GC_H */
}
