/*****************************************************************************
 * Copyright (c) 2007, Mooneer Salem
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Kite Language organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "kite_object.h"
#include "vm/kite_vm.h"

/*****************************************************************************
 * Produce a boolean from an object, based on these rules:
 * Boolean: return new reference to object
 * Integer/float: true if not 0
 * String: true if strlen() > 0
 * List: true if length > 0
 * Instance of object: call obj|bool method to return boolean value
 * False for all other cases.
 ****************************************************************************/
kite_object_t *kite_boolean_object(kite_thread_t *thd, kite_object_t *obj)
{
    int value = FALSE;

    if (obj) {
        switch(obj->type)
        {
            case OBJ_INTEGER:
                {
                    value = (obj->builtin_data.intvalue ? TRUE : FALSE);
                    break;
                }
            case OBJ_FLOAT:
                {
                    value = (obj->builtin_data.floatvalue ? TRUE : FALSE);
                    break;
                }
            case OBJ_STRING:
                {
                    value = (obj->builtin_data.stringvalue.length ? TRUE : FALSE);
                    break;
                }
            case OBJ_BOOLEAN:
                {
                    return kite_reference_object(obj);
                }
            case OBJ_LIST:
                {
                    value = (obj->builtin_data.listvalue.car != NULL);
                }
            case OBJ_INSTANCE:
                {
                    kite_vm_call_method(thd, obj, "bool", kite_new_list(thd), TRUE);
                    return kite_vm_pop(thd);
                }
            default:
                {
                    value = FALSE;
                }
        }
    }

    return kite_new_boolean(thd, value);
}

/*****************************************************************************
 * Produce a string from an object, based on these rules:
 * Boolean: "true" or "false"
 * Integer/float: actual number, given by sprintf()
 * String: return reference
 * Instance of object: call obj|str
 ****************************************************************************/
kite_object_t *kite_string_object(kite_thread_t *thd, kite_object_t *obj)
{
    char string[32];
    
    switch(obj->type)
    {
        case OBJ_INTEGER:
        {
            sprintf(string, "%ld", obj->builtin_data.intvalue);
            break;
        }
        case OBJ_FLOAT:
        {
            sprintf(string, "%f", obj->builtin_data.floatvalue);
            break;
        }
        case OBJ_STRING:
        {
            return kite_reference_object(obj);
        }
        case OBJ_NULL:
        {
            return kite_new_string(thd, "Null");
        }
        case OBJ_BOOLEAN:
        {
            if (obj->builtin_data.intvalue) {
                strcpy(string, "true");
            } else {
                strcpy(string, "false");
            }
            break;
        }
        case OBJ_LIST:
        case OBJ_INSTANCE:
        {
            kite_vm_call_method(thd, obj, "str", kite_new_list(thd), TRUE);
            return kite_vm_pop(thd);
        }
        default:
        {
            return kite_object_name(thd, obj);
        }
    }
    
    return kite_new_string(thd, string);
}

/*****************************************************************************
 * Produce a int from an object, based on these rules:
 * Boolean: 1 = true, 0 = false
 * Integer/float: number casted to appropriate type
 * String: atoi(string)
 * Instance of object: call obj|int
 * 0 otherwise.
 ****************************************************************************/
kite_object_t *kite_int_object(kite_thread_t *thd, kite_object_t *obj)
{
    int value;
    
    switch(obj->type)
    {
        case OBJ_INTEGER:
        {
            return kite_reference_object(obj);
        }
        case OBJ_FLOAT:
        {
            value = (int)obj->builtin_data.floatvalue;
            break;
        }
        case OBJ_STRING:
        {
            value = atoi(obj->builtin_data.stringvalue.string);
            break;
        }
        case OBJ_BOOLEAN:
        {
            if (obj->builtin_data.intvalue) {
                value = 1;
            } else {
                value = 0;
            }
            break;
        }
        case OBJ_LIST:
        case OBJ_INSTANCE:
        {
            kite_vm_call_method(thd, obj, "int", kite_new_list(thd), TRUE);
            return kite_vm_pop(thd);
        }
        default:
        {
            value = 0;
        }
    }
    
    return kite_new_integer(thd, value);
}

/*****************************************************************************
 * Produce a int from an object, based on these rules:
 * Boolean: 1.0 = true, 0.0 = false
 * Integer/float: number casted to appropriate type
 * String: atof(string)
 * Instance of object: call obj|float
 * 0.0 otherwise.
 ****************************************************************************/
kite_object_t *kite_float_object(kite_thread_t *thd, kite_object_t *obj) 
{
    double value;
    
    switch(obj->type)
    {
        case OBJ_INTEGER:
        {
            value = obj->builtin_data.intvalue;
            break;
        }
        case OBJ_FLOAT:
        {
            return kite_reference_object(obj);
        }
        case OBJ_STRING:
        {
            value = atof(obj->builtin_data.stringvalue.string);
            break;
        }
        case OBJ_BOOLEAN:
        {
            if (obj->builtin_data.intvalue) {
                value = 1.0;
            } else {
                value = 0.0;
            }
            break;
        }
        case OBJ_LIST:
        case OBJ_INSTANCE:
        {
            kite_vm_call_method(thd, obj, "float", kite_new_list(thd), TRUE);
            return kite_vm_pop(thd);
        }
        default:
        {
            value = 0.0;
        }
    }
    
    return kite_new_float(thd, value);
}

/*****************************************************************************
 * Place object inside a new list (e.g. 1 becomes [1] and [1,2,3] becomes
 * [[1,2,3]]).
 ****************************************************************************/
kite_object_t *kite_list_object(kite_thread_t *thd, kite_object_t *obj)
{
    kite_object_t *ret = kite_new_list(thd);
    ret->builtin_data.listvalue.car = kite_reference_object(obj);
    ret->builtin_data.listvalue.cdr = NULL;
    return ret;
}

/*****************************************************************************
 * Return fully-qualified object name of current instance
 ****************************************************************************/
static char *kite_object_name_helper(kite_object_t *obj)
{
    char *ret = NULL;
    int extra = 1;
    
    if (obj == NULL || obj == obj->owner_thread->vm->root_package) {
        return calloc(1,1);
    }
    
    ret = kite_object_name_helper(obj->parent);
    if (strlen(ret) > 0) {
        extra++;
    }
    
    ret = realloc(ret, strlen(ret) + strlen(obj->object_data.name) + extra);
    if (extra > 1) strcat(ret, ".");
    strcat(ret, obj->object_data.name);
    return ret;
}

/*****************************************************************************
 * Return fully-qualified object name of current instance
 ****************************************************************************/
kite_object_t *kite_object_name(kite_thread_t *thd, kite_object_t *obj)
{
    char *helper;
    kite_object_t *ret;

    assert(obj != NULL);
    if (obj->object_data.name) {
        if (!strcmp(obj->object_data.name, "__root"))
            helper = strdup("__root");
        else
            helper = kite_object_name_helper(obj);
    } else {
        helper = kite_object_name_helper(obj->object_data.inherit_from);
    }
    ret = kite_new_string(thd, helper);
    free(helper);
    return ret;
}

