/*****************************************************************************
 * Copyright (c) 2007, Mooneer Salem
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Kite Language organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/

#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#ifdef WIN32
#include <windows.h>
#else
#include <dlfcn.h>
#endif
#include "kite_object.h"
#include "kite_gc.h"

#define NULL_OBJECT "System.null"
#define INTEGER_OBJECT "System.integer"
#define FLOAT_OBJECT "System.float"
#define STRING_OBJECT "System.string"
#define LIST_OBJECT "System.list"
#define METHOD_OBJECT "System.method"
#define BOOLEAN_OBJECT "System.boolean"

/*****************************************************************************
 * Create a new null object
 ****************************************************************************/
kite_object_t *kite_new_null(kite_thread_t *thd) {
    kite_object_t *newobj = NULL, *cls;
    assert(thd != NULL);

    if (thd->vm->null_cached) 
        return kite_reference_object(thd->vm->null_cached);

    cls = kite_dereference_and_load(thd, NULL_OBJECT);
    assert(cls != NULL);
    
    newobj = kite_new_instance(thd, cls);
    assert(newobj != NULL);
    
    newobj->type = OBJ_NULL;
    newobj->shareable = 1;
    return newobj;
}

/*****************************************************************************
 * Create a new integer object
 ****************************************************************************/
kite_object_t *kite_new_integer(kite_thread_t *thd, long val) {
    kite_object_t *newobj = NULL, *cls;
    assert(thd != NULL);
    
    //cls = kite_dereference_and_load(thd, INTEGER_OBJECT);
    cls = thd->vm->int_cached;
    if (cls == NULL) {
        thd->vm->int_cached = kite_dereference_and_load(thd, INTEGER_OBJECT);
        cls = thd->vm->int_cached;
    }
    assert(cls != NULL);
    
    newobj = kite_new_instance(thd, cls);
    assert(newobj != NULL);
    
    newobj->builtin_data.intvalue = val;
    newobj->type = OBJ_INTEGER;
    newobj->shareable = 1;
    return newobj;
}

/*****************************************************************************
 * Create a new floating-point object
 ****************************************************************************/
kite_object_t *kite_new_float(kite_thread_t *thd, double val) {
    kite_object_t *newobj = NULL, *cls;
    assert(thd != NULL);
    
    cls = kite_dereference_and_load(thd, FLOAT_OBJECT);
    assert(cls != NULL);
    
    newobj = kite_new_instance(thd, cls);
    assert(newobj != NULL);
    
    newobj->builtin_data.floatvalue = val;
    newobj->type = OBJ_FLOAT;
    newobj->shareable = 1;
    return newobj;
}

/*****************************************************************************
 * Create a new string object
 ****************************************************************************/
kite_object_t *kite_new_string(kite_thread_t *thd, char *str) {
    kite_object_t *newobj = NULL, *cls;
    assert(thd != NULL);
    
    cls = kite_dereference_and_load(thd, STRING_OBJECT);
    assert(cls != NULL);
    
    newobj = kite_new_instance(thd, cls);
    assert(newobj != NULL);
    
    newobj->builtin_data.stringvalue.length = strlen(str);
    newobj->builtin_data.stringvalue.string = strdup(str);
    newobj->type = OBJ_STRING;
    newobj->shareable = 1;
    return newobj;
}

/*****************************************************************************
 * Create a new string object
 ****************************************************************************/
kite_object_t *kite_new_string_with_length(kite_thread_t *thd, char *str, int len) 
{
    kite_object_t *newobj = NULL, *cls;
    assert(thd != NULL);
    
    cls = kite_dereference_and_load(thd, STRING_OBJECT);
    assert(cls != NULL);
    
    newobj = kite_new_instance(thd, cls);
    assert(newobj != NULL);
    
    newobj->builtin_data.stringvalue.length = len;
    newobj->builtin_data.stringvalue.string = malloc(len + 1);
    memcpy(newobj->builtin_data.stringvalue.string, str, len);
    KITE_GET_STRING_VALUE(newobj)[len] = 0;
    
    newobj->type = OBJ_STRING;
    newobj->shareable = 1;
    return newobj;
}

/*****************************************************************************
 * Create a new identifier object
 ****************************************************************************/
kite_object_t *kite_new_ident(kite_thread_t *thd, char *str) {
    kite_object_t *newobj = NULL;
    assert(thd != NULL);
    
    newobj = kite_new_class(thd, NULL, "");
    assert(newobj != NULL);
    
    newobj->builtin_data.stringvalue.length = strlen(str);
    newobj->builtin_data.stringvalue.string = strdup(str);
    newobj->type = OBJ_IDENT;
    newobj->shareable = 1;
    return newobj;
}

/*****************************************************************************
 * Create a new list object
 ****************************************************************************/
kite_object_t *kite_new_list(kite_thread_t *thd) {
    kite_object_t *newobj = NULL, *cls;
    assert(thd != NULL);
    
    cls = kite_dereference_and_load(thd, LIST_OBJECT);
    assert(cls != NULL);
    
    newobj = kite_new_instance(thd, cls);
    assert(newobj != NULL);
        
    newobj->builtin_data.listvalue.car = NULL;
    newobj->builtin_data.listvalue.cdr = NULL;
    newobj->builtin_data.listvalue.iterator = NULL;
    newobj->type = OBJ_LIST;
    return newobj;
}

/*****************************************************************************
 * Create a new boolean object
 ****************************************************************************/
kite_object_t *kite_new_boolean(kite_thread_t *thd, int val) {
    kite_object_t *newobj = NULL, *cls;
    assert(thd != NULL);

    if (val == TRUE && thd->vm->true_cached) 
        return kite_reference_object(thd->vm->true_cached);
    if (val == FALSE && thd->vm->false_cached) 
        return kite_reference_object(thd->vm->false_cached);

    cls = kite_dereference_and_load(thd, BOOLEAN_OBJECT);
    assert(cls != NULL);
    
    newobj = kite_new_instance(thd, cls);
    assert(newobj != NULL);
    
    newobj->builtin_data.intvalue = val;
    newobj->type = OBJ_BOOLEAN;
    newobj->shareable = 1;
    return newobj;
}

/*****************************************************************************
 * Create a new method object
 ****************************************************************************/
kite_object_t *kite_new_method_compiled(kite_thread_t *thd,
                                 kite_compiled_func_t ptr,
                                                  int numargs) {
    kite_object_t *newobj = NULL, *cls;
    assert(thd != NULL);

    cls = kite_dereference_and_load(thd, METHOD_OBJECT);
    assert(cls != NULL);

    newobj = kite_new_instance(thd, cls);
    assert(newobj != NULL);
        
    newobj->builtin_data.funcvalue.functype = FUNC_COMPILED;
    newobj->builtin_data.funcvalue.funcptr = ptr;
    newobj->builtin_data.funcvalue.numargs = numargs;
    newobj->builtin_data.funcvalue.this = NULL;
    newobj->type = OBJ_METHOD;
    newobj->shareable = 1;
    return newobj;
}

/*****************************************************************************
 * Create a new method object (with documentation)
 ****************************************************************************/
kite_object_t *kite_new_method_compiled_with_docs(kite_thread_t *thd, 
                                           kite_compiled_func_t ptr,
                                                           char *doc,
                                                            int numargs, ...) 
{
    va_list ap;
    int i;
    
    kite_object_t *newobj = kite_new_method_compiled(thd, ptr, numargs);
    kite_object_t *docs = kite_new_list(thd);
    assert(newobj != NULL);
    
    va_start(ap, numargs);
    for (i = 0; i < numargs; i++)
    {
        char *param_name = va_arg(ap, char *);
        char *param_desc = va_arg(ap, char *);
        kite_object_t *ident = kite_new_ident(thd, param_name);
        kite_set_docstring(ident, param_desc);
        kite_append_list(thd, docs, ident);
    }
    
    va_end(ap);
    kite_set_arginfo(newobj, docs);
    kite_set_docstring(newobj, doc);
    return newobj;
}

/*****************************************************************************
 * Create a new method object
 ****************************************************************************/
kite_object_t *kite_new_method_bytecode(kite_thread_t *thd,
                                        kite_opcode_t *ptr,
                                                  int numargs) {
    kite_object_t *newobj = NULL, *cls;
    assert(thd != NULL);

    cls = kite_dereference_and_load(thd, METHOD_OBJECT);
    assert(cls != NULL);

    newobj = kite_new_instance(thd, cls);
    assert(newobj != NULL);
        
    newobj->builtin_data.funcvalue.functype = FUNC_BYTECODE;
    newobj->builtin_data.funcvalue.funcptr = ptr;
    newobj->builtin_data.funcvalue.numargs = numargs;
    newobj->builtin_data.funcvalue.this = NULL;
    newobj->type = OBJ_METHOD;
    newobj->shareable = 1;
    return newobj;
}

/*****************************************************************************
 * Create a new method object (with documentation)
 ****************************************************************************/
kite_object_t *kite_new_method_bytecode_with_docs(kite_thread_t *thd, 
                                                  kite_opcode_t *ptr,
                                                           char *doc,
                                                            int numargs, ...) 
{
    va_list ap;
    int i;
    
    kite_object_t *newobj = kite_new_method_bytecode(thd, ptr, numargs);
    kite_object_t *docs = kite_new_list(thd);
    assert(newobj != NULL);
    
    va_start(ap, numargs);
    for (i = 0; i < numargs; i++)
    {
        char *param_name = va_arg(ap, char *);
        char *param_desc = va_arg(ap, char *);
        kite_object_t *ident = kite_new_ident(thd, param_name);
        kite_set_docstring(ident, param_desc);
        kite_append_list(thd, docs, ident);
    }
    
    va_end(ap);
    kite_set_arginfo(newobj, docs);
    kite_set_docstring(newobj, doc);
    return newobj;
}

/*****************************************************************************
 * Create a new class object
 ****************************************************************************/
kite_object_t *kite_new_class(kite_thread_t *thd, kite_object_t *inherits,
                              char *name) {
    kite_object_t *newobj = NULL;
    assert(thd != NULL);

    newobj = kite_gc_incremental(thd);
    if (!newobj) {
#ifndef HAVE_GC_H
        newobj = (kite_object_t*)malloc(sizeof(kite_object_t));
        assert(newobj != NULL);
        memset(newobj, 0, sizeof(kite_object_t));
#else
        newobj = (kite_object_t*)GC_malloc(sizeof(kite_object_t));
        GC_register_finalizer_ignore_self(newobj, kite_finalize_object, NULL, NULL, NULL);
        /*GC_register_disappearing_link((void*)&newobj->parent);
        GC_register_disappearing_link((void*)&newobj->object_data.inherit_from);
        GC_register_disappearing_link((void*)&newobj->owner_thread);*/
#endif /* HAVE_GC_H */
    }
    
    newobj->object_data.name = strdup(name);
    newobj->object_data.inherit_from = kite_reference_object(inherits);
#ifndef WIN32
#ifdef HAVE_GC_H
    newobj->object_data.mtx = GC_malloc(sizeof(pthread_mutex_t));
#else
    newobj->object_data.mtx = malloc(sizeof(pthread_mutex_t));
#endif /* HAVE_GC_H */
    pthread_mutex_init(newobj->object_data.mtx, NULL);
#else
    newobj->object_data.mtx = CreateMutex(NULL, FALSE, NULL);
#endif /* WIN32 */
    newobj->owner_thread = thd;
    newobj->type = OBJ_CLASS;
    newobj->builtin_data.funcvalue.funcptr = NULL;
    newobj->shareable = 1;
#ifndef HAVE_GC_H
    newobj->refcount = 1;
    if (!newobj->gc_entry) {
        newobj->gc_entry = malloc(sizeof(kite_list_t));
        newobj->gc_entry->obj = newobj;
        newobj->gc_entry->prev = newobj->gc_entry->next = NULL;
    }
    kite_list_add_end(&thd->vm->gc_begin, &thd->vm->gc_end, newobj->gc_entry);
#endif /* HAVE_GC_H */
    return newobj;
}

/*****************************************************************************
 * Create a new instance of a class
 ****************************************************************************/
kite_object_t *kite_new_instance(kite_thread_t *thd, kite_object_t *class) {
    kite_object_t *newobj = NULL;
    
    newobj = kite_gc_incremental(thd);
    if (!newobj) {
#ifndef HAVE_GC_H
        newobj = (kite_object_t*)malloc(sizeof(kite_object_t));
        assert(newobj != NULL);
        memset(newobj, 0, sizeof(kite_object_t));
#else
        newobj = (kite_object_t*)GC_malloc(sizeof(kite_object_t));
        GC_register_finalizer_ignore_self(newobj, kite_finalize_object, NULL, NULL, NULL);
        /*GC_register_disappearing_link((void*)&newobj->parent);
        GC_register_disappearing_link((void*)&newobj->object_data.inherit_from);
        GC_register_disappearing_link((void*)&newobj->owner_thread);*/
#endif /* HAVE_GC_H */
    }

    newobj->object_data.name = NULL;
    newobj->object_data.inherit_from = kite_reference_object(class);
    newobj->owner_thread = thd;
    newobj->type = OBJ_INSTANCE;
    newobj->shareable = 0;
#ifndef HAVE_GC_H
    newobj->refcount = 1;
    if (!newobj->gc_entry) {
        newobj->gc_entry = malloc(sizeof(kite_list_t));
        newobj->gc_entry->obj = newobj;
        newobj->gc_entry->prev = newobj->gc_entry->next = NULL;
    }
    kite_list_add_end(&thd->vm->gc_begin, &thd->vm->gc_end, newobj->gc_entry);
#endif /* HAVE_GC_H */    
    return newobj;
}

/*****************************************************************************
 * Create a new instance of a class (and call appropriate constructor)
 ****************************************************************************/
kite_object_t *
kite_new_instance_with_constructor(kite_thread_t *thd, kite_object_t *class,
                                   kite_object_t *args) {
    kite_object_t *inst = kite_new_instance(thd, class);
    
    /* Call constructor, if one is defined. */
    kite_vm_call_constructor(thd, inst, args);
    
    return inst;
}

/*****************************************************************************
 * Create new exception object
 ****************************************************************************/
kite_object_t *kite_new_exception(kite_thread_t *thd, char *name, char *msg) {
    kite_object_t *cls = kite_dereference_and_load(thd, name);
    kite_object_t *mesg = kite_new_string(thd, msg);
    kite_object_t *args = kite_list_object(thd, mesg);
    kite_object_t *ret = kite_new_instance_with_constructor(thd, cls, args);
    
    kite_dereference_object(args);
    kite_dereference_object(mesg);
    return ret;
}

/*****************************************************************************
 * Destroy given object
 ****************************************************************************/
void
kite_destruct_object_nofree(kite_thread_t *thd, kite_object_t **obj, int r)
{
    kite_object_t *us = *obj;
    
    /* Ignore if null. */
    if (us == NULL) {
        *obj = NULL;
        return;
    }
#ifndef HAVE_GC_H
    assert(us->gc_entry != NULL);
#endif /* HAVE_GC_H */

    if (us->parent && r) {
        kite_dereference_object(us->parent);
        us->parent = NULL;
    }

    /* Free list objects and string values */
    if (us->type == OBJ_LIST && r) {
        kite_dereference_object(us->builtin_data.listvalue.car);
        kite_dereference_object(us->builtin_data.listvalue.cdr);
        kite_dereference_object(us->builtin_data.listvalue.iterator);
    } else if (us->type == OBJ_STRING || us->type == OBJ_IDENT) {
        free(us->builtin_data.stringvalue.string);
    }

    /* Free everything else */
    if (us->object_data.name) {
        free(us->object_data.name);
    }
    if (us->object_data.docstring) {
        free(us->object_data.docstring);
    }
    if (us->object_data.properties)
        kite_destruct_symtab(thd, &us->object_data.properties, r);
    if (us->type == OBJ_CLASS || us->type == OBJ_METHOD ||
        us->type == OBJ_INSTANCE) {
        if (us->builtin_data.funcvalue.functype == FUNC_BYTECODE &&
            us->builtin_data.funcvalue.this == NULL) {
            kite_free_instruction_list(thd, 
                    (kite_opcode_t*)us->builtin_data.funcvalue.funcptr);
            us->builtin_data.funcvalue.funcptr = NULL;
        }
    }
    if (r) kite_dereference_object(us->object_data.inherit_from);

    if (us->object_data.dylib_hndl)
    {
#ifdef WIN32
        FreeLibrary(us->object_data.dylib_hndl);
#else
        dlclose(us->object_data.dylib_hndl);
#endif
    }

#ifdef WIN32
    CloseHandle(us->object_data.mtx);
#else
    if (us->object_data.mtx)
    {
        pthread_mutex_destroy(us->object_data.mtx);
#ifdef HAVE_GC_H
        GC_free(us->object_data.mtx);
#else
        free(us->object_data.mtx);
#endif /* HAVE_GC_H */
    }
#endif /* WIN32 */

#ifndef HAVE_GC_H
    kite_list_remove(&thd->vm->gc_begin, &thd->vm->gc_end, (*obj)->gc_entry);
#endif /* HAVE_GC_H */
}

/*****************************************************************************
 * Finalize object.
 ****************************************************************************/
void kite_finalize_object(void *obj, void *cd)
{
    kite_object_t *tobj = (kite_object_t*)obj;
    (void)cd;
    kite_destruct_object(tobj->owner_thread, &tobj, 1);
}

/*****************************************************************************
 * Destroy given object
 ****************************************************************************/
void kite_destruct_object(kite_thread_t *thd, kite_object_t **obj, int r)
{
    kite_destruct_object_nofree(thd, obj, r);
#ifndef HAVE_GC_H
    free((*obj)->gc_entry);
    (*obj)->gc_entry = NULL;

    free(*obj);
    *obj = NULL;
#endif /* HAVE_GC_H */
}

