/*****************************************************************************
 * Copyright (c) 2007, Mooneer Salem
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Kite Language organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <sys/stat.h>
#ifdef WIN32
#include <windows.h>
#else
#include <dlfcn.h>
#endif
#include "kite_object.h"
#include "vm/kite_vm.h"

#define USER_MODULE_EXTENSION ".kt"

/* Dynamically edited by user apps through System.vm.loader.searchPath. */
char *kite_dylib_path = LIBRARY_SEARCH_PATH;

/* Prototype for module loader function */
typedef void (*kite_load_module)(kite_thread_t *, kite_object_t *);

/*****************************************************************************
 * Register dynamic library path.
 ****************************************************************************/
void kite_set_dylib_path(char *path)
{
    kite_dylib_path = path;
}

/*****************************************************************************
 * Register internal modules for loading.
 ****************************************************************************/
void 
kite_loader_register(kite_thread_t *thd, char *name, kite_load_module ptr) {
#ifndef HAVE_GC_H
    kite_symtab_t *symt = (struct kite_symtab_t*)malloc(sizeof(kite_symtab_t));
#else
    kite_symtab_t *symt = (struct kite_symtab_t*)GC_malloc(sizeof(kite_symtab_t));
#endif /* HAVE_GC_H */
    symt->global = 0;
    symt->left = symt->right = NULL;
    symt->value = (kite_object_t*)ptr;
    symt->name = strdup(name);
    symt->docstring = NULL;

    kite_symtab_insert(thd, &thd->vm->loader_refs, symt);
}

#ifdef WIN32
#define strtok_r strtok_s
#endif /* WIN32 */

/*****************************************************************************
 * OS-specific class loading (if binary), and compile initialization, if
 * user-developed module.
 ****************************************************************************/
static void 
kite_load_dylib(kite_thread_t *thd, kite_object_t *obj, char *dylib,
                char *name) {
    char *path_copy = NULL, *brkb = NULL, *prefix, *dylib_path;
    kite_object_t *package;
#ifdef WIN32
    HMODULE handle = NULL;
#else
    void *handle = NULL;
#endif
    kite_load_module load_method;
    kite_object_t *newobj;
    kite_symtab_t *found;
    struct stat st;

    KITE_FIND_IN_SYMTAB(found, thd->vm->loader_refs, name, 0);
    if (found) {
        // Internal module
        load_method = (kite_load_module)found->value;
        load_method(thd, obj);
        KITE_FIND_PROPERTY(found, obj, name, 1);
        newobj = found->value;
        if (newobj != NULL) {
            newobj->object_data.dylib_hndl = handle;
            return;
        }
    }

    assert(obj != NULL && name != NULL);
    path_copy = strdup(kite_dylib_path);
    do {
        for(prefix = strtok_r(path_copy, ":", &brkb);
            prefix;
            prefix = strtok_r(NULL, ":", &brkb) ) {
            int size = strlen(prefix) + strlen(dylib) + strlen(DYLIB_EXTENSION) + 2;
            if (size % 8) {
                /*
                 * Align on an 8-byte boundary to make Valgrind happy (yeah,
                 * yeah)
                 */
                size += 8 - (size % 8); 
            }
            dylib_path = malloc(size);
            assert(dylib_path != NULL);
        
            /* prefix + name + extension */
            strcpy(dylib_path, prefix);
            strcat(dylib_path, "/");
            strcat(dylib_path, dylib);
            strcat(dylib_path, DYLIB_EXTENSION);
        
            /* Load dynamic library */
            if (!handle) {
#ifdef WIN32
                handle = LoadLibrary(TEXT(dylib_path));
#else
                handle = dlopen(dylib_path, RTLD_LAZY | RTLD_GLOBAL);
#endif
            }
            free(dylib_path);
            if (handle) break;
        }
    
        /* 
         * If handle is valid, we found a compiled library to use. Find and
         * call library's load function to initiate library loading.
         */
        if (handle) {
            char *tmp_load_name = malloc(strlen(name) + strlen("_load_module") + 1);
            strcpy(tmp_load_name, name);
            strcat(tmp_load_name, "_load_module");
    #ifdef WIN32
            load_method = (kite_load_module)GetProcAddress(handle, tmp_load_name);
    #else
            load_method = dlsym(handle, tmp_load_name);
    #endif
            free(tmp_load_name);
            if (load_method) {
                load_method(thd, obj);
                KITE_FIND_PROPERTY(found, obj, name, 1);
                newobj = found->value;
                if (newobj == NULL) {
                    goto invalid_module;
                }
                newobj->object_data.dylib_hndl = handle;
                free(path_copy);
                return;
            } else {
                /* Module is invalid, cancel loading and try other modules */
    invalid_module:
    #ifdef WIN32
                FreeLibrary(handle);
    #else
                dlclose(handle);
    #endif
                continue;
            }
        }
    } while(0);

    /* Now try again for modules that are written in Kite. */
    free(path_copy);
    path_copy = strdup(kite_dylib_path);
    for(prefix = strtok_r(path_copy, ":", &brkb);
        prefix;
        prefix = strtok_r(NULL, ":", &brkb) ) {
        int size = strlen(prefix) + strlen(dylib) + 
            strlen(USER_MODULE_EXTENSION) + 2;
        if (size % 8) {
            /* Align again to make Valgrind happy */
            size += 8 - (size % 8);
        }
        dylib_path = malloc(size);
        assert(dylib_path != NULL);
        
        /* prefix + name + extension */
        strcpy(dylib_path, prefix);
        strcat(dylib_path, "/");
        strcat(dylib_path, dylib);
        strcat(dylib_path, USER_MODULE_EXTENSION);
        
        /* Load module */
        kite_add_property(thd, obj, name, 1, NULL);
        package =
            kite_new_class(
                thd, 
                kite_dereference_and_load(thd, "System.object"),
                ""
            );
        kite_set_property(thd, obj, name, package);
        kite_vm_compile_from_file_without_create(thd, dylib_path, package);

        if (!thd->exception && !stat(dylib_path, &st) && package->type == OBJ_CLASS) {
            /* Run initialization code, as required. */
            if (package->builtin_data.funcvalue.funcptr != NULL) {
                kite_vm_call_object(thd, package, package, kite_new_list(thd));
            }
            
            free(path_copy);
            free(package->object_data.name);
            package->object_data.name = strdup(name);
            free(dylib_path);
            return;
        } else if (thd->exception) {
            free(dylib_path);
            free(path_copy);
            return;
        } else {
            kite_remove_property(thd, obj, name, 1);
            kite_dereference_object(package);
            free(dylib_path);
        }
    }
    free(path_copy);
}

/*****************************************************************************
 * Locate class in class database. If it's not already there, find the .kt
 * or .so or other dynamic library file and load/parse into memory, then
 * do the search again. Finally return NULL if none of the stuff above finds
 * what we're looking for.
 ****************************************************************************/
kite_object_t *kite_dereference_and_load(kite_thread_t *thd, char *class) {
    char *class_duplicate, *brkd = NULL, *word, *classname = NULL;
    kite_object_t *loc, *tmp;
    kite_symtab_t *found;
    
    /* First, make sure everything is there */
    assert(thd != NULL && class != NULL);
    class_duplicate = strdup(class);
    assert(class_duplicate != NULL);
    
    /* 
     * Start from the top, and go through each part of the hierachy, till we
     * find what we're looking for. If the current part of the hierachy is 
     * empty, we load the class using either the standard dlsym functions
     * or we initiate a compile of a pure Kite file.
     */
    loc = thd->vm->root_package;
    for (word = strtok_r(class_duplicate, ".", &brkd); 
         word && loc;
         word = strtok_r(NULL, ".", &brkd)) {
        
        if (!classname) {
            classname = strdup(word);
        } else {
            classname = 
                realloc(classname, strlen(classname) + strlen(word) + 2);
            strcat(classname, "/");
            strcat(classname, word);
        }
        
        KITE_FIND_PROPERTY(found, loc, word, 1);
        if (found == NULL) {
            /* Could not find. Begin loading process. */
            kite_load_dylib(thd, loc, classname, word);
            if (thd->exception) 
            {
                free(class_duplicate);
                free(classname);
                return NULL;
            }
            KITE_FIND_PROPERTY(found, loc, word, 1);
        }
        
        /* Found this component, set loc for next iteration. */
        if (found) {
            loc = found->value;
        }
        else if (strlen(classname) >= strlen(class)) loc = NULL;
        else {
            tmp = kite_new_class(thd, NULL, word);
            kite_add_property(thd, loc, word, 1, NULL);
            kite_set_property(thd, loc, word, tmp);
            loc = tmp;
        }
    }
    
    free(class_duplicate);
    free(classname);
    if (word == NULL && loc == NULL) {
        kite_object_t *exc;

        if (strncmp(class, "System.exceptions", strlen("System.exceptions")) == 0)
        {
            printf("Fatal Kite error: could not find System.exceptions. This probably means\n");
            printf("that your Kite installation is corrupt. Please reinstall Kite from sources\n");
            printf("or your operating system's packages. If you continue to receive this error,\n");
            printf("please contact the Kite developers.\n\n");
            printf("Class attempted: %s\n", class);
            printf("Import path: %s\n", kite_dylib_path);
            assert(0);
        }
        
        /* Couldn't find module; throw exception. */
        exc = kite_new_exception(thd, "System.exceptions.PackageNotFound", 
                                 class);
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);
    }
    
    return loc;
}
