/*****************************************************************************
 * Copyright (c) 2007, Mooneer Salem
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Kite Language organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "kite_object.h"

/*****************************************************************************
 * Add property to object
 ****************************************************************************/
void kite_add_property(kite_thread_t *thd, kite_object_t *obj, 
                       char *name, int global, char *doc) {
    kite_symtab_t *propinfo;

    assert(obj != NULL);
    if (kite_exists_property(obj, name)) return;

#ifndef HAVE_GC_H
    propinfo = (kite_symtab_t*)malloc(sizeof(kite_symtab_t));
#else
    propinfo = (kite_symtab_t*)GC_malloc(sizeof(kite_symtab_t));
#endif /* HAVE_GC_H */
    assert(propinfo != NULL);
#ifndef HAVE_GC_H
    propinfo->name = strdup(name);
#else
    propinfo->name = GC_malloc_atomic(strlen(name) + 1);
    strcpy(propinfo->name, name);
#endif /* HAVE_GC_H */
    propinfo->global = global;
    if (doc)
    {
#ifndef HAVE_GC_H
        propinfo->docstring = strdup(doc);
#else
        propinfo->docstring = GC_malloc_atomic(strlen(doc) + 1);
        strcpy(propinfo->docstring, doc);
#endif /* HAVE_GC_H */
    }
    else
    {
        propinfo->docstring = NULL;
    }
    propinfo->left = NULL;
    propinfo->right = NULL;
    propinfo->value = NULL;
    
    kite_symtab_insert(thd, &obj->object_data.properties, propinfo);
}

/*****************************************************************************
 * Add method to object
 ****************************************************************************/
void kite_add_method(kite_thread_t *thd, kite_object_t *obj, char *name, 
                     kite_object_t *method) {
    kite_symtab_t *propinfo;

    assert(obj != NULL);
    //if (kite_exists_property(obj, name)) return;

#ifndef HAVE_GC_H
    propinfo = (kite_symtab_t*)malloc(sizeof(kite_symtab_t));
#else
    propinfo = (kite_symtab_t*)GC_malloc(sizeof(kite_symtab_t));
#endif /* HAVE_GC_H */
    assert(propinfo != NULL);
#ifndef HAVE_GC_H
    propinfo->name = strdup(name);
#else
    propinfo->name = GC_malloc_atomic(strlen(name) + 1);
    strcpy(propinfo->name, name);
#endif /* HAVE_GC_H */
    propinfo->docstring = NULL;
    propinfo->global = method->builtin_data.funcvalue.numargs;
    propinfo->left = NULL;
    propinfo->right = NULL;
    propinfo->docstring = NULL;
    propinfo->value = kite_reference_object(method);

    kite_symtab_insert(thd, &obj->object_data.properties, propinfo);
}

/*****************************************************************************
 * Add operator to object
 ****************************************************************************/
void kite_add_operator(kite_thread_t *thd, kite_object_t *obj, 
                       int op, kite_object_t *method) 
{
    kite_remove_operator(thd, obj, op);
    kite_add_method(thd, obj, OP_TO_METHOD(op), method);
}

/*****************************************************************************
 * Remove property from object
 ****************************************************************************/
void 
kite_remove_property(kite_thread_t *thd, kite_object_t *obj, char *name, int n) {
    assert(obj != NULL);
    kite_symtab_remove(thd, &obj->object_data.properties, name, n);
}

/*****************************************************************************
 * Remove method from object
 ****************************************************************************/
void kite_remove_method(kite_thread_t *thd, kite_object_t *obj, char *name, int n) {
    assert(obj != NULL);
    kite_symtab_remove(thd, &obj->object_data.properties, name, n);
}

/*****************************************************************************
 * Remove operator from object
 ****************************************************************************/
void kite_remove_operator(kite_thread_t *thd, kite_object_t *obj, int op) {
    kite_symtab_t *st;

    assert(obj != NULL);
    (void)thd;
    
    KITE_FIND_ANY_IN_SYMTAB(st, obj->object_data.properties, OP_TO_METHOD(op));
    
    if (st)
        kite_remove_method(thd, obj, OP_TO_METHOD(op), st->global);
}

/*****************************************************************************
 * Set property to given value
 ****************************************************************************/
void kite_set_property(kite_thread_t *thd, kite_object_t *obj, char *name, 
                       kite_object_t *val) {
    kite_symtab_t *entry;
    kite_object_t *curr = obj;
    int global = 0;
    assert(obj != NULL);
    
    do {
        KITE_FIND_PROPERTY(entry, curr, name, global);
        if (entry) {
            if (!entry->global && obj != curr) {
                break;
            }
            
            if (entry->value) kite_dereference_object(entry->value);
            else val->parent = kite_reference_object(obj);
            entry->value = kite_reference_object(val);
            return;
        }
        
        if (global) {
            curr = curr->object_data.inherit_from;
            global = 0;
        } else {
            global = 1;
        }
    } while(curr != NULL);
    
    kite_add_property(thd, obj, name, 0, NULL);
    kite_set_property(thd, obj, name, val);
}

/*****************************************************************************
 * Determine if object defines given property
 ****************************************************************************/
int kite_exists_property(kite_object_t *obj, char *name) {
    kite_symtab_t *entry;
    assert(obj != NULL);
    
    KITE_FIND_PROPERTY(entry, obj, name, 0);
    return (entry != NULL);
}

/*****************************************************************************
 * Determine if object defines given method
 ****************************************************************************/
int kite_exists_method(kite_object_t *obj, char *name) {
    kite_symtab_t *entry;
    assert(obj != NULL);
    
    KITE_FIND_METHOD(entry, obj, name, 0);
    return (entry != NULL && entry->value->type == OBJ_METHOD);
}

/*****************************************************************************
 * Determine if object defines given operator
 ****************************************************************************/
int kite_exists_operator(kite_object_t *obj, int op) {
    kite_symtab_t *st;
    
    assert(obj != NULL);
    KITE_FIND_ANY_IN_SYMTAB(st, obj->object_data.properties, OP_TO_METHOD(op));
    
    return (st != NULL);
}

/*****************************************************************************
 * Get property value
 ****************************************************************************/
kite_object_t *
kite_get_property(kite_object_t *obj, char *prop)
{
    kite_symtab_t *symt;
    KITE_FIND_PROPERTY(symt, obj, prop, 0);
    return kite_reference_object(symt->value);
}

/*****************************************************************************
 * Set docstring.
 ****************************************************************************/
void kite_set_docstring(kite_object_t *obj, char *doc)
{
    if (obj->object_data.docstring) {
        free(obj->object_data.docstring);
    }
    
    if (doc) {
        obj->object_data.docstring = strdup(doc);
    } else {
        obj->object_data.docstring = NULL;
    }
}

/*****************************************************************************
 * Get docstring.
 ****************************************************************************/
char *kite_get_docstring(kite_object_t *obj)
{
    if (obj->object_data.docstring) {
        return obj->object_data.docstring;
    }
    return "";
}

/*****************************************************************************
 * Set argument info list
 ****************************************************************************/
void kite_set_arginfo(kite_object_t *obj, kite_object_t *info)
{
    obj->builtin_data.funcvalue.arginfo = kite_reference_object(info);
}
