/*****************************************************************************
 * Copyright (c) 2008, Mooneer Salem
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Kite Language organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>
#include "microregex_internal.h"

#include "config.h"
#ifdef HAVE_GC_H
#include <gc/gc.h>
#define malloc(x) (GC_malloc(x))
#define calloc(x, y) (GC_malloc(x * y))
#define realloc(x, y) (GC_realloc(x, y))
#define strdup(x) (GC_strdup(x))
#define free(x) GC_free(x)
#endif /* HAVE_GC_H */

typedef struct tsmstList tMstList;

struct tsmstList {
  microregex_state_t *state;
  tMstList *prev;
};


static microregex_state_t
microregex_state_new(microregex_state_t *lst)
{
    microregex_state_t ret = malloc(sizeof(struct microregex_state_t));
    ret->state = NULL;
    ret->se_stack = malloc(sizeof(int));
    ret->se_stack[0] = 0;
    ret->se_size = 1;
    ret->ses = malloc(sizeof(char*));
    ret->ses[0] = malloc(1);
    *ret->ses[0] = 0;
    ret->num_ses = 1;
    ret->loop_count = 0;
    ret->prev = ret->next = NULL;
    ret->match_begin = 0;
    ret->match_end = 0;
    ret->insensitive = 0;
    if (*lst) {
      ret->prevL = (*lst);
      ret->nextL = (*lst)->nextL;
      if (ret->nextL) ret->nextL->prevL = ret;
      (*lst)->nextL = ret;
    } else {
      ret->prevL = ret->nextL = NULL;
      *lst = ret;
    }
    return ret;
}

static microregex_state_t
microregex_state_clone_ex(microregex_state_t st, int addToList)
{
    int i;
    microregex_state_t ret = malloc(sizeof(struct microregex_state_t));
    ret->state = st->state;
    ret->se_stack = malloc(sizeof(int) * st->se_size);
    ret->se_size = st->se_size;
    memcpy(ret->se_stack, st->se_stack, sizeof(int) * ret->se_size);
    
    ret->ses = malloc(sizeof(char*) * st->num_ses);
    for(i = 0; i < st->num_ses; i++)
    {
        ret->ses[i] = strdup(st->ses[i]);
    }
    ret->num_ses = st->num_ses;
    ret->prev = ret->next = NULL;
    ret->match_begin = st->match_begin;
    ret->match_end = st->match_end;
    ret->loop_count = st->loop_count;
    ret->insensitive = st->insensitive;
    if (addToList) {
      ret->prevL = st;
      ret->nextL = st->nextL;
      if (ret->nextL) ret->nextL->prevL = ret;
      st->nextL = ret;
    } else ret->prevL = ret->nextL = NULL;
    return ret;
}

static microregex_state_t
microregex_state_clone(microregex_state_t st)
{
    return microregex_state_clone_ex(st, 1);
}

void
microregex_state_destroy_ex(microregex_state_t *st, int removeFromList)
{
    int i;
    if (!st || !*st) return;
    if (removeFromList) {
      if ((*st)->nextL) (*st)->nextL->prevL = (*st)->prevL;
      if ((*st)->prevL) (*st)->prevL->nextL = (*st)->nextL;
    }
    free((*st)->se_stack);
    for (i = 0; i < (*st)->num_ses; i++)
    {
        free((*st)->ses[i]);
    }
    free((*st)->ses);
    free(*st);
    *st = NULL;
}


static void microregex_state_destroy_list (microregex_state_t st) {
  if (!st) return;
  while (st->prevL) st = st->prevL;
  while (st) {
    microregex_state_t p = st->nextL;
    microregex_state_destroy_ex(&st, 0);
    st = p;
  }
}

void microregex_state_destroy (microregex_state_t *st) {
  if (!st || !*st) return;
  microregex_state_destroy_list(*st);
  *st = NULL;
}



/******************************************************************************
 * State machine.
 *****************************************************************************/
static void
microregex_regex_push_submatch(microregex_state_t *st_lst, microregex_state_t cur, 
                               const char *str, const char *c);
static void
microregex_regex_pop_submatch(microregex_state_t *st_lst, microregex_state_t cur, 
                              const char *str, const char *c);
static void
microregex_regex_character_class(microregex_state_t *st_lst, microregex_state_t cur, 
                                 const char *str, const char *c);
static void
microregex_regex_alteration(microregex_state_t *st_lst, microregex_state_t cur, 
                            const char *str, const char *c);
static void
microregex_regex_assert_beginning(microregex_state_t *st_lst, microregex_state_t cur, 
                                  const char *str, const char *c);
static void
microregex_regex_assert_end(microregex_state_t *st_lst, microregex_state_t cur, 
                            const char *str, const char *c);
static void
microregex_regex_counted_loop(microregex_state_t *st_lst, microregex_state_t cur, 
                              const char *str, const char *c);
static void
microregex_regex_counted_end(microregex_state_t *st_lst, microregex_state_t cur, 
                             const char *str, const char *c);
static void
microregex_regex_counted_begin(microregex_state_t *st_lst, microregex_state_t cur, 
                               const char *str, const char *c);
static void
microregex_regex_pass(microregex_state_t *st_lst, microregex_state_t cur, 
                      const char *str, const char *c);
static void
microregex_regex_word_boundary(microregex_state_t *st_lst, microregex_state_t cur, 
                      const char *str, const char *c);
static void
microregex_regex_backtrack_check(microregex_state_t *st_lst, microregex_state_t cur, 
                      const char *str, const char *c);
typedef void 
(*state_machine_func_t)(microregex_state_t *, microregex_state_t, const char *, const char *);
static state_machine_func_t state_functions[] = {
    microregex_regex_push_submatch,
    microregex_regex_pop_submatch,
    microregex_regex_character_class,
    microregex_regex_alteration,
    microregex_regex_assert_beginning,
    microregex_regex_assert_end,
    microregex_regex_counted_begin,
    microregex_regex_counted_loop,
    microregex_regex_counted_end,
    microregex_regex_backtrack_check,
    microregex_regex_pass,
    microregex_regex_word_boundary,
    NULL,
};

static void
microregex_nfa_addstateback(microregex_state_t *st, microregex_state_t toadd)
{
    if (!*st)
    {
        *st = toadd;
        toadd->prev = toadd->next = toadd;
        return;
    }
    
    toadd->next = *st;
    toadd->prev = (*st)->prev;
    (*st)->prev->next = toadd;
    (*st)->prev = toadd;
}

static void
microregex_nfa_addstateforward(microregex_state_t *st, microregex_state_t toadd)
{
    if (!*st)
    {
        *st = toadd;
        toadd->prev = toadd->next = toadd;
        return;
    }
    
    toadd->next = (*st)->next;
    toadd->prev = (*st);
    (*st)->next->prev = toadd;
    (*st)->next = toadd;
}

static void
microregex_nfa_remove(microregex_state_t *st)
{
    if (!*st)
    {
        return;
    }
    else if ((*st)->next == (*st))
    {
        *st = NULL;
        return;
    }
    
    (*st)->next->prev = (*st)->prev;
    (*st)->prev->next = (*st)->next;
    (*st) = (*st)->next;
}

static void
microregex_regex_push_submatch(microregex_state_t *st_lst, microregex_state_t cur, 
                               const char *str, const char *c)
{
    (void)st_lst;
    (void)str;
    (void)c;
    cur->num_ses++;
    cur->se_size++;
    cur->se_stack = realloc(cur->se_stack, sizeof(int) * cur->se_size);
    cur->se_stack[cur->se_size - 1] = cur->num_ses - 1;

    cur->ses = realloc(cur->ses, sizeof(char*) * cur->num_ses);
    cur->ses[cur->num_ses - 1] = malloc(1);
    *cur->ses[cur->num_ses - 1] = 0;

    if (cur->state->next1)
    {
        microregex_state_t newst = microregex_state_clone(cur);
        newst->state = cur->state->next1;
        microregex_nfa_addstateforward(&cur, newst);
    }
}

static void
microregex_regex_pop_submatch(microregex_state_t *st_lst, microregex_state_t cur, 
                              const char *str, const char *c)
{
    (void)st_lst;
    (void)str;
    (void)c;
    cur->se_size--;
    cur->se_stack = realloc(cur->se_stack, sizeof(int) * cur->se_size);

    if (cur->state->next1)
    {
        microregex_state_t newst = microregex_state_clone(cur);
        newst->state = cur->state->next1;
        microregex_nfa_addstateforward(&cur, newst);
    }
}

static void
microregex_regex_pass(microregex_state_t *st_lst, microregex_state_t cur, 
                      const char *str, const char *c)
{
    (void)st_lst;
    (void)str;
    (void)c;
    if (cur->state->next1)
    {
        microregex_state_t newst = microregex_state_clone(cur);
        newst->state = cur->state->next1;
        microregex_nfa_addstateforward(&cur, newst);
    }
}

static void
microregex_regex_backtrack_check(microregex_state_t *st_lst, microregex_state_t cur, 
                                 const char *str, const char *c)
{
    char *curstr = cur->ses[cur->state->loop_to];

    (void)st_lst;
    (void)str;

    if (!cur->insensitive && curstr[cur->loop_count] != *c) return;
    if (cur->insensitive && tolower(curstr[cur->loop_count]) != tolower(*c)) return;
    
    if (cur->state->next1 && !curstr[cur->loop_count])
    {
        microregex_state_t newst = microregex_state_clone(cur);
        newst->state = cur->state->next1;
        microregex_nfa_addstateforward(&cur, newst);
    }
    else
    {
        microregex_state_t newst = microregex_state_clone(cur);
        newst->loop_count++;
        microregex_nfa_addstateback(&cur, newst);
    }
}

static void
microregex_regex_alteration(microregex_state_t *st_lst, microregex_state_t cur, 
                            const char *str, const char *c)
{
    microregex_state_t newst1 = microregex_state_clone(cur);
    (void)st_lst;
    (void)str;
    (void)c;

    if (cur->state->next2)
    {
        microregex_state_t newst2 = microregex_state_clone(cur);
        newst2->state = cur->state->next2;
        microregex_nfa_addstateforward(&cur, newst2);
    }
    if (cur->state->next1)
    {
        newst1->state = cur->state->next1;
        microregex_nfa_addstateforward(&cur, newst1);
    }
}

static void
microregex_regex_counted_loop(microregex_state_t *st_lst, microregex_state_t cur, 
                             const char *str, const char *c)
{
    microregex_state_t newst1 = microregex_state_clone(cur);
 
    (void)st_lst;
    (void)str;
    (void)c;
    
    if (newst1->loop_count >= cur->state->loop_from)
    {
        microregex_state_t newst2 = microregex_state_clone(cur);
        newst2->state = 
            (cur->state->next2->state_type == COUNTED_END) ? 
            cur->state->next2 :
            cur->state->next1;
        microregex_nfa_addstateforward(&cur, newst2);
    }
    
    newst1->loop_count++;
    if (newst1->loop_count <= cur->state->loop_to)
    {
        newst1->state = 
            (cur->state->next1->state_type == COUNTED_END) ? 
            cur->state->next2 :
            cur->state->next1;
        microregex_nfa_addstateforward(&cur, newst1);
    }
}

static void
microregex_regex_counted_end(microregex_state_t *st_lst, microregex_state_t cur, 
                             const char *str, const char *c)
{
    (void)st_lst;
    (void)str;
    (void)c;
    
    if (cur->state->next1)
    {
        microregex_state_t newst = microregex_state_clone(cur);
        newst->state = cur->state->next1;
        microregex_nfa_addstateforward(&cur, newst);
    }
}

static void
microregex_regex_counted_begin(microregex_state_t *st_lst, microregex_state_t cur, 
                               const char *str, const char *c)
{
    (void)st_lst;
    (void)str;
    (void)c;
    
    if (cur->state->next1)
    {
        microregex_state_t newst = microregex_state_clone(cur);
        newst->loop_count = 0;
        newst->state = cur->state->next1;
        microregex_nfa_addstateforward(&cur, newst);
    }
}

static void
microregex_regex_assert_end(microregex_state_t *st_lst, microregex_state_t cur, 
                            const char *str, const char *c)
{
    (void)st_lst;
    (void)str;

    if (*c == 0 && cur->state->next1)
    {
        microregex_state_t newst = microregex_state_clone(cur);
        newst->state = cur->state->next1;
        microregex_nfa_addstateforward(&cur, newst);
    }
}

static void
microregex_regex_assert_beginning(microregex_state_t *st_lst, microregex_state_t cur, 
                                  const char *str, const char *c)
{
    (void)st_lst;
    
    if (c == str && cur->state->next1)
    {
        microregex_state_t newst = microregex_state_clone(cur);
        newst->state = cur->state->next1;
        microregex_nfa_addstateforward(&cur, newst);
    }
}

static void
microregex_regex_character_class(microregex_state_t *st_lst, microregex_state_t cur, 
                                 const char *str, const char *c)
{
    char *tmp = cur->state->character_class;
    int res = cur->state->invert, i;
    char ma = *c;
 
    (void)st_lst;
    (void)str;
    
    while(*tmp)
    {
        char ch = *tmp++;

        if (cur->insensitive)
        {
            ch = tolower(ch);
            ma = tolower(ma);
        } 
        
        if (cur->state->invert) res &= (ch != ma);
        else res |= (ch == ma);
    }
    
    if (!res) 
    {
        return;
    }
    
    for (i = 0; i < cur->se_size; i++)
    {
        char *ptr = cur->ses[cur->se_stack[i]];
        int len = strlen(ptr);
        ptr = realloc(ptr, len + 2);
        ptr[len] = *c;
        ptr[len + 1] = 0;
        cur->ses[cur->se_stack[i]] = ptr;
    }

    cur->match_end++;

    if (cur->state->next1)
    {
        microregex_state_t newst = microregex_state_clone(cur);
        newst->state = cur->state->next1;
        microregex_nfa_addstateback(&cur, newst);
    }
}

static void
microregex_regex_word_boundary(microregex_state_t *st_lst, microregex_state_t cur, 
                               const char *str, const char *c)
{
    int left = 0, right = 0, res;
    (void)st_lst;

    if (c == str) 
    {
        left = 1;
        right = isalnum(*c) || (*c == '_');
    }
    else if (!*c) 
    {
        right = 1;
        left = isalnum(*(c-1)) || (*(c-1) == '_');
    }
    else
    {
        if ((left = !(isalnum(*c) || (*c == '_'))))
        {
            right = isalnum(*(c+1)) || (*(c+1) == '_');
        }
        else
        {
            right = !(isalnum(*(c+1)) || (*(c+1) == '_'));
        }
    }
    
    if (cur->state->invert)
        res = !(left && right);
    else
        res = (left && right);

    if (cur->state->next1 && res)
    {
        microregex_state_t newst = microregex_state_clone(cur);
        newst->state = cur->state->next1;
        microregex_nfa_addstateforward(&cur, newst);
    }
}

microregex_state_t 
microregex_match(microregex_t regex_obj, const char *str, int start)
{
    microregex_state_t st_lst = NULL;
    microregex_state_t flist = NULL;
    microregex_state_t begin = microregex_state_new(&flist);
    microregex_state_t cur = begin, ret = NULL, tmp;
    const char *curstr = str + start;

    begin->state = regex_obj->nfa;
    begin->insensitive = regex_obj->case_insensitive;
    begin->match_begin = begin->match_end = start;
    microregex_nfa_addstateforward(&st_lst, microregex_state_new(&flist));
    microregex_nfa_addstateforward(&st_lst, begin);

    while(*curstr)
    {
        while (cur != st_lst)
        {
            microregex_state_t next;

            if (cur->state->state_type != REGEX_END)
            {
                (state_functions[cur->state->state_type])(&st_lst, cur, str, curstr);
                next = cur->next;
                tmp = cur;
                microregex_nfa_remove(&cur);
                if (flist == tmp)
                {
                    flist = tmp->nextL;
                    flist->prevL = NULL;
                    microregex_state_destroy_ex(&tmp, 0);
                }
                else
                {
                    microregex_state_destroy_ex(&tmp, 1);
                }
            }
            else
            {
                next = cur->next;
            }

            cur = next;
        }
        
        /* Add beginning state for next letter */
        if (*(curstr + 1))
        {
            microregex_state_t newbegin = microregex_state_new(&flist);
            newbegin->state = regex_obj->nfa;
            newbegin->insensitive = regex_obj->case_insensitive;
            newbegin->match_begin = newbegin->match_end = (curstr - str) + 1;
            microregex_nfa_addstateback(&st_lst, newbegin);
            cur = cur->next;
        }
        
        curstr++;
    }

    /* Loop one more time to process end-of-string conditions */
    cur = st_lst->next;
    while (cur != st_lst)
    {
        microregex_state_t next;

        if (cur->state->state_type != REGEX_END)
        {
            (state_functions[cur->state->state_type])(&st_lst, cur, str, curstr);
            next = cur->next;
            microregex_nfa_remove(&cur);
        }
        else
        {
            ret = microregex_state_clone_ex(cur, 0);
            microregex_state_destroy_list(flist);
            return ret;
        }
        
        cur = next;
    }
    microregex_state_destroy_list(flist);
    return ret;
}
                                                                                                 
int                                                                                                    
microregex_state_get_num_captured(microregex_state_t st)                                               
{                                                                                                      
    return st->num_ses;                                                                                
}                                                                                                      

int                                                                                                    
microregex_state_get_match_begin(microregex_state_t st)                                               
{                                                                                                      
    return st->match_begin;                                                                                
}
                                                                                                                                                                                                  
char *                                                                                                 
microregex_state_get_subexpression(microregex_state_t st, int n)                                       
{                                                                                                      
    if (n < 0 || n > microregex_state_get_num_captured(st))                                            
    {                                                                                                  
        return NULL;                                                                                   
    }                                                                                                  
                                                                                                       
    return st->ses[n];                                                                                 
}
