/*****************************************************************************
* Copyright (c) 2008, Mooneer Salem
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*     * Neither the name of the Kite Language organization nor the
*       names of its contributors may be used to endorse or promote products
*       derived from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
* THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>
#include "microregex_internal.h"

#include "config.h"
#ifdef HAVE_GC_H
#include <gc/gc.h>
#define malloc(x) (GC_malloc(x))
#define calloc(x, y) (GC_malloc(x * y))
#define realloc(x, y) (GC_realloc(x, y))
#define free(x) GC_free(x)
#endif /* HAVE_GC_H */

#define LOWERCASE_NEXT_LETTER (1 << 0)
#define LOWERCASE_UNTIL_END (1 << 1)
#define UPPERCASE_NEXT_LETTER (1 << 2)
#define UPPERCASE_UNTIL_END (1 << 3)
#define DISABLE_METACHARACTERS (1 << 4)
#define SET_REGEX_PARSER_FLAG(re, fl) re->parser_flags |= fl
#define CLEAR_REGEX_PARSER_FLAG(re, fl) re->parser_flags &= ~fl
#define IS_REGEX_PARSER_FLAG_SET(re, fl) (re->parser_flags & fl)

static microregex_nfa_t microregex_parse_regex(microregex_t, const char *, char **, int, microregex_nfa_t *);

static int
microregex_nfa_append(microregex_nfa_t *nfa, microregex_nfa_t end)
{
    int added = 0;
    if (!*nfa)
    {
        *nfa = end;
        return 1;
    }
    if ((*nfa)->state_type != PASS && (*nfa)->state_type != REGEX_END && (*nfa)->next1 != end)
        added = microregex_nfa_append(&(*nfa)->next1, end);
    if ((*nfa)->state_type == ALTERATION || (*nfa)->state_type == COUNTED_LOOP)
        if ((*nfa)->next2 != end) {
            int added2 = microregex_nfa_append(&(*nfa)->next2, end);
            added |= added2;
        }
    return added;
}


static void microregex_nfa_tolist (microregex_nfa_t ret, microregex_nfa_t *lst) {
  if (*lst) {
    ret->prevL = (*lst);
    ret->nextL = (*lst)->nextL;
    if (ret->nextL) ret->nextL->prevL = ret;
    (*lst)->nextL = ret;
  } else {
    ret->prevL = ret->nextL = NULL;
    *lst = ret;
  }
}


static microregex_nfa_t
microregex_nfa_backtrack_check(int backtrack, microregex_nfa_t *lst)
{
    microregex_nfa_t ret = (microregex_nfa_t)malloc(sizeof(struct microregex_nfa_t));
    microregex_nfa_tolist(ret, lst);
    ret->state_type = BACKTRACK_CHECK;
    ret->character_class = NULL;
    ret->invert = 0;
    ret->loop_to = backtrack;
    ret->next1 = NULL;
    ret->next2 = NULL;
    return ret;
}

static microregex_nfa_t
microregex_nfa_alteration(microregex_nfa_t next1, microregex_nfa_t next2, microregex_nfa_t *lst)
{
    microregex_nfa_t ret = (microregex_nfa_t)malloc(sizeof(struct microregex_nfa_t));
    microregex_nfa_tolist(ret, lst);
    ret->state_type = ALTERATION;
    ret->character_class = NULL;
    ret->invert = 0;
    ret->next1 = next1;
    ret->next2 = next2;
    return ret;
}

static microregex_nfa_t
microregex_nfa_assert_beginning(microregex_nfa_t *lst)
{
    microregex_nfa_t ret = (microregex_nfa_t)malloc(sizeof(struct microregex_nfa_t));
    microregex_nfa_tolist(ret, lst);
    ret->state_type = ASSERT_BEGINNING;
    ret->character_class = NULL;
    ret->invert = 0;
    ret->next1 = NULL;
    ret->next2 = NULL;
    return ret;
}

static microregex_nfa_t
microregex_nfa_assert_end(microregex_nfa_t *lst)
{
    microregex_nfa_t ret = (microregex_nfa_t)malloc(sizeof(struct microregex_nfa_t));
    microregex_nfa_tolist(ret, lst);
    ret->state_type = ASSERT_END;
    ret->character_class = NULL;
    ret->invert = 0;
    ret->next1 = NULL;
    ret->next2 = NULL;
    return ret;
}

static microregex_nfa_t
microregex_nfa_counted_begin(microregex_nfa_t *lst)
{
    microregex_nfa_t ret = (microregex_nfa_t)malloc(sizeof(struct microregex_nfa_t));
    microregex_nfa_tolist(ret, lst);
    ret->state_type = COUNTED_BEGIN;
    ret->character_class = NULL;
    ret->invert = 0;
    ret->next1 = NULL;
    ret->next2 = NULL;
    return ret;
}

static microregex_nfa_t
microregex_nfa_counted_loop(int from, int to, microregex_nfa_t *lst)
{
    microregex_nfa_t ret = (microregex_nfa_t)malloc(sizeof(struct microregex_nfa_t));
    microregex_nfa_tolist(ret, lst);
    ret->state_type = COUNTED_LOOP;
    ret->character_class = NULL;
    ret->invert = 0;
    ret->loop_from = from;
    ret->loop_to = to;
    ret->next1 = NULL;
    ret->next2 = NULL;
    return ret;
}

static microregex_nfa_t
microregex_nfa_counted_end(microregex_nfa_t *lst)
{
    microregex_nfa_t ret = (microregex_nfa_t)malloc(sizeof(struct microregex_nfa_t));
    microregex_nfa_tolist(ret, lst);
    ret->state_type = COUNTED_END;
    ret->character_class = NULL;
    ret->invert = 0;
    ret->next1 = NULL;
    ret->next2 = NULL;
    return ret;
}

static microregex_nfa_t
microregex_nfa_push_submatch(microregex_nfa_t *lst)
{
    microregex_nfa_t ret = (microregex_nfa_t)malloc(sizeof(struct microregex_nfa_t));
    microregex_nfa_tolist(ret, lst);
    ret->state_type = PUSH_SUBMATCH;
    ret->character_class = NULL;
    ret->invert = 0;
    ret->next1 = NULL;
    ret->next2 = NULL;
    return ret;
}

static microregex_nfa_t
microregex_nfa_pop_submatch(microregex_nfa_t *lst)
{
    microregex_nfa_t ret = (microregex_nfa_t)malloc(sizeof(struct microregex_nfa_t));
    microregex_nfa_tolist(ret, lst);
    ret->state_type = POP_SUBMATCH;
    ret->character_class = NULL;
    ret->invert = 0;
    ret->next1 = NULL;
    ret->next2 = NULL;
    return ret;
}

static microregex_nfa_t
microregex_nfa_end(microregex_nfa_t *lst)
{
    microregex_nfa_t ret = (microregex_nfa_t)malloc(sizeof(struct microregex_nfa_t));
    microregex_nfa_tolist(ret, lst);
    ret->state_type = REGEX_END;
    ret->character_class = NULL;
    ret->invert = 0;
    ret->next1 = NULL;
    ret->next2 = NULL;
    return ret;
}

static microregex_nfa_t
microregex_nfa_pass(microregex_nfa_t *lst)
{
    microregex_nfa_t ret = (microregex_nfa_t)malloc(sizeof(struct microregex_nfa_t));
    microregex_nfa_tolist(ret, lst);
    ret->state_type = PASS;
    ret->character_class = NULL;
    ret->invert = 0;
    ret->next1 = NULL;
    ret->next2 = NULL;
    return ret;
}

static microregex_nfa_t
microregex_nfa_oneplus(microregex_nfa_t one, int greedy, microregex_nfa_t *lst)
{
    microregex_nfa_t pass = microregex_nfa_pass(lst);
    microregex_nfa_t ret = 
        microregex_nfa_alteration(
            greedy ? pass : NULL,
            greedy ? NULL : pass
        , lst);
    microregex_nfa_append(&one, ret);
    pass->next1 = one;
    return one;
}

static microregex_nfa_t
microregex_nfa_zeroplus(microregex_nfa_t one, int greedy, microregex_nfa_t *lst)
{
    microregex_nfa_t ret = 
        microregex_nfa_alteration(
            greedy ? one : NULL, 
            greedy ? NULL : one
        , lst);
    microregex_nfa_t pass = microregex_nfa_pass(lst);
    microregex_nfa_append(&one, pass);
    pass->next1 = ret;
    return ret;
}

static microregex_nfa_t
microregex_nfa_zeroone(microregex_t regex, microregex_nfa_t one, int greedy, int p, microregex_nfa_t *lst)
{
    microregex_nfa_t pass = microregex_nfa_pass(lst);
    microregex_nfa_t ce = microregex_nfa_counted_end(lst);
    microregex_nfa_t parens = NULL;
    int i;
    microregex_nfa_t ret;

    for (i = p + 1; i <= regex->max_paren_count; i++) {
        microregex_nfa_append(&parens, microregex_nfa_push_submatch(lst));
    }
    for (i = regex->max_paren_count; i >= p + 1; i--) {
        microregex_nfa_append(&parens, microregex_nfa_pop_submatch(lst));
    }
    microregex_nfa_append(&parens, ce);

    ret = 
        microregex_nfa_alteration(
            greedy ? one : parens, 
            greedy ? parens : one
        , lst);
    microregex_nfa_append(&one, pass);
    pass->next1 = ce;
    return ret;
}

static const char *
microregex_nfa_charliteral_process(const microregex_t rgx_obj, const char *regex, char *buf)
{
    int i, j;

    memset(buf, 0, 256);

    if (IS_REGEX_PARSER_FLAG_SET(rgx_obj, DISABLE_METACHARACTERS))
    {
        goto make_literal;
    }
    
    if (*regex == '.')
    {
        for (i = 1; i <= 255; i++)
        {
            buf[i - 1] = i;
        }
    }
    else if (*regex == '\\' && *(regex + 1))
    {
        regex++;
        switch(*regex)
        {
            case 'c':
            {
                char esc_type = tolower(*(regex + 2));
                if (esc_type >= 'a' && esc_type <= 'z')
                {
                    buf[0] = 1 + (esc_type - 'a');
                }
                else if (esc_type == '[')
                {
                    buf[0] = '\x1b';
                }
                else if (esc_type == '\\')
                {
                    buf[0] = '\x1c';
                }
                else if (esc_type == ']')
                {
                    buf[0] = '\x1d';
                }
                else if (esc_type == '^')
                {
                    buf[0] = '\x1e';
                }
                else if (esc_type == '_')
                {
                    buf[0] = '\x1f';
                }
                
                if (*buf) regex++;
                else buf[0] = 'c';
                break;
            }
            case 't':
            {
                buf[0] = '\t';
                break;
            }
            case 'n':
            {
                buf[0] = '\n';
                break;
            }
            case 'r':
            {
                buf[0] = '\r';
                break;
            }
            case 'b':
            {
                buf[0] = '\b';
                break;
            }
            case 'f':
            {
                buf[0] = '\f';
                break;
            }
            case 'a':
            {
                buf[0] = '\a';
                break;
            }
            case 'e':
            {
                buf[0] = '\x1b';
                break;
            }
            case '0':
            {   char *tt;
                buf[0] = (char)strtol(regex, &tt, 8);
                regex = tt;
                break;
            }
            case 'x':
            {   char *tt;
                buf[0] = (char)strtol(regex, &tt, 16);
                regex = tt;
                break;
            }
            case 'w':
            {
                for (i = 1, j = 0; i <= 255; i++)
                {
                    if (isalnum(i) || ((char)i) == '_')
                    {
                        buf[j++] = (char)i;
                    }
                }
                break;
            }
            case 'W':
            {
                for (i = 1, j = 0; i <= 255; i++)
                {
                    if (!isalnum(i) && (!((char)i) == '_'))
                    {
                        buf[j++] = (char)i;
                    }
                }
                break;
            }
            case 's':
            {
                for (i = 1, j = 0; i <= 255; i++)
                {
                    if (isspace(i))
                    {
                        buf[j++] = (char)i;
                    }
                }
                break;
            }
            case 'S':
            {
                for (i = 1, j = 0; i <= 255; i++)
                {
                    if (!isspace(i))
                    {
                        buf[j++] = (char)i;
                    }
                }
                break;
            }
            case 'd':
            {
                for (i = 1, j = 0; i <= 255; i++)
                {
                    if (isdigit(i))
                    {
                        buf[j++] = (char)i;
                    }
                }
                break;
            }
            case 'D':
            {
                for (i = 1, j = 0; i <= 255; i++)
                {
                    if (!isdigit(i))
                    {
                        buf[j++] = (char)i;
                    }
                }
                break;
            }
            default:
            {
                buf[0] = *regex;
                break;
            }
        }
    }
    else
    {
make_literal:
        if (IS_REGEX_PARSER_FLAG_SET(rgx_obj, LOWERCASE_NEXT_LETTER))
        {
            buf[0] = tolower(*regex);
            CLEAR_REGEX_PARSER_FLAG(rgx_obj, LOWERCASE_NEXT_LETTER);
        }
        else if (IS_REGEX_PARSER_FLAG_SET(rgx_obj, UPPERCASE_NEXT_LETTER))
        {
            buf[0] = toupper(*regex);
            CLEAR_REGEX_PARSER_FLAG(rgx_obj, UPPERCASE_NEXT_LETTER);
        }
        else if (IS_REGEX_PARSER_FLAG_SET(rgx_obj, LOWERCASE_UNTIL_END))
        {
            buf[0] = tolower(*regex);
        }
        else if (IS_REGEX_PARSER_FLAG_SET(rgx_obj, UPPERCASE_UNTIL_END))
        {
            buf[0] = toupper(*regex);
        }
        else
        {
            buf[0] = *regex;
        }
    }

    return ++regex;
}

static microregex_nfa_t
microregex_nfa_charliteral(const microregex_t rgx_obj, const char *regex, const char **end_location, microregex_nfa_t *lst)
{
    microregex_nfa_t ret = (microregex_nfa_t)malloc(sizeof(struct microregex_nfa_t));
    char tmp_buf[256];
    microregex_nfa_t bt;

    microregex_nfa_tolist(ret, lst);

    ret->state_type = CHARACTER_CLASS;
    ret->character_class = NULL;
    ret->invert = 0;
    ret->next1 = NULL;
    ret->next2 = NULL;

    memset(tmp_buf, 0, 256);
    ret->character_class = malloc(1);
    ret->character_class[0] = 0;

try_another:
    if (!*regex)
    {
        free(ret->character_class);
        free(ret);
        return NULL;
    }
    
    if (IS_REGEX_PARSER_FLAG_SET(rgx_obj, DISABLE_METACHARACTERS))
    {
        goto literal_chars;
    }
    
    if (*regex == '[')
    {
        const char *tmp = regex + 1;
        if (*tmp == '^') 
        {
            ret->invert++;
            tmp++;
        }
        regex = tmp;
        while(*tmp && *tmp != ']')
        {
            if (strlen(tmp_buf) == 1 && *tmp == '-' && *(tmp + 1) != ']')
            {
                char from = tmp_buf[0];
                char b[2];
                tmp = microregex_nfa_charliteral_process(rgx_obj, tmp + 1, tmp_buf);
                if (strlen(tmp_buf) == 1 && tmp_buf[0] > from)
                {
                    for (; from <= tmp_buf[0]; from++)
                    {
                        ret->character_class = 
                            realloc(
                                ret->character_class, 
                                strlen(ret->character_class) + 2
                            );
                        b[0] = from;
                        b[1] = 0;
                        strcat(ret->character_class, b);
                    }
                }
                else
                {
                    ret->character_class = 
                        realloc(
                            ret->character_class, 
                            strlen(ret->character_class) + 3 + strlen(tmp_buf)
                        );
                    strcat(ret->character_class, tmp_buf);
                    b[0] = from;
                    b[1] = 0;
                    strcat(ret->character_class, b);
                    b[0] = '-';
                    strcat(ret->character_class, b);
                }
                continue;
            }
            else
            {
                ret->character_class = 
                    realloc(
                        ret->character_class, 
                        strlen(ret->character_class) + strlen(tmp_buf) + 1
                    );
                strcat(ret->character_class, tmp_buf);
            }

            tmp = microregex_nfa_charliteral_process(rgx_obj, tmp, tmp_buf);
            if (!*tmp)
            {
                ret->character_class = 
                    realloc(
                        ret->character_class, 
                        strlen(ret->character_class) + strlen(tmp_buf) + 1
                    );
                strcat(ret->character_class, tmp_buf);
            }
        }

        if (strlen(tmp_buf) > 0)
        {
            ret->character_class = 
                realloc(
                    ret->character_class, 
                    strlen(ret->character_class) + strlen(tmp_buf) + 1
                );
            strcat(ret->character_class, tmp_buf);
        }

        regex = tmp;
    }
    else
    {
literal_chars:
        if (IS_REGEX_PARSER_FLAG_SET(rgx_obj, DISABLE_METACHARACTERS) && *regex == '\\' && *(regex+1) == 'E')
        {
            rgx_obj->parser_flags = 0;
            regex += 2;
            goto try_another;
        }
        else if (IS_REGEX_PARSER_FLAG_SET(rgx_obj, DISABLE_METACHARACTERS))
        {
            goto default_literal_chars;
        }
        
        if (*regex == '\\' && *(regex+1) == 'l')
        {
            SET_REGEX_PARSER_FLAG(rgx_obj, LOWERCASE_NEXT_LETTER);
            CLEAR_REGEX_PARSER_FLAG(rgx_obj, UPPERCASE_NEXT_LETTER);
            regex += 2;
            goto try_another;
        }
        else if (*regex == '\\' && *(regex+1) == 'u')
        {
            SET_REGEX_PARSER_FLAG(rgx_obj, UPPERCASE_NEXT_LETTER);
            CLEAR_REGEX_PARSER_FLAG(rgx_obj, LOWERCASE_NEXT_LETTER);
            regex += 2;
            goto try_another;
        }
        else if (*regex == '\\' && *(regex+1) == 'L')
        {
            SET_REGEX_PARSER_FLAG(rgx_obj, LOWERCASE_UNTIL_END);
            CLEAR_REGEX_PARSER_FLAG(rgx_obj, UPPERCASE_UNTIL_END);
            regex += 2;
            goto try_another;
        }
        else if (*regex == '\\' && *(regex+1) == 'U')
        {
            SET_REGEX_PARSER_FLAG(rgx_obj, UPPERCASE_UNTIL_END);
            CLEAR_REGEX_PARSER_FLAG(rgx_obj, LOWERCASE_UNTIL_END);
            regex += 2;
            goto try_another;
        }
        else if (*regex == '\\' && *(regex+1) == 'Q')
        {
            rgx_obj->parser_flags = 0;
            SET_REGEX_PARSER_FLAG(rgx_obj, DISABLE_METACHARACTERS);
            regex += 2;
            goto try_another;
        }
        else if (*regex == '\\' && *(regex+1) == 'E')
        {
            rgx_obj->parser_flags = 0;
            regex += 2;
            goto try_another;
        }
        else if (*regex == '\\' && *(regex+1) == 'b')
        {
            /* word boundary */
            regex += 2;
            ret->state_type = WORD_BOUNDARY;
        }
        else if (*regex == '\\' && *(regex+1) == 'B')
        {
            /* word boundary */
            regex += 2;
            ret->state_type = WORD_BOUNDARY;
            ret->invert = 1;
        }
        else if (*regex == '\\' && *(regex+1) == 'A')
        {
            /* Beginning of string */
            regex += 2;
            ret->state_type = ASSERT_BEGINNING;
        }
        else if (*regex == '\\' && *(regex+1) == 'Z')
        {
            /* End of string */
            regex += 2;
            ret->state_type = ASSERT_END;
        }
        else if (*regex == '\\' && *(regex+1) == 'z')
        {
            /* End of string */
            regex += 2;
            ret->state_type = ASSERT_END;
        }
        else if (*regex == '\\' && *(regex+1) >= '1' && *(regex+1) <= '9')
        {
            /* backtracking support */
            int backtrack;
            char *endp;
            
            backtrack = strtol(regex+1, &endp, 10);
            regex = endp;
            
            ret->state_type = COUNTED_BEGIN;
            bt = microregex_nfa_backtrack_check(backtrack, lst);
            ret->next1 = bt;
        }
        else
        {
default_literal_chars:
            regex = microregex_nfa_charliteral_process(rgx_obj, regex, tmp_buf);
            ret->character_class = 
                realloc(
                    ret->character_class, 
                    strlen(ret->character_class) + strlen(tmp_buf) + 1
                );
            strcat(ret->character_class, tmp_buf);
        }
        regex--;
    }

    if (end_location) *end_location = regex;
    return ret;
}

/**
 * Create regular expression object.
 * @param regex The regular expression to parse.
 * @return The created regular expression object.
 */
microregex_t
microregex_create(const char *regex)
{
    microregex_t ret = (microregex_t)malloc(sizeof(struct microregex_t));
    ret->paren_count = 0;
    ret->max_paren_count = 0;
    ret->regex = regex;
    ret->err_string = NULL;
    ret->nlist = NULL;
    ret->parser_flags = 0;
    ret->nfa = microregex_parse_regex(ret, regex, NULL, 0, &ret->nlist);
    microregex_nfa_append(&ret->nfa, microregex_nfa_end(&ret->nlist));

    return ret;
}

/**
 * Destroy regular expression object.
 * @param regex_obj The regular expression object to destroy.
 */
void
microregex_destroy(microregex_t *regex_obj)
{
    microregex_nfa_t lst;
    microregex_nfa_t nn;

    if (regex_obj == NULL || (*regex_obj) == NULL) return;
    if ((*regex_obj)->err_string) free((*regex_obj)->err_string);

    lst = (*regex_obj)->nlist;
    while (lst) {
      nn = lst->nextL;
      free(lst->character_class);
      free(lst);
      lst = nn;
    }

    free(*regex_obj);
    *regex_obj = NULL;
}

/**
 * Set case insensitivity of regular expression.
 * @param regex_obj The regular expression object to edit.
 * @param val Value to set flag to.
 */
void microregex_set_case_insensitivity(microregex_t regex_obj, int val)
{
    regex_obj->case_insensitive = val;
}

/**
 * Parse regular expression into an NFA.
 * @param regex_obj The regular expression object with state.
 * @param regex The regular expression to parse.
 * @param end_char The location where it stopped parsing
 * @param paren_start The starting nesting level.
 * @return The created NFA object.
 */
static microregex_nfa_t
microregex_parse_regex(microregex_t regex_obj, const char *regex, char **end_char, int paren_start, microregex_nfa_t *lst)
{
    int cur_paren = regex_obj->paren_count;
    microregex_nfa_t nfa = NULL;
    microregex_nfa_t prev = NULL;
    char *end_location;
    int greedy;

    while(*regex && regex_obj->paren_count >= cur_paren) 
    {
        if (prev && *regex != '+' && *regex != '?' && *regex != '*' && *regex != '{')
        {
            microregex_nfa_append(&nfa, prev);
            prev = NULL;
        } 
        else if (prev && IS_REGEX_PARSER_FLAG_SET(regex_obj, DISABLE_METACHARACTERS))
        {
            microregex_nfa_append(&nfa, prev);
            prev = NULL;
        }
        
        if (IS_REGEX_PARSER_FLAG_SET(regex_obj, DISABLE_METACHARACTERS))
        {
            goto default_character;
        }
        
        switch(*regex)
        {
            case '(':
            {
                regex_obj->paren_count++;
                regex_obj->max_paren_count++;
                microregex_nfa_append(
                    &prev,
                    microregex_nfa_push_submatch(lst)
                );
                microregex_nfa_append(
                    &prev,
                    microregex_parse_regex(
                        regex_obj,
                        regex + 1,
                        &end_location,
                        paren_start + 1,
                        lst
                    )
                );

                regex = end_location;
                break;
            }
            case ')':
            {
                if (regex_obj->paren_count == paren_start)
                {
                    microregex_nfa_append(
                        &prev,
                        microregex_nfa_pop_submatch(lst)
                    );
                }
                
                regex_obj->paren_count--;
                regex--; /* We increment at the bottom of the loop */
                break;
            }
            case '|':
            {
                /* Alternation */
                microregex_nfa_t alt;

                regex_obj->paren_count++;
                
                alt = microregex_nfa_alteration(
                    nfa,
                    microregex_parse_regex(
                        regex_obj,
                        regex + 1,
                        &end_location,
                        paren_start,
                        lst
                    ), lst
                );

                regex = end_location - 1;
                
                nfa = alt;
                break;
            }
            case '+':
            {
                greedy = (*(++regex) != '?');

                /* One or more */
                prev =
                    microregex_nfa_oneplus(
                        prev,
                        greedy,
                        lst
                    );

                if (*regex != '?') regex--;
                break;
            }
            case '*':
            {
                greedy = (*(++regex) != '?');

                /* Zero or more */
                prev = 
                    microregex_nfa_zeroplus(
                        prev,
                        greedy,
                        lst
                    );

                if (*regex != '?') regex--;
                break;
            }
            case '?':
            {
                greedy = (*(++regex) != '?');

                /* Zero or one */
                prev =    
                    microregex_nfa_zeroone(
                        regex_obj,
                        prev,
                        greedy,
                        cur_paren,
                        lst
                    );

                if (*regex != '?') regex--;
                break;
            }
            case '{':
            {
                /* Counted repetition */
                const char *cur_pos = regex; char *tt;
                int from = strtol(regex + 1, &tt, 10);
                int to;
                microregex_nfa_t tmpnfa, counted_loop, tmppass;
                regex = tt;
                if (!*regex || (*regex != ',' && *regex != '}'))
                {
                    regex = cur_pos;
                    goto default_character;
                }
                else if (*regex == ',')
                {
                    to = strtol(regex + 1, &tt, 10);
                    regex = tt;
                    if (*regex != '}')
                    {
                        regex = cur_pos;
                        goto default_character;
                    }
                    else if (to == 0)
                    {
                        to = INT_MAX;
                    }
                }
                else
                {
                    to = from;
                }

                tmpnfa = microregex_nfa_counted_begin(lst);
                counted_loop = microregex_nfa_counted_loop(from, to, lst);
                greedy = (*(++regex) != '?');
                tmppass = microregex_nfa_pass(lst);
                tmppass->next1 = counted_loop;

                counted_loop->next1 = greedy ? prev : microregex_nfa_counted_end(lst);
                counted_loop->next2 = greedy ? microregex_nfa_counted_end(lst) : prev;
                microregex_nfa_append(
                    &prev,
                    tmppass
                );

                tmpnfa->next1 = counted_loop;

                prev = tmpnfa;
                if (greedy) regex--;
                break;
            }
            case '^':
            {
                /* Beginning of string */
                microregex_nfa_append(
                    &nfa,
                    microregex_nfa_assert_beginning(lst)
                );
                break;
            }
            case '$':
            {
                /* End of string */
                microregex_nfa_append(
                    &nfa,
                    microregex_nfa_assert_end(lst)
                );
                break;
            }
            case '[':
            default:
            {
default_character:
                /* Character class/literal */
                prev = microregex_nfa_charliteral(
                    regex_obj,
                    regex,
                    (const char **)(&end_location),
                    lst
                );
                regex = end_location;
            }
        };

        regex++;
    }

    if (prev) microregex_nfa_append(&nfa, prev);
    if (end_char) *end_char = (char *)regex;
    if (!*regex) regex_obj->parser_flags = 0;
    return nfa;
}
