/*****************************************************************************
 * Copyright (c) 2007, Mooneer Salem
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Kite Language organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#define KITE_EXECUTE_C
#include "objs/kite_object.h"
#include "kite_vm.h"
#include "kite_opcodes.h"

/*
 * Valid operator strings
 */
char *OPERATOR_STRINGS[] = {
    "+",
    "-",
    "*",
    "/",
    "%",
    "unary +",
    "unary -",
    "=",
    "==",
    "!=",
    "<",
    ">",
    "<=",
    ">=",
    "and",
    "or",
    "not",
    "xor",
    "<<",
    ">>",
    "map",
    "reduce",
    "[]",
    "[]=",
    "|",
    ".",
    NULL,
};

char *OPERATOR_NAME_STRINGS[] = {
    "plus",
    "minus",
    "multiply",
    "divide",
    "mod",
    "unplus",
    "unminus",
    "assign",
    "equals",
    "nequals",
    "lt",
    "gt",
    "leq",
    "geq",
    "and",
    "or",
    "not",
    "xor",
    "lshift",
    "rshift",
    "map",
    "reduce",
    "array",
    "array_set",
    "call",
    "property",
    "number_of_operators"
};

char *OPERATOR_METHODS[] = {
    "__op_plus__",
    "__op_minus__",
    "__op_multiply__",
    "__op_divide__",
    "__op_mod__",
    "__op_unplus__",
    "__op_unminus__",
    "__op_assign__",
    "__op_equals__",
    "__op_nequals__",
    "__op_lt__",
    "__op_gt__",
    "__op_leq__",
    "__op_geq__",
    "__op_and__",
    "__op_or__",
    "__op_not__",
    "__op_xor__",
    "__op_lshift__",
    "__op_rshift__",
    "__op_map__",
    "__op_reduce__",
    "__op_array__",
    "__op_array_set__",
    "__op_call__",
    "__op_property__",
    "__op_number_of_operators__"
};

/*****************************************************************************
 * Kite opcode: Push onto execution stack
 ****************************************************************************/
static kite_opcode_t *
kite_vm_execute_push(kite_thread_t *thd, kite_opcode_push *ins)
{
    assert(ins->obj != NULL);
    kite_vm_push(thd, ins->obj);
    return ins->common.next;
}

/*****************************************************************************
 * Kite opcode: Place function arguments into symbol table.
 ****************************************************************************/
static kite_opcode_t *
kite_vm_execute_funcargs(kite_thread_t *thd, kite_opcode_funcargs *ins)
{
    kite_object_t *args, *cur_array, *cur_args, *arg;
    kite_symtab_t *symt;
    kite_funccall_info_t *stackentry;
    int deref_array = 0;
    assert(ins->args != NULL);
    
    args = kite_vm_pop(thd);
    if (args->type != OBJ_LIST) {
        deref_array = 1;
        args = kite_list_object(thd, args);
    }

    cur_array = args;
    cur_args = ins->args;
    
    while(cur_array && cur_args) {
        stackentry = (kite_funccall_info_t*)kite_top_stack(thd->func_stack);
        arg = cur_args->builtin_data.listvalue.car;
        
        if (arg != NULL) {
#ifndef HAVE_GC_H
            symt = (kite_symtab_t*)malloc(sizeof(kite_symtab_t));
            symt->name = strdup(arg->builtin_data.stringvalue.string);
#else
            symt = (kite_symtab_t*)GC_malloc(sizeof(kite_symtab_t));
            symt->name = (void*)GC_malloc_atomic(strlen(KITE_GET_STRING_VALUE(arg)) + 1);
            strcpy(symt->name, KITE_GET_STRING_VALUE(arg));
#endif /* HAVE_GC_H */
            symt->global = 0;
            symt->left = symt->right = NULL;
            symt->docstring = NULL;
            symt->value = 
                kite_reference_object(cur_array->builtin_data.listvalue.car);
            kite_symtab_insert(thd, &stackentry->locals, symt);
        }
    
        cur_array = cur_array->builtin_data.listvalue.cdr;
        cur_args = cur_args->builtin_data.listvalue.cdr;
    }

    if (deref_array) kite_dereference_object(args);
    assert(cur_array == NULL && cur_args == NULL);
    return ins->common.next;
}

/*****************************************************************************
 * Kite opcode: Dereference variable
 ****************************************************************************/
static kite_opcode_t *
kite_vm_execute_deref_1(kite_thread_t *thd, kite_opcode_deref_1 *ins)
{
    kite_object_t *ident, *res = NULL, *exc;
    kite_stack_entry_t *cur = 
        &thd->func_stack->stack[thd->func_stack->length - 1];
    kite_symtab_t *found;
    kite_funccall_info_t *stackentry, *fc_entry, *oldentry, *fc;
    int searchthis = 0;
    char *ident_string;
    int old_sys;

    assert(cur != NULL);

    ident = kite_vm_pop(thd);
    assert(ident->type == OBJ_IDENT);
    ident_string = ident->builtin_data.stringvalue.string;

    if (((kite_funccall_info_t*)cur->obj)->last_deref &&
        !strcmp(((kite_funccall_info_t*)cur->obj)->last_deref, ident_string)) {
        found = ((kite_funccall_info_t*)cur->obj)->last_deref_obj;
        kite_reference_object(found->value);
        kite_push_stack(&thd->running_stack, TRUE, found);
        return ins->common.next;
    } else if (strcmp("this", ident_string) == 0) {
        res = kite_reference_object(((kite_funccall_info_t*)cur->obj)->this);
    } else if (strcmp("__root", ident_string) == 0) {
        res = kite_reference_object(thd->vm->root_package);
    } else if (strcmp("base", ident_string) == 0) {
        kite_symtab_t *cur_method;
        
        KITE_FIND_IN_SYMTAB(
            cur_method, ((kite_funccall_info_t*)cur->obj)->locals,
            "__current", 0
        );
        
        if (cur_method && cur_method->value && cur_method->value->parent)
        {
            res = kite_reference_object(cur_method->value->parent->object_data.inherit_from);
        }
        else
        {
            res = kite_new_null(thd);
        }
    } else {
        do {
            stackentry = (kite_funccall_info_t*)cur->obj;
            if (searchthis && stackentry->this) {
                KITE_FIND_ANY_IN_SYMTAB(found, 
                    stackentry->this->object_data.properties,
                    ident_string);
            } else {
                KITE_FIND_ANY_IN_SYMTAB(found, stackentry->locals,
                    ident_string);
            }
            if (found) {
                if (found->value) res = found->value;
                else res = found->value = kite_new_null(thd);
                break;
            } else {
                if (searchthis) {
                    searchthis = 0;
                    cur--;
                } else {
                    searchthis = 1;
                }
            }
        } while(cur >= thd->func_stack->stack);
        
        /* If not in the stack, find in the root namespace */
        if (res == NULL) {
            KITE_FIND_PROPERTY(found, thd->vm->root_package, 
                ident_string, 1);
            if (found) {
                res = found->value;
            }
        }
        
        if (res == NULL && ins->create) {
            /* Create entry in symbol table */
            stackentry = (kite_funccall_info_t*)kite_top_stack(thd->func_stack);
#ifndef HAVE_GC_H
            found = (kite_symtab_t*)malloc(sizeof(kite_symtab_t));
            found->name = strdup(ident_string);
#else
            found = (kite_symtab_t*)GC_malloc(sizeof(kite_symtab_t));
            found->name = GC_malloc_atomic(strlen(ident_string) + 1);
            strcpy(found->name, ident_string);
#endif /* HAVE_GC_H */
            found->global = 0;
            found->docstring = NULL;
            found->left = found->right = NULL;
            found->value = kite_new_null(thd);
            kite_symtab_insert(thd, &stackentry->locals, found);
            res = found->value;
        } else if (res == NULL && !ins->create) {
            char buf[1024];
            fc = kite_top_stack(thd->func_stack);
            PUSH_FUNC_STACK_WITH_RET_VALUE(NULL, ins->common.next);
            fc_entry->sys_or_user = fc->sys_or_user;
            
            sprintf(buf, "Could not find property or variable %s", 
                ident_string);
            exc = kite_new_exception(thd, "System.exceptions.ObjectNotFound", 
                                     buf);
            kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);
            POP_FUNC_STACK;
            return ins->common.next;
        }
        
        if (!found->value->object_data.inherit_from)
        {
            kite_object_t *sysobj =
                kite_dereference_and_load(thd, "System.object");
            if (found->value != sysobj)
                found->value->object_data.inherit_from = sysobj;
        }

        if (!found->value->shareable && found->value->owner_thread != thd) 
            goto thread_access_err;
        
        kite_reference_object(found->value);
        kite_push_stack(&thd->running_stack, TRUE, found);
        kite_dereference_object(ident);

        stackentry->last_deref = ident_string;
        stackentry->last_deref_obj = found;

        return ins->common.next;
    }

    /* Wrap it in a reference */
    if (!res->shareable && res->owner_thread != thd) 
        goto thread_access_err;
        
    kite_dereference_object(ident);
    kite_vm_push(thd, res);
    return ins->common.next;

thread_access_err:
    {
        PUSH_FUNC_STACK_WITH_RET_VALUE(NULL, ins->common.next);
        exc = 
            kite_new_exception(
                thd, 
                "System.exceptions.AccessDenied", 
                "Cannot access unshared object."
            );
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);
    }
    return ins->common.next;
}

/*****************************************************************************
 * Kite opcode: Define variable
 ****************************************************************************/
static kite_opcode_t *
kite_vm_execute_defprop(kite_thread_t *thd, kite_opcode_defprop *ins)
{
    kite_object_t *ident;
    kite_symtab_t *found;
    kite_funccall_info_t *stackentry;

    (void)ins;
    ident = kite_vm_pop(thd);
    assert(ident->type == OBJ_IDENT);

    /* Create entry in symbol table */
    stackentry = (kite_funccall_info_t*)kite_top_stack(thd->func_stack);
    KITE_FIND_IN_SYMTAB(found, stackentry->locals, ident->builtin_data.stringvalue.string, 0)
    if (found == NULL) {
#ifndef HAVE_GC_H
        found = (kite_symtab_t*)malloc(sizeof(kite_symtab_t));
        found->name = (char*)strdup(KITE_GET_STRING_VALUE(ident));
#else
        found = (kite_symtab_t*)GC_malloc(sizeof(kite_symtab_t));
        found->name = (char*)GC_malloc_atomic(strlen(KITE_GET_STRING_VALUE(ident)) + 1);
        strcpy(found->name, KITE_GET_STRING_VALUE(ident));
#endif /* HAVE_GC_H */
        found->global = 0;
        found->docstring = NULL;
        found->left = found->right = NULL;
        found->value = kite_new_null(thd);
        kite_symtab_insert(thd, &stackentry->locals, found);
        stackentry->last_deref = NULL;
        stackentry->last_deref_obj = NULL;
    }

    kite_dereference_object(ident);
    return ins->common.next;
}

/*****************************************************************************
 * Kite opcode: Dereference object property
 ****************************************************************************/
static kite_opcode_t *
kite_vm_execute_deref_2(kite_thread_t *thd, kite_opcode_deref_2 *ins)
{
    kite_object_t *ident, *obj, *res = NULL, *curr, *exc;
    kite_symtab_t *entry;
    char buf[1024];
    char *name;
    kite_funccall_info_t *oldentry, *fc_entry;
    int old_sys;

    ident = kite_vm_pop(thd);
    obj = kite_reference_object(kite_vm_pop(thd));
    
    curr = obj;
    name = ident->builtin_data.stringvalue.string;
    
    do {
        KITE_FIND_ANY_IN_SYMTAB(entry, curr->object_data.properties, name);
        if (entry) {
            if ((!entry->global && obj != curr) ||
                (entry->value && entry->value->type == OBJ_METHOD && 
                 obj != curr)) {
                break;
            }
            
            if (!entry->value) entry->value = kite_new_null(thd);
            else kite_reference_object(entry->value);
            
            if (!entry->value->shareable && entry->value->owner_thread != thd) goto thread_access_err;
            
            kite_push_stack(&thd->running_stack, TRUE, entry);
            kite_dereference_object(ident);
            kite_dereference_object(obj);
            return ins->common.next;
        }
        
        curr = curr->object_data.inherit_from;
    } while(curr != NULL);
    
    if (curr) {
        kite_add_property(thd, obj, name, 0, NULL);
        
        if (entry->value && entry->value->type == OBJ_METHOD) {
            if (entry->value->builtin_data.funcvalue.functype == FUNC_BYTECODE)
                res = kite_new_method_bytecode(thd, NULL, 0);
            else
                res = kite_new_method_compiled(thd, NULL, 0);            
            memcpy(&res->builtin_data.funcvalue,
                   &entry->value->builtin_data.funcvalue,
                   sizeof(kite_function_t));
            res->builtin_data.funcvalue.this = obj;
        } else {
            res = kite_new_null(thd);
        }
        kite_set_property(thd, obj, name, res);
        
        KITE_FIND_IN_SYMTAB(entry, obj->object_data.properties, name, 0);
        
        if (!entry->value->shareable && entry->value->owner_thread != thd) goto thread_access_err;
        
        kite_push_stack(&thd->running_stack, TRUE, entry);
        kite_dereference_object(res);
        kite_dereference_object(ident);
        kite_dereference_object(obj);
        return ins->common.next;
    }
    
    curr = kite_new_list(thd);
    curr->builtin_data.listvalue.car = kite_new_string(thd, name);
    curr->builtin_data.listvalue.cdr = NULL;
            
    if (kite_vm_call_operator(thd, obj, OP_PROPERTY, curr, 0)) {
        kite_dereference_object(curr);
        return ins->common.next;
    }
        
    kite_dereference_object(curr);
    
    {
        kite_funccall_info_t *fc = 
            kite_top_stack(thd->func_stack);
        PUSH_FUNC_STACK_WITH_RET_VALUE(obj, ins->common.next);
        fc_entry->sys_or_user = fc->sys_or_user;
        
        sprintf(buf, "Could not find property %s", name);
        exc = kite_new_exception(thd, "System.exceptions.ObjectNotFound", 
                                 buf);
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);
        return ins->common.next;
    }

thread_access_err:
    {
        PUSH_FUNC_STACK_WITH_RET_VALUE(NULL, ins->common.next);
        exc = 
            kite_new_exception(
                thd, 
                "System.exceptions.AccessDenied", 
                "Cannot access unshared object."
            );
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);
    }
    return ins->common.next;
}

/*****************************************************************************
 * Kite opcode: Call method
 ****************************************************************************/
static kite_opcode_t *
kite_vm_execute_call(kite_thread_t *thd, kite_opcode_call *ins)
{
    kite_object_t *args, *method, *this, *exc, *tmp;
    int curstack;
    kite_funccall_info_t *fc;
    kite_funccall_info_t *fc_entry, *oldentry;
    int old_sys;

    args = kite_vm_pop(thd);
    method = kite_reference_object(kite_vm_pop(thd));
    this = kite_vm_pop(thd);

    curstack = thd->running_stack->length - 1;

    if (method->type == OBJ_METHOD && method->builtin_data.funcvalue.this) {
        this = method->builtin_data.funcvalue.this;
    }
    
    fc = kite_top_stack(thd->func_stack);
    PUSH_FUNC_STACK_WITH_RET_VALUE(this, ins->common.next);
    fc_entry->sys_or_user = fc->sys_or_user;
    
    if (method->type == OBJ_METHOD ||
        method->type == OBJ_CLASS) {
        if (kite_list_count(thd, args) != 
            method->builtin_data.funcvalue.numargs) {
            exc = kite_new_exception(thd, "System.exceptions.InvalidArgument", 
                                     "Specified arguments do not match.");
            kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);
            return ins->common.next;
        }
        kite_vm_call_object(thd, this, method, args);
    } else if (method->type == OBJ_IDENT) {
        kite_vm_call_method(thd, this, method->builtin_data.stringvalue.string,
                            args, TRUE);
    } else {
        exc = kite_new_exception(thd, "System.exceptions.TypeMismatch", 
                                 "Cannot call object as a method.");
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);
        return ins->common.next;
    }
    
    kite_dereference_object(args);
    kite_dereference_object(method);

    if (thd->exception) return ins->common.next;

    /* Remove anything other than the return value from the top of the stack */
    if ((thd->running_stack->length - 1) != curstack) {
        kite_symtab_t *st;
        kite_object_t *ret = NULL;
        
        if (thd->running_stack->stack[thd->running_stack->length - 1].reference == 2)
        {
            st = ((kite_symtab_t*)thd->running_stack->stack[thd->running_stack->length - 1].obj);
            (void)kite_vm_pop(thd);
        }
        else
        {
            ret = kite_vm_pop(thd);
        }
        
        while((thd->running_stack->length - 1) != curstack) {
            if (thd->running_stack->stack[thd->running_stack->length - 1].reference) {
                (void)kite_vm_pop(thd);
            } else {
                tmp = kite_vm_pop(thd);
                kite_dereference_object(tmp);
            }
        }
        if (strcmp("__construct__", method->builtin_data.stringvalue.string) != 0)
        {
            if (ret != NULL) kite_vm_push(thd, ret);
            else kite_push_stack(&thd->running_stack, TRUE, st);
        }
        POP_FUNC_STACK;
    } else {
        POP_FUNC_STACK;
        if (strcmp("__construct__", method->builtin_data.stringvalue.string) != 0)
        {
            kite_vm_push(thd, kite_new_null(thd));
            return ins->common.next;
        }
    }
    
    return ins->common.next;
}

/*****************************************************************************
 * Kite opcode: Execute code
 ****************************************************************************/
static kite_opcode_t *
kite_vm_execute_eval(kite_thread_t *thd, kite_opcode_eval *ins)
{
    kite_object_t *code;

    (void)ins;
    code = kite_vm_pop(thd);
    kite_vm_execute_user_method(thd, code->builtin_data.funcvalue.funcptr);
    kite_dereference_object(code);      
    
    return ins->common.next;          
}

/*****************************************************************************
 * Kite opcode: Construct a list
 ****************************************************************************/
static kite_opcode_t *
kite_vm_execute_list_cons(kite_thread_t *thd, kite_opcode_list_cons *ins)
{
    kite_object_t *car, *cdr, *res;

    (void)ins;

    cdr = kite_vm_pop(thd);
    car = kite_vm_pop(thd);
    res = kite_new_list(thd);
    
    if (cdr->type != OBJ_LIST) {
        res->builtin_data.listvalue.cdr = kite_list_object(thd, cdr);
        kite_dereference_object(cdr);
    } else {
        res->builtin_data.listvalue.cdr = cdr;
    }
    res->builtin_data.listvalue.car = car;
    
    kite_vm_push(thd, res);
    kite_dereference_object(res);
    return ins->common.next;
}

/*****************************************************************************
 * Kite opcode: Construct a list
 ****************************************************************************/
static kite_opcode_t *
kite_vm_execute_list_cons_1(kite_thread_t *thd, kite_opcode_list_cons_1 *ins)
{
    kite_object_t *car, *res;
    
    (void)ins;

    car = kite_vm_pop(thd);
    res = kite_new_list(thd);
    res->builtin_data.listvalue.car = car;
    
    kite_vm_push(thd, res);
    kite_dereference_object(res);
    
    return ins->common.next;
}

/*****************************************************************************
 * Kite opcode: Construct a list
 ****************************************************************************/
static kite_opcode_t *
kite_vm_execute_list_cons_0(kite_thread_t *thd, kite_opcode_list_cons_0 *ins)
{
    kite_object_t *res;
    
    (void)ins;

    res = kite_new_list(thd);
    
    kite_vm_push(thd, res);
    kite_dereference_object(res);
    
    return ins->common.next;
}

/*****************************************************************************
 * Kite opcode: Duplicate top of stack
 ****************************************************************************/
static kite_opcode_t *
kite_vm_execute_dupe_top(kite_thread_t *thd, kite_opcode_dupe_top *ins)
{
    kite_object_t *top;
    kite_symtab_t *symt;

    if (thd->running_stack->stack[thd->running_stack->length - 1].reference) {
        symt = (kite_symtab_t*)(thd->running_stack->stack[thd->running_stack->length - 1].obj);
        if (!ins->dupe_ref) kite_reference_object(symt->value);
        kite_push_stack(&thd->running_stack, TRUE, symt);
    } else {
        top = kite_vm_pop(thd);
        kite_vm_push(thd, top);
        kite_vm_push(thd, top);
        kite_dereference_object(top);
    }
    
    return ins->common.next;
}
    
/*****************************************************************************
 * Kite opcode: Jump if true
 ****************************************************************************/
static void *
kite_vm_execute_jumptrue(kite_thread_t *thd, kite_opcode_jump_true *ins)
{
    kite_object_t *obj, *onstack;
    void *newpc;
    assert(ins->jumpto != NULL && thd->func_stack != NULL);

    onstack = kite_vm_pop(thd);
    obj = kite_boolean_object(thd, onstack);
    if (obj->builtin_data.intvalue)
    {
        newpc = ins->jumpto;
    }
    else
    {
        newpc = ins->common.next;
    }
    kite_dereference_object(obj);
    kite_dereference_object(onstack);
    return newpc;
}

/*****************************************************************************
 * Kite opcode: Jump if false
 ****************************************************************************/
static void *
kite_vm_execute_jumpfalse(kite_thread_t *thd, kite_opcode_jump_false *ins)
{
    kite_object_t *obj, *onstack;
    void *newpc;
    assert(ins->jumpto != NULL && thd->func_stack != NULL);
 
    onstack = kite_vm_pop(thd);
    obj = kite_boolean_object(thd, onstack);
    if (!obj->builtin_data.intvalue)
    {
        newpc = ins->jumpto;
    }
    else
    {
        newpc = ins->common.next;
    }
    kite_dereference_object(obj);
    kite_dereference_object(onstack);
    return newpc;
}

/*****************************************************************************
 * Kite opcode: Jump unconditional
 ****************************************************************************/
static void *
kite_vm_execute_jumpuncond(kite_thread_t *thd, kite_opcode_jump_unconditional *ins)
{
    assert(ins->jumpto != NULL && thd->func_stack != NULL);
    return ins->jumpto;
}

/*****************************************************************************
 * Kite opcode: Handle exception
 ****************************************************************************/
static void *kite_vm_execute_exception_hndl(
    kite_thread_t *thd, kite_opcode_exception_hndl *ins)
{
    int tos = thd->running_stack->length;
    int func_stack = thd->func_stack->length;
    kite_symtab_t *propinfo;
    kite_funccall_info_t *current;

    kite_vm_execute_user_method(thd, ins->common.next);
    if (thd->exception)
    {
        while(thd->running_stack && thd->running_stack->length != tos) {
            kite_object_t *obj = kite_vm_pop(thd);
            kite_dereference_object(obj);
        }
        while(thd->func_stack && thd->func_stack->length != func_stack) {
            kite_funccall_info_t *obj = kite_pop_stack(&thd->func_stack);
#ifndef HAVE_GC_H
            free(obj);
#else
            (void)obj;
#endif /* HAVE_GC_H */
        }
        
#ifndef HAVE_GC_H
        propinfo = (kite_symtab_t*)malloc(sizeof(kite_symtab_t));
#else
        propinfo = (kite_symtab_t*)GC_malloc(sizeof(kite_symtab_t));
#endif /* HAVE_GC_H */
        assert(propinfo != NULL);
#ifndef HAVE_GC_H
        propinfo->name = strdup("__exc");
#else
        propinfo->name = GC_malloc_atomic(strlen("__exc") + 1);
        strcpy(propinfo->name, "__exc");
#endif /* HAVE_GC_H */
        propinfo->global = 0;
        propinfo->docstring = NULL;
        propinfo->left = NULL;
        propinfo->right = NULL;
        propinfo->value = thd->exception;
        
        current = 
            (kite_funccall_info_t*)(thd->func_stack->stack[thd->func_stack->length - 1].obj);
        kite_symtab_insert(thd, &current->locals, propinfo);
        current->last_deref = NULL;
        current->last_deref_obj = NULL;

        thd->exception = NULL;
        return ins->handler;
    }
    
    return NULL;
}

/*****************************************************************************
 * Kite opcode: Handle exception
 ****************************************************************************/
static kite_opcode_t * kite_vm_execute_pop_exception_hndl(
    kite_thread_t *thd, kite_opcode_pop_exception_hndl *ins)
{
    kite_symtab_t *symt;
    kite_funccall_info_t *current = 
        (kite_funccall_info_t*)(thd->func_stack->stack[thd->func_stack->length - 1].obj);
    (void)ins;

    KITE_FIND_IN_SYMTAB(symt, current->locals, "__exc", 0);
    if (symt) kite_symtab_remove(thd, &current->locals, "__exc", 0);
    
    return ins->common.next;
}

/*****************************************************************************
 * Kite opcode: Get "this"
 ****************************************************************************/
static kite_opcode_t * kite_vm_execute_this(kite_thread_t *thd, kite_opcode_this *ins)
{
    kite_vm_push(thd, 
        ((kite_funccall_info_t*)kite_top_stack(thd->func_stack))->this);
    return ins->common.next;
}

/*****************************************************************************
 * Kite opcode: Execute arithmetic operation
 ****************************************************************************/
static kite_opcode_t *
kite_vm_execute_arithop(kite_thread_t *thd, kite_opcode_arithop *ins)
{
    kite_object_t *lhs, *rhs, *idx = NULL;
    kite_funccall_info_t *fc, *fc_entry, *oldentry;
    int old_sys;
    int curstack;

    assert(ins->operation < NUM_OPERATORS);
    rhs = kite_vm_pop(thd);

    if (ins->operation == OP_ARRAY_SET)
        idx = kite_vm_pop(thd);

    if (thd->running_stack->stack[thd->running_stack->length - 1].reference &&
        ins->operation == OP_ASSIGNMENT) {
        kite_symtab_t *symt = 
            (kite_symtab_t*)thd->running_stack->stack[thd->running_stack->length - 1].obj;
        kite_object_t *oldobj;
        if (symt->global == -1)
        {
            oldobj = *(kite_object_t**)symt->value;
            *(kite_object_t**)symt->value = kite_reference_object(rhs);
        }
        else
        {
            oldobj = symt->value;
            symt->value = kite_reference_object(rhs);
        }
        kite_dereference_object(oldobj);
        kite_dereference_object(oldobj->parent);
        if (oldobj->type != OBJ_CLASS)
            oldobj->parent = NULL;
        kite_dereference_object(kite_vm_pop(thd));
        return ins->common.next;
    }
    lhs = kite_vm_pop(thd);

    /* 
     * Short-circuit, most common operations--we don't need to call the method.
     */ 
    if (lhs->type == rhs->type && (
            lhs->type == OBJ_INTEGER ||
            lhs->type == OBJ_FLOAT ||
            lhs->type == OBJ_BOOLEAN
        ) &&
        ins->operation != OP_ASSIGNMENT)
    {
        kite_symtab_t *st;

        KITE_FIND_ANY_IN_SYMTAB(st, 
            lhs->object_data.inherit_from->object_data.properties, 
            OP_TO_METHOD(ins->operation));
        
        if (st)
        {
            kite_compiled_func_t func =
                (st->value->builtin_data.funcvalue.funcptr);
            (*func)(thd, lhs, rhs);
            kite_dereference_object(lhs);
            kite_dereference_object(rhs);
            return ins->common.next;
        }
    }

    fc = kite_top_stack(thd->func_stack);
    PUSH_FUNC_STACK_WITH_RET_VALUE(lhs, ins->common.next);
    fc_entry->sys_or_user = fc->sys_or_user;

    curstack = thd->running_stack->length - 1;
    if (idx)
    {
        idx = kite_list_object(thd, idx);
        kite_append_list(thd, idx, rhs);
        kite_vm_call_operator(thd, lhs, ins->operation, idx, TRUE);
    }
    else
        kite_vm_call_operator(thd, lhs, ins->operation, rhs, TRUE);
    if (thd->exception) return ins->common.next;

    kite_dereference_object(lhs);
    kite_dereference_object(rhs);

    /* Remove anything other than the return value from the top of the stack */
    if ((thd->running_stack->length - 1) != curstack) {
        kite_symtab_t *st;
        kite_object_t *ret = NULL;
        if (thd->running_stack->stack[thd->running_stack->length - 1].reference == 2)
        {
            st = ((kite_symtab_t*)thd->running_stack->stack[thd->running_stack->length - 1].obj);
            (void)kite_vm_pop(thd);
        }
        else
        {
            ret = kite_vm_pop(thd);
        }
        while((thd->running_stack->length - 1) != curstack) {
            kite_dereference_object(kite_vm_pop(thd));
        }
        if (ret)
        {
            kite_vm_push(thd, ret);
            kite_dereference_object(ret);
        }
        else
        {
            kite_push_stack(&thd->running_stack, 2, st);
        }
        POP_FUNC_STACK;
    } else {
        POP_FUNC_STACK;
        kite_vm_push(thd, kite_new_null(thd));
    }
    
    return ins->common.next;
}

/*****************************************************************************
 * Kite opcode: Execute nop operation
 ****************************************************************************/
static kite_opcode_t*
kite_vm_execute_nop(kite_thread_t *thd, kite_opcode_nop *ins)
{
    (void)thd;
    return ins->common.next;
}

/*****************************************************************************
 * Kite opcode: Execute make operation
 ****************************************************************************/
static kite_opcode_t *
kite_vm_execute_make(kite_thread_t *thd, kite_opcode_make *ins)
{
    kite_object_t *args, *obj, *newobj;
    int current_stack;
    kite_funccall_info_t *fc, *fc_entry, *oldentry;
    int old_sys;

    args = kite_vm_pop(thd);
    obj = kite_vm_pop(thd);
    current_stack = thd->running_stack->length;

    fc = kite_top_stack(thd->func_stack);
    PUSH_FUNC_STACK_WITH_RET_VALUE(obj, ins->common.next);
    fc_entry->sys_or_user = fc->sys_or_user;
    
    newobj = kite_new_instance_with_constructor(thd, obj, args);
    if (thd->exception) return ins->common.next;
    POP_FUNC_STACK;

    if (thd->running_stack->length != current_stack) {
        while(thd->running_stack->length != current_stack) {
            obj = kite_vm_pop(thd);
            kite_dereference_object(obj);
        }
    }

    kite_vm_push(thd, newobj);
    kite_dereference_object(newobj);
    return ins->common.next;
}

/*****************************************************************************
 * Kite opcode: Execute is/isof operation
 ****************************************************************************/
static kite_opcode_t *
kite_vm_execute_objis(kite_thread_t *thd, kite_opcode_objis *ins)
{
    kite_object_t *class, *obj, *cur;
    int res = FALSE;
    
    class = kite_vm_pop(thd);
    obj = kite_vm_pop(thd);
    
    if (obj->type != OBJ_CLASS) {
        cur = obj->object_data.inherit_from;
    } else {
        cur = obj;
    }
    
    do {
        if (cur == class) res = TRUE;
        
        cur = cur->object_data.inherit_from;
    } while(cur && !res && ins->isof);

    obj = kite_new_boolean(thd, res);
    kite_vm_push(thd, obj);
    return ins->common.next;
    //kite_dereference_object(obj);
}

/*****************************************************************************
 * Call method on object programmatically
 ****************************************************************************/
void kite_vm_call_method(kite_thread_t *thd, kite_object_t *obj, 
                         char *name, struct kite_object_t *args, int err)
{
    kite_symtab_t *found = NULL;
    kite_object_t *curr = obj, *exc;
    char buf[1024];
    int numargs = kite_list_count(thd, args);
    
    do {
        KITE_FIND_METHOD(found, curr, name, numargs);
        if (found && found->value->type == OBJ_METHOD) {
            kite_vm_call_object(thd, obj, found->value, args);
            return;
        } else if (found) {
            sprintf(buf, "%s is not callable.", name);
            exc = kite_new_exception(thd, "System.exceptions.TypeMismatch", 
                                     buf);
            kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);
            return;
        }
        
        curr = curr->object_data.inherit_from;
    } while(curr);
    
    /* Call function operator if method's not found */
    if (strcmp("__construct__", name) != 0 &&
        strcmp("__destruct__", name) != 0) {
        curr = kite_new_list(thd);
        curr->builtin_data.listvalue.car = kite_new_string(thd, name);
        curr->builtin_data.listvalue.cdr = kite_new_list(thd);
        curr->builtin_data.listvalue.cdr->builtin_data.listvalue.car =
            kite_reference_object(args);
            
        if (kite_vm_call_operator(thd, obj, OP_METHOD_CALL, curr, 0)) {
            kite_dereference_object(curr);
            return;
        }
        
        kite_dereference_object(curr);
    }
    
    /* Method not found, raise exception */
    if (err) {
        if (strcmp("__construct__", name) == 0) {
            sprintf(buf, "Could not find constructor that takes %d argument(s)",
                    numargs);
        } else {
            sprintf(buf, "Could not find method %s that takes %d argument(s)",
                    name, numargs);
        }
        exc = kite_new_exception(thd, "System.exceptions.ObjectNotFound", 
                                 buf);
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);
        return;
    }
}

/*****************************************************************************
 * Call method on object programmatically
 ****************************************************************************/
void 
kite_vm_call_object(kite_thread_t *thd, struct kite_object_t *this, 
                    struct kite_object_t *method, struct kite_object_t *args)
{
    kite_compiled_func_t func;
    int freestack = 0;
    kite_object_t *oldthis = NULL;
    kite_symtab_t *propinfo;
    kite_funccall_info_t *current;

    /*
     * [Ticket 5] Replace with current this if we're simply calling a parent 
     * class's method.
     */
    if (this->type != OBJ_INSTANCE && thd->func_stack && 
        thd->func_stack->length > 1) {
        kite_funccall_info_t *current = 
            (kite_funccall_info_t*)(thd->func_stack->stack[thd->func_stack->length - 2].obj);
        kite_object_t *cur_this = current->this;
        while(cur_this != NULL) {
            if (cur_this == this && current->this->type == OBJ_INSTANCE) {
                oldthis = this;
                this = current->this;
                break;
            }
            cur_this = cur_this->object_data.inherit_from;
        }
    }
    
    assert(method->type == OBJ_METHOD || method->type == OBJ_CLASS ||
           method->type == OBJ_INSTANCE);
    if (method->builtin_data.funcvalue.functype == FUNC_COMPILED) {
        func = method->builtin_data.funcvalue.funcptr;
        func(thd, this, args);
        return;
    }
    
    /* Non-compiled method */
    if (!thd->func_stack || thd->func_stack->length == 0) {
        kite_funccall_info_t *entry = calloc(sizeof(kite_funccall_info_t), 1);
        freestack = 1;
        entry->this = this;
        entry->file = "(Kite loader)";
        kite_push_stack(&thd->func_stack, FALSE, entry);
    } else {
        kite_funccall_info_t *entry = 
            (kite_funccall_info_t*)kite_top_stack(thd->func_stack);
        entry->this = this;
        entry->last_deref = NULL;
        entry->last_deref_obj = NULL;
    }
    
    // Request by ozzloy @ irc.freenode.net/#sdcolleges 6/5/2008
    // __caller variable will point to currently running method.
    // Perfect for anonymous recursive functions.    
#ifndef HAVE_GC_H
    propinfo = (kite_symtab_t*)malloc(sizeof(kite_symtab_t));
#else
    propinfo = (kite_symtab_t*)GC_malloc(sizeof(kite_symtab_t));
#endif /* HAVE_GC_H */
    assert(propinfo != NULL);
#ifndef HAVE_GC_H
    propinfo->name = strdup("__current");
#else
    propinfo->name = GC_malloc_atomic(strlen("__current") + 1);
    strcpy(propinfo->name, "__current");
#endif /* HAVE_GC_H */
    propinfo->global = 0;
    propinfo->docstring = NULL;
    propinfo->left = NULL;
    propinfo->right = NULL;
    propinfo->value = kite_reference_object(method);
    
    current = 
        (kite_funccall_info_t*)(thd->func_stack->stack[thd->func_stack->length - 1].obj);
    kite_symtab_insert(thd, &current->locals, propinfo);
    
    // Push arguments onto stack and begin executing function.
    kite_vm_push(thd, args);
    kite_vm_execute_user_method(thd, method->builtin_data.funcvalue.funcptr);
    
    if (oldthis) {
        kite_funccall_info_t *entry = 
            (kite_funccall_info_t*)kite_top_stack(thd->func_stack);
        entry->this = oldthis;
    }

    if (freestack) {
        kite_funccall_info_t *f = kite_top_stack(thd->func_stack);
        if (f && f->locals) kite_destruct_symtab(thd, &f->locals, 0);
#ifndef HAVE_GC_H
        free(kite_pop_stack(&thd->func_stack));
#else
        kite_pop_stack(&thd->func_stack);
#endif /* HAVE_GC_H */
    }
}

/*****************************************************************************
 * Call constructor on object programmatically
 ****************************************************************************/
void kite_vm_call_constructor(kite_thread_t *thd, struct kite_object_t *this, 
                              struct kite_object_t *args) 
{
    kite_symtab_t *symt;
    kite_object_t *curr = this->object_data.inherit_from;
    
    /* 
     * Find __construct__ (with any arguments) in the object first. If we
     * can't find any, just ignore this call. Otherwise, call constructor
     * and have the system throw an exception if the arguments don't match
     * up.
     *
     * Note [Ticket 5]: we're only looking in the current object, not the
     * parents. Any constructors in the current object are responsible for
     * calling parent constructors as needed (by using 
     * classname.__construct__).
     *
     * Note2: We also want to spit out an error if we're trying to call a
     * constructor that takes arguments that does not exist.
     */    
    KITE_FIND_ANY_IN_SYMTAB(symt, curr->object_data.properties, 
        "__construct__");
    if (symt != NULL || kite_list_count(thd, args) > 0) {
        kite_vm_call_method(thd, this, "__construct__", args, TRUE);
        return;
    }
}

/*****************************************************************************
 * Call operator method on object programmatically
 ****************************************************************************/
int kite_vm_call_operator(kite_thread_t *thd, struct kite_object_t *this, 
                          int op, struct kite_object_t *arg, int err)
{
    kite_object_t *cur = this, *exc;
    kite_symtab_t *st;
    
    do {
        KITE_FIND_ANY_IN_SYMTAB(st, cur->object_data.properties, OP_TO_METHOD(op));
        if (st) {
            kite_vm_call_object(thd, this, st->value, arg);
            return 1;
        }
        
        cur = cur->object_data.inherit_from;
    } while(cur);
    
    /* Operator not defined, raise exception */
    if (err) {
        exc = kite_new_exception(thd, "System.exceptions.OperatorNotDefined", 
                                 OP_TO_STRING(op));
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);
    }
    return 0;
}

/*****************************************************************************
 * Handle exception thrown in user code or Kite standard library/VM
 ****************************************************************************/
void kite_vm_execute_exception(kite_thread_t *thd, int exit_now)
{
    kite_object_t *name = kite_object_name(thd, thd->exception);
    kite_symtab_t *msg, *trace;
    kite_object_t *cur = NULL;
    
    KITE_FIND_PROPERTY(msg, thd->exception, "msg", 0);
    KITE_FIND_PROPERTY(trace, thd->exception, "trace", 0);
    
    printf("%s: %s\n", name->builtin_data.stringvalue.string,
                       msg->value->builtin_data.stringvalue.string);
    if (trace) cur = trace->value;
    while(cur != NULL) {
        printf("  %s\n", 
            cur->builtin_data.listvalue.car->builtin_data.stringvalue.string);
        cur = cur->builtin_data.listvalue.cdr;
    }               
    kite_dereference_object(name);

    if (exit_now)
    {
        if (thd->vm->thread_list[0] == thd)
            exit(-1);

#ifndef WIN32
        pthread_exit(-1);
#else
        ExitThread(-1);
#endif /* WIN32 */
    }
}

/*****************************************************************************
 * Main execution loop for Kite opcodes
 ****************************************************************************/
#ifndef WIN32
void *kite_vm_execute(void *ptr)
#else
DWORD WINAPI kite_vm_execute(LPVOID ptr)
#endif /* WIN32 */
{
    kite_thread_t *thd = (kite_thread_t*)ptr;
    kite_object_t *args;
    
    if (thd->running_stack && thd->running_stack->length) {
        args = kite_vm_pop(thd);
    } else {
        args = kite_new_null(thd);
    }
    
    kite_vm_call_object(thd, thd->start, thd->start, args);
    kite_dereference_object(args);
    
    if (thd->exception) kite_vm_execute_exception(thd, TRUE);

#ifdef WIN32
    return (DWORD)0;
#else
    return NULL;
#endif /* WIN32 */
}

typedef kite_opcode_t *(*kite_opcode_func_t)(kite_thread_t*, kite_opcode_t*);

static kite_opcode_func_t opcode_funcs[] = 
{
    (kite_opcode_func_t)kite_vm_execute_jumptrue,
    (kite_opcode_func_t)kite_vm_execute_jumpfalse,
    (kite_opcode_func_t)kite_vm_execute_jumpuncond,
    (kite_opcode_func_t)kite_vm_execute_call,
    (kite_opcode_func_t)kite_vm_execute_deref_1,
    (kite_opcode_func_t)kite_vm_execute_deref_2,
    (kite_opcode_func_t)kite_vm_execute_dupe_top,
    (kite_opcode_func_t)kite_vm_execute_exception_hndl,
    (kite_opcode_func_t)kite_vm_execute_pop_exception_hndl,
    (kite_opcode_func_t)kite_vm_execute_list_cons,
    (kite_opcode_func_t)kite_vm_execute_list_cons_0,
    (kite_opcode_func_t)kite_vm_execute_list_cons_1,
    (kite_opcode_func_t)kite_vm_execute_push,
    (kite_opcode_func_t)kite_vm_execute_arithop,
    (kite_opcode_func_t)kite_vm_execute_funcargs,
    (kite_opcode_func_t)kite_vm_execute_make,
    (kite_opcode_func_t)kite_vm_execute_eval,
    (kite_opcode_func_t)kite_vm_execute_objis,
    (kite_opcode_func_t)kite_vm_execute_nop,
    (kite_opcode_func_t)kite_vm_execute_defprop,
    (kite_opcode_func_t)kite_vm_execute_this,
    NULL
};

/*****************************************************************************
 * Main execution loop for Kite opcodes
 ****************************************************************************/
static void kite_vm_execute_user_method_debug(kite_thread_t *thd, void *pc)
{
    kite_opcode_t *cur = (kite_opcode_t*)pc;
    kite_funccall_info_t *fc_entry, *oldentry;
    int old_sys;
    kite_object_t *args;
    kite_object_t *usethis;
    kite_funccall_info_t *fc = kite_top_stack(thd->func_stack);

    while(cur && cur->opcode != RETURN_NOW)
    {
        kite_opcode_call *ins = (kite_opcode_call*)cur;
        int curstack = thd->running_stack->length - 1;
        
        PUSH_FUNC_STACK(fc->this);
        fc_entry->sys_or_user = 1;
        
        /* Function arguments */
        args = kite_new_list(thd);
        kite_append_list(thd, args, kite_new_integer(thd, cur->opcode));
        kite_append_list(thd, args, kite_new_string(thd, cur->file));
        kite_append_list(thd, args, kite_new_integer(thd, cur->line));
        usethis = thd->vm->instruction_tracer->builtin_data.funcvalue.this;
        kite_vm_call_object(thd, usethis ? usethis : fc->this, thd->vm->instruction_tracer, args);
        kite_dereference_object(args);
        
        /* Ignore return value and anything else pushed onto the stack. */
        if ((thd->running_stack->length - 1) != curstack) {
            kite_object_t *ret = kite_vm_pop(thd);
            while((thd->running_stack->length - 1) != curstack) {
                if (thd->running_stack->stack[thd->running_stack->length - 1].reference) {
                    (void)kite_vm_pop(thd);
                } else {
                    kite_object_t *tmp = kite_vm_pop(thd);
                    kite_dereference_object(tmp);
                }
            }
            kite_dereference_object(ret);
        }
        
        POP_FUNC_STACK;

        /* Decode, execute */
        cur = (*opcode_funcs[cur->opcode])(thd, cur);
        if (thd->exception) return;
    }
    
    if (cur && cur->opcode == RETURN_NOW)
    {
        /* Exit out of method. */
        kite_dereference_object(thd->exception);
        thd->exception = NULL;
        return;
    }
}

/*****************************************************************************
 * Main execution loop for Kite opcodes
 ****************************************************************************/
void kite_vm_execute_user_method(kite_thread_t *thd, void *pc)
{
    kite_opcode_t *cur = (kite_opcode_t*)pc;
    kite_funccall_info_t *fc = kite_top_stack(thd->func_stack);

    if (!fc || !fc->sys_or_user)
    {
        if (thd->vm && thd->vm->instruction_tracer && thd->vm->instruction_tracer->type == OBJ_METHOD)
        {
            kite_vm_execute_user_method_debug(thd, pc);
            return;
        }
    }
    
    while(cur && cur->opcode != RETURN_NOW)
    {
        /* Decode, execute */
        cur = (*opcode_funcs[cur->opcode])(thd, cur);
        if (thd->exception) return;
    }
    
    if (cur && cur->opcode == RETURN_NOW)
    {
        /* Exit out of method. */
        kite_dereference_object(thd->exception);
        thd->exception = NULL;
        return;
    }
}
