/*****************************************************************************
 * Copyright (c) 2007, Mooneer Salem
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Kite Language organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "objs/kite_object.h"
#include "kite_vm.h"
#include "kite_opcodes.h"

#define ALLOCATE_OPCODE(name) (name*)kite_compile_common(sizeof(name))

/*****************************************************************************
 * Common function for compilation.
 ****************************************************************************/
static kite_opcode_t *kite_compile_common(size_t size)
{
#ifndef HAVE_GC_H
    kite_opcode_t *ret = (kite_opcode_t*)calloc(size, 1);
#else
    kite_opcode_t *ret = (kite_opcode_t*)GC_malloc(size);
#endif /* HAVE_GC_H */
    ret->length = size;
    return ret;
}

/*****************************************************************************
 * Compile an arithmetic operation
 ****************************************************************************/
kite_opcode_t *kite_compile_arithop(unsigned int operation)
{
    kite_opcode_arithop *op = ALLOCATE_OPCODE(kite_opcode_arithop);
    op->operation = operation;
    op->common.opcode = ARITH_OP;
    return (kite_opcode_t*)op;
}

/*****************************************************************************
 * Compile a push operation
 ****************************************************************************/
kite_opcode_t *kite_compile_push(struct kite_object_t *obj)
{
    kite_opcode_push *op = ALLOCATE_OPCODE(kite_opcode_push);
    op->obj = obj;
    op->common.opcode = PUSH;
    return (kite_opcode_t*)op;
}

/*****************************************************************************
 * Compile a "get this" operation
 ****************************************************************************/
kite_opcode_t *kite_compile_this()
{
    kite_opcode_t *opc = 
        (kite_opcode_t*)ALLOCATE_OPCODE(kite_opcode_this);
    opc->opcode = PUSH_THIS;
    return opc;
}

/*****************************************************************************
 * Compile a list construct operation
 ****************************************************************************/
kite_opcode_t *kite_compile_list_cons()
{
    kite_opcode_t *opc = 
        (kite_opcode_t*)ALLOCATE_OPCODE(kite_opcode_list_cons);
    opc->opcode = LIST_CONS;
    return opc;
}

/*****************************************************************************
 * Compile a list construct operation
 ****************************************************************************/
kite_opcode_t *kite_compile_list_cons_0()
{
    kite_opcode_t *opc = 
        (kite_opcode_t*)ALLOCATE_OPCODE(kite_opcode_list_cons_0);
    opc->opcode = LIST_CONS_0;
    return opc;
}

/*****************************************************************************
 * Compile a list construct operation
 ****************************************************************************/
kite_opcode_t *kite_compile_list_cons_1()
{
    kite_opcode_t *opc = 
        (kite_opcode_t*)ALLOCATE_OPCODE(kite_opcode_list_cons);
    opc->opcode = LIST_CONS_1;
    return opc;
}

/*****************************************************************************
 * Compile a deref_1 operation
 ****************************************************************************/
kite_opcode_t *kite_compile_deref_1(int create)
{
    kite_opcode_deref_1 *opc = ALLOCATE_OPCODE(kite_opcode_deref_1);
    opc->common.opcode = DEREF_1;
    opc->create = create;
    return (kite_opcode_t*)opc;
}

/*****************************************************************************
 * Compile a deref_2 operation
 ****************************************************************************/
kite_opcode_t *kite_compile_deref_2()
{
    kite_opcode_t *opc = (kite_opcode_t*)ALLOCATE_OPCODE(kite_opcode_deref_2);
    opc->opcode = DEREF_2;
    return opc;
}

/*****************************************************************************
 * Compile a call operation
 ****************************************************************************/
kite_opcode_t *kite_compile_call()
{
    kite_opcode_t *opc = (kite_opcode_t*)ALLOCATE_OPCODE(kite_opcode_call);
    opc->opcode = CALL;
    return opc;
}

/*****************************************************************************
 * Compile a jump if true operation
 ****************************************************************************/
kite_opcode_t *kite_compile_jump_true(kite_opcode_t *jumpto)
{
    kite_opcode_jump_true *op = ALLOCATE_OPCODE(kite_opcode_jump_true);
    op->jumpto = jumpto;
    op->common.opcode = JUMP_TRUE;
    return (kite_opcode_t*)op;
}

/*****************************************************************************
 * Compile a jump if false operation
 ****************************************************************************/
kite_opcode_t *kite_compile_jump_false(kite_opcode_t *jumpto)
{
    kite_opcode_jump_false *op = ALLOCATE_OPCODE(kite_opcode_jump_false);
    op->jumpto = jumpto;
    op->common.opcode = JUMP_FALSE;
    return (kite_opcode_t*)op;
}

/*****************************************************************************
 * Compile a jump unconditional
 ****************************************************************************/
kite_opcode_t *kite_compile_jump_uncond(kite_opcode_t *jumpto)
{
    kite_opcode_jump_unconditional *op = ALLOCATE_OPCODE(kite_opcode_jump_unconditional);
    op->jumpto = jumpto;
    op->common.opcode = JUMP_UNCOND;
    return (kite_opcode_t*)op;
}

/*****************************************************************************
 * Compile an exception handle operation
 ****************************************************************************/
kite_opcode_t *kite_compile_exception_hndl(kite_opcode_t *handler)
{
    kite_opcode_exception_hndl *op = 
        ALLOCATE_OPCODE(kite_opcode_exception_hndl);
    op->handler = handler;
    op->common.opcode = EXCEPTION_HNDL;
    return (kite_opcode_t*)op;
}

/*****************************************************************************
 * Compile a pop exception handler operation
 ****************************************************************************/
kite_opcode_t *kite_compile_pop_exception_hndl()
{
    kite_opcode_t *opc =
        (kite_opcode_t*)ALLOCATE_OPCODE(kite_opcode_pop_exception_hndl);
    opc->opcode = POP_EXCEPTION_HNDL;
    return opc;
}

/*****************************************************************************
 * Compile a duplicate top of stack operation
 ****************************************************************************/
kite_opcode_t *kite_compile_dupe_top(int dupe_ref)
{
    kite_opcode_dupe_top *opc = 
        (kite_opcode_dupe_top*)ALLOCATE_OPCODE(kite_opcode_dupe_top);
    opc->common.opcode = DUPE_TOP;
    opc->dupe_ref = dupe_ref;
    return (kite_opcode_t*)opc;
}

/*****************************************************************************
 * Compile funcargs instruction
 ****************************************************************************/
kite_opcode_t *kite_compile_funcargs(struct kite_object_t *args)
{
    kite_opcode_funcargs *op = ALLOCATE_OPCODE(kite_opcode_funcargs);
    op->args = args;
    op->common.opcode = FUNCARGS;
    return (kite_opcode_t*)op;
}

/*****************************************************************************
 * Compile a make operation
 ****************************************************************************/
kite_opcode_t *kite_compile_make()
{
    kite_opcode_t *opc = (kite_opcode_t*)ALLOCATE_OPCODE(kite_opcode_make);
    opc->opcode = MAKE_OBJ;
    return opc;
}

/*****************************************************************************
 * Compile an eval operation
 ****************************************************************************/
kite_opcode_t *kite_compile_eval()
{
    kite_opcode_t *opc = (kite_opcode_t*)ALLOCATE_OPCODE(kite_opcode_eval);
    opc->opcode = EVAL_OBJ;
    return opc;
}

/*****************************************************************************
 * Compile a nop operation
 ****************************************************************************/
kite_opcode_t *kite_compile_nop()
{
    kite_opcode_t *opc = (kite_opcode_t*)ALLOCATE_OPCODE(kite_opcode_nop);
    opc->opcode = NOP;
    return opc;
}

/*****************************************************************************
 * Compile a return operation
 ****************************************************************************/
kite_opcode_t *kite_compile_return()
{
    kite_opcode_t *opc = (kite_opcode_t*)ALLOCATE_OPCODE(kite_opcode_return);
    opc->opcode = RETURN_NOW;
    return opc;
}

/*****************************************************************************
 * Compile a defprop operation
 ****************************************************************************/
kite_opcode_t *kite_compile_defprop()
{
    kite_opcode_t *opc = (kite_opcode_t*)ALLOCATE_OPCODE(kite_opcode_defprop);
    opc->opcode = DEFPROP;
    return opc;
}

/*****************************************************************************
 * Compile is/isof instruction
 ****************************************************************************/
kite_opcode_t *kite_compile_objis(int isof)
{
    kite_opcode_objis *op = ALLOCATE_OPCODE(kite_opcode_objis);
    op->isof = isof;
    op->common.opcode = IS_ISOF;
    return (kite_opcode_t*)op;
}

/*****************************************************************************
 * Add instruction to instruction list.
 ****************************************************************************/
kite_opcode_t *
kite_add_to_instruction_list(kite_opcode_t **list, kite_opcode_t *opc)
{
    kite_opcode_t *cur = *list;
    
    if (*list == NULL) {
        *list = opc;
        return opc;
    }
    
    for( ; cur->next != NULL; cur = cur->next )
    {
        /* empty */
    }
    
    cur->next = opc;
    return opc;
}

/*****************************************************************************
 * Free memory associated with instruction list.
 ****************************************************************************/
void kite_free_instruction_list(kite_thread_t *thd, kite_opcode_t *opc)
{
    kite_opcode_t *tmp;
    (void)thd;
    
    while(opc != NULL)
    {
        tmp = opc;
        opc = opc->next;
#ifndef HAVE_GC_H
        free(tmp->file);
        free(tmp);
#endif /* HAVE_GC_H */
    }
}

/*****************************************************************************
 * Copy instruction list
 ****************************************************************************/
kite_opcode_t *kite_copy_instruction_list(kite_thread_t *thd, kite_opcode_t *opc)
{
    kite_opcode_t *ret, *cur;
    (void)thd;
    
#ifndef HAVE_GC_H
    ret = (kite_opcode_t*)malloc(opc->length);
#else
    ret = (kite_opcode_t*)GC_malloc(opc->length);
#endif /* HAVE_GC_H */
    memcpy(ret, opc, opc->length);
#ifndef HAVE_GC_H
    if (ret->file) ret->file = strdup(opc->file);
#else
    if (ret->file)
    {
        ret->file = GC_malloc_atomic(strlen(opc->file) + 1);
        strcpy(ret->file, opc->file);
    }
#endif /* HAVE_GC_H */
    cur = ret;
    while(cur->next) {
#ifndef HAVE_GC_H
        cur->next = (kite_opcode_t*)malloc(opc->next->length);
#else
        cur->next = (kite_opcode_t*)GC_malloc(opc->next->length);
#endif /* HAVE_GC_H */
        memcpy(cur->next, opc->next, opc->next->length);
#ifndef HAVE_GC_H
        if (cur->next->file) cur->next->file = strdup(opc->next->file);
#else
        if (cur->next->file)
        {
            cur->next->file = GC_malloc_atomic(strlen(opc->next->file) + 1);
            strcpy(cur->next->file, opc->next->file);
        }
#endif /* HAVE_GC_H */
        cur = cur->next;
        opc = opc->next;
    }

    return ret;
}

