/*****************************************************************************
 * Copyright (c) 2007, Mooneer Salem
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Kite Language organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/

#include <stdlib.h>
#include <assert.h>
#include "objs/kite_object.h"
#include "objs/kite_gc.h"
#include "kite_vm.h"

#ifdef WIN32
#define WIN32_LEAN_AND_MEAN
//#define _WIN32_WINNT 0x501
#include <winsock2.h>
#include <ws2tcpip.h>
#endif

KITE_CLASS_METHOD(VM_InstructionTracer)
{
    kite_object_t *arg;

    (void)this;
    KITE_GET_METHOD_ARGUMENT(arg, 1);

    kite_dereference_object(thd->vm->instruction_tracer);
    thd->vm->instruction_tracer = kite_reference_object(arg);
}

#ifdef HAVE_GC_H
static void kite_ignore_gc_warnings(char *msg, void *arg)
{
    /* ignore */
    (void)msg;
    (void)arg;
}
#endif /* HAVE_GC_H */

/*****************************************************************************
 * Initialize Kite.
 ****************************************************************************/
void kite_app_init()
{
#ifdef HAVE_GC_H
    GC_init();
#endif /* HAVE_GC_H */
}

/*****************************************************************************
 * Create a new instance of the VM
 ****************************************************************************/
kite_vm_t *kite_new_vm(char **args)
{
    kite_vm_t *ret;
    kite_thread_t *thd;
    int i = 0;
#ifdef WIN32
    WSADATA wsaData;
    int iResult;
#endif /* WIN32 */
#ifndef HAVE_GC_H
    ret = (kite_vm_t*)calloc(1, sizeof(kite_vm_t));
#else
    ret = (kite_vm_t*)GC_malloc(sizeof(kite_vm_t));
#endif /* HAVE_GC_H */
    if (!ret) return NULL;
    
    ret->numthreads = 0;
    ret->thread_list = NULL;
    thd = kite_new_thread_without_start(ret);
    ret->root_package = kite_new_class(thd, NULL, "__root");
    ret->instruction_tracer = NULL;
    ret->args = args;
#ifndef HAVE_GC_H
    ret->gc_begin = ret->gc_end = NULL;
#endif /* HAVE_GC_H */
#ifndef WIN32
    pthread_mutex_init(&ret->gc_mtx, NULL);
#else
    ret->gc_mtx = CreateMutex(NULL, FALSE, NULL);
#endif /* WIN32 */

    /* Clear signals */
#ifndef WIN32
    for(i = 0; i < NSIG; i++) {
        if (i == SIGPROF) continue;
        signal(i, SIG_DFL);
    }
#endif

    /* Force linkage of built-in modules */
    kite_module_initialize_now(thd);
    thd->vm->true_cached = kite_reference_object(kite_new_boolean(thd, TRUE));
    thd->vm->false_cached = kite_reference_object(kite_new_boolean(thd, FALSE));
    thd->vm->null_cached = kite_reference_object(kite_new_null(thd));
    thd->vm->int_cached = kite_dereference_and_load(thd, "System.integer");
    
    /* Force load System and interface (to ensure everything gets loaded). */
    kite_dereference_and_load(thd, "System");

    kite_add_method(thd, ret->root_package, "__instruction_tracer", 
        kite_new_method_compiled_with_docs(thd, VM_InstructionTracer, 
                                  "Set method for instruction tracing.", 1,
                                  "mthd", "The method to use (or null to disable)"));

#ifdef WIN32
    /* Win32: initialize Winsock. */
    iResult = WSAStartup(MAKEWORD(2,2), &wsaData);
    if (iResult != 0) {
        printf("WSAStartup failed: %d\n", iResult);
        return NULL;
    }
#endif

#ifdef HAVE_GC_H
    ret->old_proc = GC_set_warn_proc(kite_ignore_gc_warnings);
#endif /* HAVE_GC_H */

    return ret;
}

/*****************************************************************************
 * Destroys a VM
 ****************************************************************************/
void kite_free_vm(kite_vm_t **vm)
{
    kite_vm_t *cur;
    kite_thread_t *thd;
#ifndef WIN32
    int i;
#endif /* WIN32 */

    assert(vm != NULL && (*vm) != NULL);

#ifndef WIN32
    /* Clear signals */
    for(i = 0; i < NSIG; i++) {
        if (i == SIGPROF) continue;
        signal(i, SIG_DFL);
    }
#endif

    /* Kill all threads */
    cur = *vm;
    for ( ; cur->numthreads > 0; cur->numthreads--)
    {
        kite_exit_thread(cur, cur->thread_list[cur->numthreads - 1]);
    }

    /* Begin garbage collection until everything's collected */
    thd = kite_new_thread_without_start(*vm);
    kite_gc_destroy_all(thd);
    kite_destruct_object(thd, &(*vm)->root_package, FALSE);
    kite_destruct_symtab(thd, &cur->loader_refs, 0);
    kite_exit_thread(cur, thd);

#ifndef WIN32
    pthread_mutex_destroy(&cur->gc_mtx);
#else
    CloseHandle(cur->gc_mtx);
#endif /* WIN32 */

#ifndef HAVE_GC_H
    free(cur);
#else
    GC_set_warn_proc(cur->old_proc);
#endif /* HAVE_GC_H */
    *vm = NULL;
}

